package testprocess;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.annotation.Resource;
import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.Filer;
import javax.annotation.processing.Messager;
import javax.annotation.processing.ProcessingEnvironment;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.Element;
import javax.lang.model.element.TypeElement;
import javax.tools.Diagnostic.Kind;
import javax.tools.FileObject;
import javax.tools.JavaFileManager;
import javax.tools.StandardLocation;
import javax.tools.JavaFileManager.Location;
import javax.tools.JavaFileObject;

import at.cosimo.verocl.metamodel.manager.EcoreVM;
import at.cosimo.verocl.util.ecore.AnnotationIface;

@SupportedAnnotationTypes(value={"*"})
@SupportedSourceVersion(SourceVersion.RELEASE_5)
public class MyProcessor extends AbstractProcessor {
	
	private Filer filer;
	private Messager messager;
	
	static {
		try {
			EcoreVM.class.newInstance();
		} catch (InstantiationException | IllegalAccessException e) {}
	}
	
	//private ManagerManager manager = new ManagerManager();
	
	@Override
	public void init(ProcessingEnvironment env) {
		filer = env.getFiler();
		messager = env.getMessager();
	}
	
	public boolean deleteEmptySubs(File file) {
		boolean nonDir = false;
		for (File f: file.listFiles()) {
			if (!f.isDirectory()) {
				nonDir = true;
			} else {
				if (!deleteEmptySubs(f)) {
					nonDir = true;
				}
			}
		}
		if (!nonDir) {
			file.delete();
			return true;
		}
		return false;
	}
	
	@Override
	public boolean process(Set<? extends TypeElement> elements, RoundEnvironment env) {
		ByteArrayOutputStream errorMsg = new ByteArrayOutputStream();
		PrintWriter errorWriter = new PrintWriter(errorMsg);
		try {
		StringBuilder input = new StringBuilder();

		/*String myType = null;
		
		for (Element el: env.getRootElements()) {
			if (el instanceof TypeElement) {
				TypeElement tel = (TypeElement)el;
				myType = tel.getQualifiedName().toString();
			}
		}
		
		if (myType == null) {
			StringBuilder builder = new StringBuilder();
			builder.append("Have root elements: ");
			for (Element el: env.getRootElements()) {
				builder.append("Root element " + el.getSimpleName() +" of class " + el.getClass().getCanonicalName()+"\n");
			}
			messager.printMessage(Kind.ERROR, builder.toString());
			return true;
		}*/
		
		messager.printMessage(Kind.NOTE, "Starting processing");
		
		Map<String,String> annotationNames = new HashMap<>();

		Set<String> possibleDelete = new HashSet<>();
		
		Set<String> myTypes = new HashSet<String>();
		
		boolean haveAnnot = false;
	
		Map<String,Element> elementMap = new HashMap<>();
		FileObject res = null;
		try {
			res = filer.getResource(StandardLocation.SOURCE_OUTPUT, "", "id.txt");
		} catch (IOException e1) {
			e1.printStackTrace(errorWriter);
			errorWriter.flush();
			messager.printMessage(Kind.ERROR, "Exception m2:  " + errorMsg.toString() + " as " + e1.getMessage() );
		}
		
		String basePath = res.toUri().toString();
		if (basePath.indexOf('/') != -1) {
			basePath = basePath.substring(0,basePath.lastIndexOf('/'));
			if (basePath.indexOf('/') != -1) {
				basePath = basePath.substring(0,basePath.lastIndexOf('/'));
			}
		}
		
		for (Element elem: env.getElementsAnnotatedWith(EcoreVM.class)) {
			if (elem.getAnnotation(EcoreVM.class) != null) {
				messager.printMessage(Kind.WARNING, "My path is " + basePath,elem);
				haveAnnot = true;
				EcoreVM annotation = elem.getAnnotation(EcoreVM.class);
				Element encl = elem;
				TypeElement tencl = null;
				while (encl.getEnclosingElement() != null) {
					if (encl instanceof TypeElement) {
						tencl = (TypeElement)encl;
					}
					encl = encl.getEnclosingElement();
				}
				if (tencl == null) {
					messager.printMessage(Kind.ERROR, "No enclosing type element for " + elem.getSimpleName() + " found",elem);
				} else {
					messager.printMessage(Kind.WARNING, "Enclosing element " + tencl.getSimpleName() + " of Class " + tencl.getClass().getCanonicalName(),elem);
				}
				
				String realMyType = tencl.getQualifiedName().toString();
				myTypes.add(realMyType);
			//}
			//for (EcoreVM annotation: elem.getAnnotationsByType(EcoreVM.class)) {
				StringBuilder builder = new StringBuilder();
				builder.append("Annotation " + annotation.name() + " found: Merging ");
				if (annotationNames.containsKey(annotation.name())) {
					messager.printMessage(Kind.ERROR, "Duplicate annotation name " + annotation.name() + " in file " + annotationNames.get(annotation.name()),elem);
				}  else {
					annotationNames.put(annotation.name(), realMyType);
				}
				boolean first = true;
				for (String ecore: annotation.ecoreFiles()) {
					if (first) {first = false;} else {builder.append(", ");}
					
					builder.append(ecore);
					
				}
				elementMap.put(annotation.name(), elem);
				messager.printMessage(Kind.WARNING, builder.toString(),elem);
				messager.printMessage(Kind.WARNING, builder.toString());
				input.append("+"+annotation.name()+"\n");
				for (String ecore: annotation.ecoreFiles()) {
					if (!ecore.startsWith("/")) {
						ecore = basePath+"/"+ecore; 
					}
					input.append("#"+ecore+"\n");
				}
			}
		}
		
		String sourcePath = "";
		try {
			BufferedReader reader = new BufferedReader(res.openReader(true));
			String s;
			while ((s = reader.readLine()) != null) {
				if (s.isEmpty()) {continue;}
				String[] pre = s.split("@");
				//Long fileDate = Long.valueOf(pre[1]);
				String[] annot = pre[0].split(":");
				if (!myTypes.contains(annot[1])) {
					if (annotationNames.containsKey(annot[0])) {
						messager.printMessage(Kind.ERROR, "Duplicate annotation name " + annot[0] + " in file " + annotationNames.get(annot[0]),elementMap.get(annot[0]));
					} else {
						annotationNames.put(annot[0],annot[1]);
					}
				} else {
					possibleDelete.add(annot[0]);
				}
			}
		} catch (IOException e1) {
			e1.printStackTrace(errorWriter);
			errorWriter.flush();
			messager.printMessage(Kind.ERROR, "Exception m2:  " + errorMsg.toString() + " as " + e1.getMessage() );
		}
		
		

		
		if (!haveAnnot) {
			return false;
		}
		
		possibleDelete.removeAll(annotationNames.keySet());
		
		/*try {
			res = filer.createResource(StandardLocation.SOURCE_OUTPUT, "", "id.txt");
			StringBuilder builder = new StringBuilder();
			for (Entry<String,String> entry: annotationNames.entrySet()) {
				builder.append(entry.getKey()+":"+entry.getValue());
				builder.append("\n");
			}
			Writer writer = res.openWriter();
			writer.append(builder.toString());
			writer.flush();
			writer.close();
		} catch (IOException e1) {
			e1.printStackTrace(errorWriter);
			errorWriter.flush();
			messager.printMessage(Kind.ERROR, "Exception m3:  " + errorMsg.toString() + " as " + e1.getMessage() );
		}*/
		
		input.append("\n");
		
		for (String infoStr: possibleDelete) {
			messager.printMessage(Kind.NOTE, "Deleting basePkg " + infoStr);

			try {
				res = filer.getResource(StandardLocation.SOURCE_OUTPUT, infoStr, "info.txt");
				BufferedReader reader = new BufferedReader(res.openReader(true));
				String s;
				Set<String> classes = new HashSet<>();
				int curNum = 1;
				while ((s = reader.readLine()) != null) {
					if (s.isEmpty()) {continue;}
					if (s.startsWith("+")) {
						classes.add(infoStr+".virtual.Factoryvt"+(curNum++));
					} else if (s.startsWith("<")) {
						String[] spl = s.substring(1).split("<");
						classes.add(spl[0]);
						classes.add(spl[1]);
					} else if (s.startsWith(":")) {
						String[] spl = s.substring(1).split(":",2);
						classes.add(spl[0]);
					}
				}
				classes.add(infoStr+".BasicObject");
				classes.add(infoStr+".Manager");
				classes.add(infoStr+".virtual.ModelFactory");
				for (String cl: classes) {
					String pkg = "";
					int lind = cl.lastIndexOf('.');
					if (lind != -1) {
						pkg = cl.substring(0,lind);
						cl = cl.substring(lind+1);
					}
					FileObject fo = filer.getResource(StandardLocation.SOURCE_OUTPUT, pkg,cl+".java");
					messager.printMessage(Kind.NOTE, "Deleting file " + fo.toUri());
					File file = new File(fo.toUri());
					if (file.exists()) {
						file.delete();
					}
					
				}
				reader.close();
				File file = new File(res.toUri());
				messager.printMessage(Kind.NOTE, "Deleting file " + res.toUri());
				if (file.exists()) {
					file.delete();
				}
				file = file.getParentFile();
				deleteEmptySubs(file);
			} catch (IOException e1) {
				e1.printStackTrace(errorWriter);
				errorWriter.flush();
				messager.printMessage(Kind.ERROR, "Exception m4:  " + errorMsg.toString() + " as " + e1.getMessage() );
			}
		}
		
		for (String infoStr: annotationNames.keySet()) {
			try {
				//Paketname, verwendete Ecore-Dateien
				//+Paketname
				//#Ecore-File1...
				//#Ecore-Filen
				//:fullyqualifiedname:EClassURI
				//<subclass<superclass
				//
				res = filer.getResource(StandardLocation.SOURCE_OUTPUT, infoStr, "info.txt");
				input.append("?="+infoStr+"\n");
				BufferedReader reader = new BufferedReader(res.openReader(true));
				String s;
				List<String> ecoreFiles = new ArrayList<>();
				Map<String,String> eclassUris = new HashMap<>();
				Map<String,Collection<String>> subToSuper = new HashMap<>();
				while ((s = reader.readLine()) != null) {
					if (s.isEmpty()) {continue;}
					if (s.startsWith("+")) {
						ecoreFiles.add(s.substring(1));
						input.append("?"+s+"\n");
					} else if (s.startsWith("<")) {
						String[] spl = s.substring(1).split("<");
						Collection<String> sup = subToSuper.get(spl[0]);
						if (sup == null) {
							subToSuper.put(spl[0],sup = new ArrayList<>());
						}
						sup.add(spl[1]);
						input.append("?"+s+"\n");
					} else if (s.startsWith(":")) {
						String[] spl = s.substring(1).split(":",2);
						eclassUris.put(spl[0], spl[1]);
						input.append("?"+s+"\n");
					}
				}
				
			} catch (IOException e1) {
				e1.printStackTrace(errorWriter);
				errorWriter.flush();
				messager.printMessage(Kind.ERROR, "Exception m5:  " + errorMsg.toString() + " as " + e1.getMessage() );
			}
		}
		
		StringBuilder output = new StringBuilder();
		try {
			ByteArrayInputStream bis = new ByteArrayInputStream(input.toString().getBytes());
			BufferedReader br = new BufferedReader(new InputStreamReader(bis));
			StringWriter sw = new StringWriter();
			BufferedWriter wri = new BufferedWriter(sw);
			AnnotationIface.doMain(br, wri);
			br.close();
			wri.flush();
			wri.close();
			messager.printMessage(Kind.NOTE, "Done main: " + sw.toString());
			//p = Runtime.getRuntime().exec("java -jar \"" + Paths.get("lib"+File.separator+"annot.jar").toAbsolutePath().toString()+"\" \""+StandardLocation.SOURCE_OUTPUT+"\"");
			/*InputStream is = p.getInputStream();
			
			BufferedReader reader = new BufferedReader(new InputStreamReader(is));*/
			BufferedReader reader = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(sw.getBuffer().toString().getBytes())));
			
			/*
			OutputStream os = p.getOutputStream();
			os.write(input.toString().getBytes());
			os.flush();
			os.close();*/
			
			String s;
			StringBuilder cur = new StringBuilder();
			Map<String,String> packageToJavaMap = new HashMap<>();
			String curName = null;
			while ((s = reader.readLine()) != null) {
				output.append(s);
				output.append("\n");
				if (s.isEmpty()) {continue;}
				if (s.startsWith("+")) {
					if (curName != null) {
						packageToJavaMap.put(curName, cur.toString());
						cur = new StringBuilder();
					}
					curName = s.substring(1);
					//idwriter.append(curName+"\n");
				} else if (s.startsWith("#")) {
					cur.append(s.substring(1));
					cur.append("\n");
				}
				
			}
			
			if (curName != null) {
				packageToJavaMap.put(curName, cur.toString());
			}
			
			/*BufferedReader errReader = new BufferedReader(new InputStreamReader(p.getErrorStream()));
			
			while ((s = errReader.readLine()) != null) {
				output.append(s);
				output.append("\n");
			}*/
			
			
			

			
			for (Entry<String,String> entry: packageToJavaMap.entrySet()) {
				messager.printMessage(Kind.NOTE, "Writing source file " + entry.getKey()+":\n"+entry.getValue());
				if (entry.getKey().endsWith("info-txt")) {
					int lastPoint = entry.getKey().lastIndexOf('.');
					String pkg = "";
					String name = entry.getKey().replace('-', '.');
					if (lastPoint != -1) {
						pkg = entry.getKey().substring(0,lastPoint);
						name = entry.getKey().substring(lastPoint+1).replace('-', '.');
					}
					FileObject fo = filer.createResource(StandardLocation.SOURCE_OUTPUT, pkg, name);
					Writer writer = fo.openWriter();
					writer.write(entry.getValue());
					writer.flush();
					writer.close();
				} else {
					JavaFileObject jfo = filer.createSourceFile(entry.getKey());
					Writer writer = jfo.openWriter();
					writer.write(entry.getValue());
					
					writer.flush();
					writer.close();
				}
			}
			
			messager.printMessage(Kind.WARNING, output.toString());
			
		} catch (IOException e) {
			e.printStackTrace(errorWriter);
			errorWriter.flush();
			messager.printMessage(Kind.ERROR, "Exception l1:  " + errorMsg.toString() + " as " + e.getMessage() );
		}	
		} catch (Exception e) {
			e.printStackTrace(errorWriter);
			errorWriter.flush();
			messager.printMessage(Kind.ERROR, "Exception l2:  " + errorMsg.toString() + " as " + e.getMessage() );
		}
		return false;		
	}
	
	public static void main(String[] args) {
		new MyProcessor();
	}
	
}
