package at.cosimo.verocl;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.compare.Comparison;
import org.eclipse.emf.compare.Conflict;
import org.eclipse.emf.compare.EMFCompare;
import org.eclipse.emf.compare.Match;
import org.eclipse.emf.compare.scope.DefaultComparisonScope;
import org.eclipse.emf.compare.scope.IComparisonScope;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.EcoreResourceFactoryImpl;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;

import at.cosimo.verocl.metamodel.ManagerManager;
import at.cosimo.verocl.metamodel.MergeClass;
import at.cosimo.verocl.metamodel.MergeManager;
import at.cosimo.verocl.metamodel.output.OutputClass;
import at.cosimo.verocl.metamodel.output.OutputManager;
import at.cosimo.verocl.metamodel.output.OutputPackage;
import at.cosimo.verocl.metamodel.output.PseudoOutputManager;
import at.cosimo.verocl.util.ecore.GeneralTraverser;
import at.cosimo.verocl.util.ecore.GeneralTraverser.TraverserIFace;

public class SimpleMerge {
	
	{
	
		
	}
	
	private ResourceSet ecoreSet;
	private List<Resource> allEcores = new ArrayList<>();
	private Resource modelResource;
	
	GeneralTraverser traverser = new GeneralTraverser();
	
	static Map<String,ObjectMerger> objectMergerForClass = new HashMap<>();
	
	public static interface ObjectMerger {
		public EObject mergeObjects(EObject ret, Collection<EObject> objects, EClass cl, Map<EObject,EObject> mapped);
	}
	
	static {
		objectMergerForClass.put("EClass",
				new ObjectMerger() {
					
					@Override
					public EObject mergeObjects(EObject ret, Collection<EObject> objects, EClass cl, Map<EObject,EObject> mapped) {
						for (EStructuralFeature feat: cl.getEAllStructuralFeatures()) {
							if ("interface".equals(feat.getName()) || "abstract".equals(feat.getName())) {
								boolean myFeat = true;
								for (EObject obj: objects) {
									myFeat&= (Boolean)obj.eGet(feat);
								}
								ret.eSet(feat, myFeat);
							}
							if ("name".equals(feat.getName()) || "instanceClassName".equals(feat.getName()) ||
									"instanceTypeName".equals(feat.getName()) || "defaultValue".equals(feat.getName())
									|| "instanceClass".equals(feat.getName())) {
								for (EObject obj: objects) {
									Object have = obj.eGet(feat);
									if (have != null) {
										ret.eSet(feat, have);
										break;
									}
								}								
							}
							if ("ePackage".equals(feat.getName())) {
								//TODO: Erstelle eigenes
							}
							if ("eIDAttribute".equals(feat.getName())) {
								//TODO: Irgendwie zusamenbasteln
							}
							if ("eTypeParameters".equals(feat.getName())) {
								//TODO: Was macht man da?
							}
							if ("eOperations".equals(feat.getName())) {
							}
							if ("eStructuralFeatures".equals(feat.getName())) {
								Set<EAttribute> superTypes = new HashSet<>();
								for (EObject eobj: objects) {
									for (EAttribute ecl: (EList<EAttribute>)eobj.eGet(feat)) {
										superTypes.add((EAttribute)mapped.get(ecl));
									}
								}
								EList<EAttribute> eClassList = new BasicEList<>(); 
								eClassList.addAll(superTypes);
								ret.eSet(feat, eClassList);
							}/*
							if ("eAttributes".equals(feat.getName())) {
								Set<EAttribute> superTypes = new HashSet<>();
								for (EObject eobj: objects) {
									for (EAttribute ecl: (EList<EAttribute>)eobj.eGet(feat)) {
										superTypes.add((EAttribute)mapped.get(ecl));
									}
								}
								EList<EAttribute> eClassList = new BasicEList<>(); 
								eClassList.addAll(superTypes);
								ret.eSet(feat, eClassList);
							}
							if ("eReferences".equals(feat.getName())) {
								Set<EReference> superTypes = new HashSet<>();
								for (EObject eobj: objects) {
									for (EReference ecl: (EList<EReference>)eobj.eGet(feat)) {
										superTypes.add((EReference)mapped.get(ecl));
									}
								}
								//TODO: Eigenschaften der Referenz setzen - wobei: Das wird wohl woanders gemacht
								EList<EReference> eClassList = new BasicEList<>(); 
								eClassList.addAll(superTypes);
								ret.eSet(feat, eClassList);
							}*/
							if ("eAnnotations".equals(feat.getName())) {
								
							}
							if ("eSuperTypes".equals(feat.getName())) {
								Set<EClass> superTypes = new HashSet<>();
								for (EObject eobj: objects) {
									for (EClass ecl: (EList<EClass>)eobj.eGet(feat)) {
										superTypes.add((EClass)mapped.get(ecl));
									}
								}
								EList<EClass> eClassList = new BasicEList<>(); 
								eClassList.addAll(superTypes);
								ret.eSet(feat, eClassList);
							}
						}
						
						return ret;
					}
				}
				);
		objectMergerForClass.put("EAttribute",
				new ObjectMerger() {
					
					@Override
					public EObject mergeObjects(EObject ret, Collection<EObject> objects, EClass cl, Map<EObject,EObject> mapped) {
						for (EStructuralFeature feat: cl.getEAllStructuralFeatures()) {
							if ("unique".equals(feat.getName()) || "id".equals(feat.getName())) {
								boolean myFeat = true;
								for (EObject obj: objects) {
									myFeat&= (Boolean)obj.eGet(feat);
								}
								ret.eSet(feat, myFeat);
							}
							if ("changeable".equals(feat.getName())) {
								//Eigentlich auch dann nicht, wenn es in mind. einem nicht vorkommt
								boolean chgable = true;
								for (EObject obj: objects) {
									chgable&= (Boolean)obj.eGet(feat);
								}
								ret.eSet(feat, chgable);
							}
							if ("ordered".equals(feat.getName())) {
								boolean myFeat = false;
								for (EObject obj: objects) {
									myFeat|= (Boolean)obj.eGet(feat);
								}
								ret.eSet(feat, myFeat);
							}
							if ("lowerBound".equals(feat.getName())) {
								int minValue = Integer.MAX_VALUE;
								for (EObject obj: objects) {
									minValue = Math.min(minValue,(Integer)obj.eGet(feat));
								}
								ret.eSet(feat, minValue);
							}
							if ("upperBound".equals(feat.getName())) {
								int maxValue = 0;
								for (EObject obj: objects) {
									Integer upBound = (Integer)obj.eGet(feat);
									if (upBound == -1) {
										maxValue = -1;
									} else if (upBound > maxValue && maxValue >= 0) {
										maxValue = upBound;
									}
								}								
								ret.eSet(feat, maxValue);
							}
							/*if ("eContainingClass".equals(feat.getName())) {
								//TODO: Das muss angepasst werden, aber ich bezweifle dass das hier richtig ist
								for (EObject obj: objects) {
									Object have = obj.eGet(feat);
									if (have != null) {
										if (mapped.get(have) != null) {
											have = mapped.get(have);
										}
										ret.eSet(feat, have);
										break;
									}
								}		
							}*/
							if ("eType".equals(feat.getName())/* || "eGenericType".equals(feat.getName())*/) {
								//TODO: Nimm den "obertyp"
								//TODO: Das muss angepasst werden, aber ich bezweifle dass das hier richtig ist
								//TODO: Der GenericType macht irgendwie nicht das was ich will - vielleicht weil er unvollstndig ist?
								for (EObject obj: objects) {
									Object have = obj.eGet(feat);
									System.out.println("Have " + feat.getName() + ": " + have);
									if (have != null) {										
										if (mapped.get(have) != null) {
											have = mapped.get(have);
										}
										System.out.println("Which is later : " + have);
										ret.eSet(feat, have);
										break;
									}
								}							
							}
							if ("name".equals(feat.getName()) /*|| "instanceClassName".equals(feat.getName()) ||
									"instanceTypeName".equals(feat.getName()) || "defaultValue".equals(feat.getName())
									|| "instanceClass".equals(feat.getName())*/) {
								for (EObject obj: objects) {
									Object have = obj.eGet(feat);
									if (have != null) {
										ret.eSet(feat, have);
										break;
									}
								}								
							}
							
							if ("ePackage".equals(feat.getName())) {
								//TODO: Erstelle eigenes
							}
							if ("eIDAttribute".equals(feat.getName())) {
								//TODO: Irgendwie zusamenbasteln
							}
							if ("eTypeParameters".equals(feat.getName())) {
								//TODO: Was macht man da?
							}

							if ("eAnnotations".equals(feat.getName())) {
								
							}

						}
						
						return ret;
					}
				}
				);
		objectMergerForClass.put("EReference",
				new ObjectMerger() {
					
					@Override
					public EObject mergeObjects(EObject ret, Collection<EObject> objects, EClass cl, Map<EObject,EObject> mapped) {
						for (EStructuralFeature feat: cl.getEAllStructuralFeatures()) {
							if ("unique".equals(feat.getName()) || "id".equals(feat.getName())) {
								boolean myFeat = true;
								for (EObject obj: objects) {
									myFeat&= (Boolean)obj.eGet(feat);
								}
								ret.eSet(feat, myFeat);
							}
							if ("changeable".equals(feat.getName())) {
								//Eigentlich auch dann nicht, wenn es in mind. einem nicht vorkommt
								boolean chgable = true;
								for (EObject obj: objects) {
									chgable&= (Boolean)obj.eGet(feat);
								}
								ret.eSet(feat, chgable);
							}
							if ("ordered".equals(feat.getName())) {
								boolean myFeat = false;
								for (EObject obj: objects) {
									myFeat|= (Boolean)obj.eGet(feat);
								}
								ret.eSet(feat, myFeat);
							}
							if ("lowerBound".equals(feat.getName())) {
								int minValue = Integer.MAX_VALUE;
								for (EObject obj: objects) {
									minValue = Math.min(minValue,(Integer)obj.eGet(feat));
								}
								ret.eSet(feat, minValue);
							}
							if ("upperBound".equals(feat.getName())) {
								int maxValue = 0;
								for (EObject obj: objects) {
									Integer upBound = (Integer)obj.eGet(feat);
									if (upBound == -1) {
										maxValue = -1;
									} else if (upBound > maxValue && maxValue >= 0) {
										maxValue = upBound;
									}
								}								
								ret.eSet(feat, maxValue);
							}
							if ("eContainingClass".equals(feat.getName())) {
								//TODO: Das muss angepasst werden, aber ich bezweifle dass das hier richtig ist
								for (EObject obj: objects) {
									Object have = obj.eGet(feat);
									if (have != null) {
										if (mapped.get(have) != null) {
											have = mapped.get(have);
										}
										ret.eSet(feat, have);
										break;
									}
								}		
							}
							if ("eType".equals(feat.getName()) || "eGenericType".equals(feat.getName())||
									"eReferenceType".equals(feat.getName())) {
								//TODO: Nimm den "obertyp"
								//TODO: Das muss angepasst werden, aber ich bezweifle dass das hier richtig ist
								for (EObject obj: objects) {
									Object have = obj.eGet(feat);
									if (have != null) {
										if (mapped.get(have) != null) {
											have = mapped.get(have);
										}
										ret.eSet(feat, have);
										break;
									}
								}							
							}
							if ("name".equals(feat.getName()) || "instanceClassName".equals(feat.getName()) ||
									"instanceTypeName".equals(feat.getName()) || "defaultValue".equals(feat.getName())
									|| "instanceClass".equals(feat.getName())) {
								for (EObject obj: objects) {
									Object have = obj.eGet(feat);
									if (have != null) {
										ret.eSet(feat, have);
										break;
									}
								}								
							}
							
							if ("ePackage".equals(feat.getName())) {
								//TODO: Erstelle eigenes
							}
							if ("eIDAttribute".equals(feat.getName())) {
								//TODO: Irgendwie zusamenbasteln
							}
							if ("eTypeParameters".equals(feat.getName())) {
								//TODO: Was macht man da?
							}

							if ("eAnnotations".equals(feat.getName())) {
								
							}

						}
						
						return ret;
					}
				}
				);
		objectMergerForClass.put("EGenericType",
				new ObjectMerger() {
					
					@Override
					public EObject mergeObjects(EObject ret, Collection<EObject> objects, EClass cl, Map<EObject,EObject> mapped) {
						for (EStructuralFeature feat: cl.getEAllStructuralFeatures()) {
							
							if ("eTypeArguments".equals(feat.getName())) {
								//TODO: Irgendwie zusamenbasteln
							}
							if (/*"eRawType".equals(feat.getName()) || */"eLowerBound".equals(feat.getName()) || 
									"eUpperBound".equals(feat.getName())) {
								//TODO: Was macht man da?
								for (EObject obj: objects) {
									Object have = obj.eGet(feat);
									if (have != null) {
										if (mapped.get(have) != null) {
											have = mapped.get(have);
										}
										ret.eSet(feat, have);
										break;
									}
								}	
							}

							if ("eTypeParameter".equals(feat.getName())) {
								//TODO: Irgendwie zusamenbasteln
							}
							if ("eClassifier".equals(feat.getName())) {
								//TODO: ??
							}

						}
						
						return ret;
					}
				}
				);
	}
	
	
	public void mergeObjects(Map<EObject,EObject> mapped, Map<EObject,Map<EObject,Resource>> equivalenceClass) {
		
		for (Entry<EObject,Map<EObject,Resource>> entry: equivalenceClass.entrySet()) {
			Map<EObject,Resource> objectMap = entry.getValue();
			//Ensure all have same EClass
			EObject ret = null;
			EClass cl = null;
			for (EObject o: objectMap.keySet()) {
				if (ret == null) {
					ret = mapped.get(o);
				}
				if (o != null && o.eClass() != null) {
					if (cl != null && cl != o.eClass()) {
						throw new RuntimeException("Cannot merge objects of different class");
					}
					cl = o.eClass();
				}
			}
			if (cl == null) {
				continue;
			}
			ObjectMerger merger = objectMergerForClass.get(cl.getName());
			if (merger == null) {
				System.err.println("Unknown merger for Class " + cl.getName());
				continue;
			}
			
			merger.mergeObjects(ret, objectMap.keySet(), cl, mapped);
		}		 
//		//Attribute haben werte, Referenzen sind auf andere Modellelemente - aber hier sollte es ja ein gemeinsames
//		//Metamodell geben (hier Ecore, allgemeiner auch CSL usw.)
//		
//		for (EAttribute attr: cl.getEAllAttributes()) {
//			//Mglicherweise gibt die Ecore-API gar nicht alles notwendige her - hier gbe es ja
//			//Unterschiedliche Attribute etc., die man vielleicht so direkt gar nicht verbinden kann
//			//Die Attribute selbst (Also von einer eigenen Klasse die Attribute) kann man ja zusammenfgen, 
//			//Aber hier ist das Problem, dass man die Klasseneigenschaften selbst bestimmen msste, und das geht vielleicht
//			//nicht.
//		}
//		
//		for (EReference ref: cl.getEAllReferences()) {
//			
//		}
		
	}
	
	
	public static SimpleMerge createSimpleMerge(File xmiFile) {
		SimpleMerge ret = new SimpleMerge();
		ResourceSet resourceSet = ret.ecoreSet = new ResourceSetImpl();
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("xmi", new XMIResourceFactoryImpl());
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("ecore", new EcoreResourceFactoryImpl());
		
		URI mainEcoreURI = URI.createFileURI(new File("model"+File.separator+"verocl.ecore").getAbsolutePath());
		Resource mainEcore = ret.modelResource = resourceSet.createResource(mainEcoreURI);
		try {
			mainEcore.load(new HashMap<>());
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		for (EObject obj: mainEcore.getContents()) {
			if (obj instanceof EPackage) {
				EPackage pkg = (EPackage)obj;
				resourceSet.getPackageRegistry().put(pkg.getNsURI(),pkg);
				System.out.println("Register package with nsURI: " + pkg.getNsURI());
			}
		}
		
		Resource xmiResource = resourceSet.createResource(URI.createFileURI(xmiFile.getAbsolutePath()));
		try {
			xmiResource.load(new HashMap<>());
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		EObject obj = xmiResource.getContents().get(0);
		TraverserIFace tv = ret.traverser.traverser(obj);
		
		
		List<String> ecoreFileNames = (List<String>)(List)tv.attr("models").attr("filename").getValue();
		ManagerManager managermanager = new ManagerManager();
		MergeManager manager = managermanager.createManager("virtualpkg",ecoreFileNames.toArray(new String[]{}));
		List<String> lessEcore = ecoreFileNames.subList(0,1);//2, ecoreFileNames.size());
		MergeManager manager2 = managermanager.createManager("virtualpkg2",lessEcore.toArray(new String[]{}));
		
		
		
		OutputManager om = OutputManager.getOutput(manager, new OutputManager(), Collections.<PseudoOutputManager>emptyList());
		for (OutputPackage pkg: om.getOutputPkg()) {
			for (OutputClass ocl: pkg.getContainedClasses()) {
				StringBuilder outputbuilder = new StringBuilder();
				ocl.putOut(outputbuilder, 0);
				System.out.println(outputbuilder);
			}
		}
		
		OutputManager om2 = OutputManager.getOutput(manager2, new OutputManager(), Collections.<PseudoOutputManager>emptyList());
		for (OutputPackage pkg: om2.getOutputPkg()) {
			for (OutputClass ocl: pkg.getContainedClasses()) {
				StringBuilder outputbuilder = new StringBuilder();
				ocl.putOut(outputbuilder, 0);
				System.out.println(outputbuilder);
			}
		}

		
		try {
			om.outputIn(new File("src"));
			om2.outputIn(new File("src"));
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		return ret;
	}
	
	
	private static void printEObj(EObject eobj) {
		System.out.println(eobj.eClass().getName()+": {");
		for (EStructuralFeature feat: eobj.eClass().getEAllStructuralFeatures()) {
			System.out.println(feat.getName()+" = " + eobj.eGet(feat));
		}
		System.out.println("}");
	}


	public static void main(String[] args) {
		File mi = new File("model/ModelIndex.xmi");
		System.out.println(mi.getAbsolutePath()+" exists: " + mi.exists());
		SimpleMerge merge = createSimpleMerge(mi);
		
	}
}
