package at.cosimo.verocl.metamodel;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.compare.Comparison;
import org.eclipse.emf.compare.Conflict;
import org.eclipse.emf.compare.EMFCompare;
import org.eclipse.emf.compare.Match;
import org.eclipse.emf.compare.scope.DefaultComparisonScope;
import org.eclipse.emf.compare.scope.IComparisonScope;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.EcoreResourceFactoryImpl;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;

import at.cosimo.verocl.metamodel.output.OutputManager;
import at.cosimo.verocl.metamodel.output.PseudoOutputManager;

public class ManagerManager {

	
	private Map<Set<String>,MergeManager> mergeManagers = new HashMap<>();
	
	private ResourceSet resourceSet = new ResourceSetImpl(); {
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("xmi", new XMIResourceFactoryImpl());
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("ecore", new EcoreResourceFactoryImpl());
	}
	
	public static class EObjectPairMatch {
		public EObjectPairMatch(Resource first, Resource second, Match match) {
			this.first = first;
			this.second = second;
			this.match = match;
		}
		
		private Resource first;
		private Resource second;
		private Match match;
		
		public String toString() {
			return String.valueOf(match) + " between " + first + " and " + second;
		}
	}
	
	public MergeManager createManager(String packageName, String... ecoreFileNames) {
		List<ResourceSet> emfCompareSet = new ArrayList<>();
		List<Resource> emfCompareRSet = new ArrayList<>();
		List<Resource> allEcores = new ArrayList<>();
		for (String str: ecoreFileNames) {
			System.out.println("Model filename " + str);
			URI uri;
			if (str.contains(":/")) {
				uri = URI.createURI(str);
			} else {
				uri = URI.createFileURI(new File(str).getAbsolutePath());
			}
			Resource toLoadResource = resourceSet.getResource(uri,true);
			if (!toLoadResource.isLoaded()) {
				try {
					toLoadResource.load(new HashMap<>());
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
			ResourceSet emfComp = new ResourceSetImpl();
			emfComp.getResourceFactoryRegistry().getExtensionToFactoryMap().put("xmi", new XMIResourceFactoryImpl());
			emfComp.getResourceFactoryRegistry().getExtensionToFactoryMap().put("ecore", new EcoreResourceFactoryImpl());
			emfCompareSet.add(emfComp);
			emfCompareRSet.add(emfComp.getResource(uri, true));
			allEcores.add(toLoadResource);
		}
		
		EMFCompare comparator = EMFCompare.builder().build();
		
		Map<EObject,Map<Resource,EObject>> eobjectSet = new HashMap<>();
		Map<EObject,Map<EObject,Resource>> eobjectRSet = new HashMap<>();
		
		Map<EObject,EObject> createdMap = new HashMap<>();
		
		List<EObjectPairMatch> allMatches = new ArrayList<ManagerManager.EObjectPairMatch>(); 
		
		for (int i = 0; i < emfCompareRSet.size(); ++i) {
			Resource first = emfCompareRSet.get(i);
			for (int j = i+1; j < emfCompareRSet.size(); ++j) {
				Resource snd = emfCompareRSet.get(j);

				IComparisonScope scope = new DefaultComparisonScope(emfCompareSet.get(i), emfCompareSet.get(j), null);
				Comparison comp = comparator.compare(scope);
				for (Conflict conflict: comp.getConflicts()) {
					System.out.println("Conflict: " + conflict);
				}
				for (Match match: comp.getMatches()) {
					System.out.println("Match: " + match);
					for (Match submatch: match.getAllSubmatches()) {
						System.out.println("Submatch: " + submatch);
						allMatches.add(new EObjectPairMatch(first, snd, submatch));
					}
				}
			}
		}
		allMatches.sort(new Comparator<EObjectPairMatch>(){

			@Override
			public int compare(EObjectPairMatch o1, EObjectPairMatch o2) {
				int ret = Integer.compare(o1.match.getDifferences().size(),o2.match.getDifferences().size());
				if (ret != 0) {
					return ret;
				}
				ret = -Integer.compare(o1.match.getSubmatches().size(), o2.match.getSubmatches().size());
				return ret;
			}});
		for (EObjectPairMatch match: allMatches) {
			Resource first = match.first;
			Resource snd = match.second;
			Match submatch = match.match;
						
			if (submatch.getLeft() != null && submatch.getRight() != null) {
				Map<Resource,EObject> map1a = eobjectSet.get(submatch.getLeft());
				Map<EObject,Resource> map1b = eobjectRSet.get(submatch.getLeft());
				Map<Resource,EObject> map2a = eobjectSet.get(submatch.getRight());
				Map<EObject,Resource> map2b = eobjectRSet.get(submatch.getRight());
				
				//if (map1a != null && map2a != null && (map1a.containsKey(snd) || map2a.containsKey(first))) {
				//	//Es gibt schon einen Match, ignoriere das daher
				//	System.out.println("Ignoring match " + submatch);
				//} else {
					if (map1a == null || map2a == null) {
						if (map1a == null) {
							map1a = new HashMap<>();
							map1b = new HashMap<>();
							eobjectSet.put(submatch.getLeft(), map1a);
							eobjectRSet.put(submatch.getLeft(), map1b);
						}
						if (map2a == null) {
							map2a = new HashMap<>();
							map2b = new HashMap<>();
							eobjectSet.put(submatch.getRight(), map2a);
							eobjectRSet.put(submatch.getRight(), map2b);
						}
					}
					if (map1a != map2a) {
						int notCompatibleReason = -1;
						//Check compatibility
						boolean compatible = true;
						EClass cur = null;
						for (Entry<Resource,EObject> entr: map2a.entrySet()) {
							EObject frst = map1a.get(entr.getKey());
							if (frst != null && frst.eClass() != null) {
								cur = frst.eClass();
							}
							if (frst != null && entr.getValue() != null && frst != entr.getValue()) {
								compatible = false;
								notCompatibleReason = 1;
								break;
							}
						}
						for (EObject eobj: map1b.keySet()) {
							if (eobj != null && eobj.eClass() != null && cur != null && eobj.eClass() != cur) {
								compatible = false;
								notCompatibleReason = 2;
								break;
							}
						}
						//Ensure same eclass
						for (EObject eobj: map2b.keySet()) {
							if (eobj != null && eobj.eClass() != null && cur != null && eobj.eClass() != cur) {
								compatible = false;
								notCompatibleReason = 3;
								break;
							}
						}
						
						if (cur != null && submatch.getLeft().eClass() != null && submatch.getLeft().eClass() != cur) {
							compatible = false;
							notCompatibleReason = 4;
						}
						
						if (cur != null && submatch.getRight().eClass() != null && submatch.getRight().eClass() != cur) {
							compatible = false;
							notCompatibleReason = 5;
						}
						 
						if (compatible) {
							System.out.println("Compatible submatch " + submatch);
							eobjectSet.put(submatch.getRight(), map1a);
							eobjectRSet.put(submatch.getRight(), map1b);
							
							for (Entry<Resource,EObject> entr: map2a.entrySet()) {
								if (entr.getValue() != null) {
									map1a.put(entr.getKey(), entr.getValue());
								}
							}
							
							for (Entry<EObject,Resource> entr: map2b.entrySet()) {
								if (entr.getValue() != null) {
									map1b.put(entr.getKey(), entr.getValue());
								}
							}
							map1a.put(first, submatch.getLeft());
							map1b.put(submatch.getLeft(), first);
							map1a.put(snd, submatch.getRight());
							map1b.put(submatch.getRight(), snd);
						} else {
							System.out.println("Incompatible submatch " + submatch + " with reason " + notCompatibleReason);
						}
					} else {
						System.out.println("Already match " + submatch);
					}
				//}
			}
		}
		
		//Jetzt stehen in den Maps Gleichheiten drin, es gibt aber einiges, was nirgends vorkommt
		for (int i = 0; i < emfCompareRSet.size(); ++i) {
			Resource first = emfCompareRSet.get(i);
			Iterator<EObject> iter = first.getAllContents();
			while (iter.hasNext()) {
				EObject content = iter.next();
				if (eobjectSet.get(content) == null) {
					HashMap<EObject,Resource> hm = new HashMap<>();
					HashMap<Resource,EObject> hm2 = new HashMap<>();
					hm.put(content,first);
					hm2.put(first, content);
					eobjectRSet.put(content, hm);
					eobjectSet.put(content, hm2);					
				}
				
			}
		}
		
		//Jetzt stehen berall gleichheiten drin und man sollte mergen knnen ...
		//Kopiere einmal das eobjectRSet
		
	
		/*ret.mergeObjects(createdMap, eobjectRSet);		
		
		for (EObject eobj: createdMap.values()) {
			printEObj(eobj);
		}*/
		
		MergeManager ret = MergeManager.getMergeManager(packageName,allEcores, eobjectSet, eobjectRSet);
		return addMergeManager(ret);
	}
	
	public MergeManager addMergeManager(MergeManager man) {
		Set<String> resourceURI = new HashSet<String>();
		for (Resource res: man.getResources()) {
			resourceURI.add(res.getURI().toString());
		}
		MergeManager cur = mergeManagers.get(resourceURI);
		if (cur != null) {
			return cur;
		}
		for (Entry<Set<String>,MergeManager> entry: mergeManagers.entrySet()) {
			if (entry.getKey().containsAll(resourceURI)) {
				System.out.println("New sub of old ");
				//Vorhandenes ist subset von dem
				entry.getValue().addSubManager(man);
			} else if (resourceURI.containsAll(entry.getKey())) {
				System.out.println("Old sub of new");
				//Das ist subset von dem
				//Beides sollte nicht erfllt sein, weil sonst gleich und das sollte lt. oberem verhindert werden
				man.addSubManager(entry.getValue());
			}
		}
		
		
		
		mergeManagers.put(resourceURI, man);
		return man;
	}
	
	public static Map<MergeManager,OutputManager> ommap = new HashMap<MergeManager,OutputManager>();
	
	public static List<PseudoOutputManager> pseudooms = new ArrayList<>();
	
	public static void addPseudoOM(PseudoOutputManager om) {
		pseudooms.add(om);
	}
	
	public static void setPseudoOM(List<PseudoOutputManager> poms) {
		pseudooms = poms;
	}
	
	public static OutputManager getOrCreateOutput(MergeManager manager) {
		OutputManager ret = ommap.get(manager);
		if (ret == null) {
			ommap.put(manager,ret = new OutputManager());
			OutputManager.getOutput(manager, ret, pseudooms);
		}
		return ret;
	}
}
