package at.cosimo.verocl.metamodel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

public class MergeClass extends MergeClassifier {

	protected MergeClass(MergeManager manager, List<EClass> baseClasses) {
		super(manager);
		this.baseClasses = new ArrayList<>();
		this.baseClasses.addAll(baseClasses);
		this.baseClassSet = new HashSet<>();
		this.baseClassSet.addAll(baseClasses);
		for (EClass child: baseClasses) {
			for (EClass parent: child.getESuperTypes()) {
				addSuperType(parent, child);
			}
		}
	}
	
	public boolean isAbstract() {
		return abstract_;
	}
	public void setAbstract(boolean abstract_) {
		this.abstract_ = abstract_;
		EObject bla;
	}

	public boolean isInterface() {
		for (EClass baseClass: baseClasses) {
			if (!baseClass.isInterface()) {
				return false;
			}
		}
		return true;
	}
	
	private Set<EClass> baseClassSet = new HashSet<>();
	private List<EClass> baseClasses = new ArrayList<>();
	
	public Collection<EClass> getBase() {
		return baseClasses;
	}

	private boolean abstract_;
	private Collection<MergeClass> superTypes = new ArrayList<>();
	private Collection<MergeGenericType> genericSuperTypes = new ArrayList<>();
	private Collection<MergeStructuralFeature> structuralFeatures = new ArrayList<>();
	
	
	public Map<String,MergeClass> getPseudoSubclasses() {
		Map<String,MergeClass> ret = new HashMap<String,MergeClass>();
		for (MergeManager man: this.getManager().getSubManagers()) {
			Set<MergeClass> findSuper = new HashSet<MergeClass>();
			for (EClass cl: baseClasses) {
				MergeClass trg = man.getForeignEClass(cl);
				if (trg != null) {
					findSuper.add(trg);
				}
			}
			MergeClass trg = MergeStructuralFeature.getCommonSuperType(findSuper);
			if (trg != null) {
				ret.put(man.getBasePackage(),trg);
			}
		}
		return ret;
	}
	

	public Map<String,MergeClass> getPseudoSuperclasses() {
		Map<String,MergeClass> ret = new HashMap<String,MergeClass>();
		for (MergeManager man: this.getManager().getSuperManagers()) {
			Set<MergeClass> findSuper = new HashSet<MergeClass>();
			for (EClass cl: baseClasses) {
				MergeClass trg = man.getForeignEClass(cl);
				if (trg != null) {
					findSuper.add(trg);
				}
			}
			MergeClass trg = MergeStructuralFeature.getCommonSuperType(findSuper);
			if (trg != null) {
				ret.put(man.getBasePackage(),trg);
			}
		}
		return ret;
	}
	
	/*private Map<String,MergeClass> pseudoSubClass = new HashMap<>();
	
	public void addPseudoSubClass(String str, MergeClass mergeClass) {
		pseudoSubClass.put(str, mergeClass);
	}
	
	public Map<String,MergeClass> getPseudoSubClass() {
		return pseudoSubClass;
	}*/
	
	public Collection<MergeStructuralFeature> getStructuralFeatures() {
		Collection<MergeStructuralFeature> ret = new HashSet<>();
		if (getName().equals("Study")) {
			System.out.print("");
		}
		for (EClass cl: baseClasses) {
			for (EStructuralFeature feat: cl.getEStructuralFeatures()) {
				if (feat instanceof EAttribute) {
					ret.add(mergeManager.getMatching((EAttribute)feat));
				} else if (feat instanceof EReference) {
					ret.add(mergeManager.getMatching((EReference)feat));
				} else {
					System.err.println("Unknown feature type " + feat.getClass());
				}
			}
		}
		if (getName().equals("Study")) {
			System.out.println("Study-Attrs:");
			Set<String> multi = new HashSet<String>();
			for (MergeStructuralFeature retF: ret) {
				System.out.print(retF.getName()+" ");
				if (!multi.add(retF.getName())) {
					System.out.print("");
					ret = new HashSet<>();
					for (EClass cl: baseClasses) {
						for (EStructuralFeature feat: cl.getEStructuralFeatures()) {
							if (feat instanceof EAttribute) {
								ret.add(mergeManager.getMatching((EAttribute)feat));
							} else if (feat instanceof EReference) {
								ret.add(mergeManager.getMatching((EReference)feat));
							} else {
								System.err.println("Unknown feature type " + feat.getClass());
							}
						}
					}
				}
			}
			System.out.println();
		}
		return ret;
	}
	
	public Collection<MergeClass> getSuperTypes() {
		Collection<MergeClass> ret = new HashSet<>();
		for (EClass cl: baseClasses) {
			for (EClass sup: cl.getESuperTypes()) {
				MergeClass match = mergeManager.getMatching(sup);
				ret.add(match);
			}
		}
		return ret;
	}
	
	public static Map<EClass,Collection<EClass>> subTypes = new HashMap<EClass, Collection<EClass>>();
	
	public void addSuperType(EClass parent, EClass child) {
		Collection<EClass> cls = subTypes.get(parent);
		if (cls == null) {
			subTypes.put(parent, cls = new ArrayList<EClass>());
		}
		cls.add(child);
	}
	
	public Collection<EClass> getSubTypes(EClass parent) {
		return subTypes.getOrDefault(parent, Collections.emptyList());
	}
	

	public Collection<MergeClass> getSubTypes() {
		Collection<MergeClass> ret = new HashSet<>();
		for (EClass cl: baseClasses) {
			for (EClass sup: getSubTypes(cl)) {
				MergeClass match = mergeManager.getMatching(sup);
				ret.add(match);
			}
		}
		return ret;
	}
	
	public Collection<MergeGenericType> getGenericSuperTypes() {
		return genericSuperTypes;
	}

	@Override
	public String getName() {
		//TODO: Besseres Merging
		String ret = null;
		for (EClass cl: baseClasses) {
			if (cl.getName() != null) {
				if (ret == null || ret.compareTo(cl.getName()) > 1) {
					ret = cl.getName();
				}				
			}
		}
		return ret;
	}


	

}
