package at.cosimo.verocl.metamodel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;

public abstract class MergeClassifier extends MergeObject {

	protected MergeClassifier(MergeManager manager) {
		super(manager);
	}
	
	public String getInstanceClassName() {
		return instanceClassName;
	}
	public void setInstanceClassName(String instanceClassName) {
		this.instanceClassName = instanceClassName;
	}

	public String getInstanceTypeName() {
		return instanceTypeName;
	}

	public void setInstanceTypeName(String instanceTypeName) {
		this.instanceTypeName = instanceTypeName;
	}

	public Object getDefaultValue() {
		return defaultValue;
	}

	public void setDefaultValue(Object defaultValue) {
		this.defaultValue = defaultValue;
	}

	public Class<?> getInstanceClass() {
		return instanceClass;
	}

	public void setInstanceClass(Class<?> instanceClass) {
		this.instanceClass = instanceClass;
	}

	private String instanceClassName;
	private Class<?> instanceClass;
	private Object defaultValue;
	private String instanceTypeName;
	
	protected static final Map<String,String> simpleMap = new HashMap<>();
	static {
		simpleMap.put("int", "Integer");
		simpleMap.put("boolean", "Boolean");
		simpleMap.put("double", "Double");
		simpleMap.put("char", "Character");
		simpleMap.put("long", "Long");
		simpleMap.put("byte", "Byte");
	}
	
	public static Class<?> getCommonSuperClass(Set<Class<?>> baseClasses) {
		Set<Class<?>> commonClasses = null;
		Set<Class<?>> referenceTypes = new HashSet<>();
		referenceTypes.addAll(baseClasses);
		for (Class<?> curClass: referenceTypes) {
			Set<Class<?>> closedClasses = new HashSet<>();
			Set<Class<?>> openClasses = new HashSet<>();
			openClasses.add(curClass);
			closedClasses.add(curClass);
			while (!openClasses.isEmpty()) {
				Iterator<Class<?>> iter = openClasses.iterator();
				Class<?> mc = iter.next();
				iter.remove();
				Collection<Class<?>> sups = new ArrayList<Class<?>>();
				for (int i = 0; i < mc.getInterfaces().length; ++i) {
					sups.add(mc.getInterfaces()[i]);
				}
				sups.add(mc.getSuperclass());
				if (sups != null) {
					for (Class<?> sup: sups) {
						if (closedClasses.add(sup)) {
							openClasses.add(sup);
						}
					}
				}
			}
			if (commonClasses == null) {
				commonClasses = closedClasses;
			} else {
				commonClasses.retainAll(closedClasses);
			}
		}
		if (commonClasses == null) {
			return null;
		}
		Set<Class<?>> toDelete = new HashSet<>();
		for (Class<?> mc: commonClasses) {
			Collection<Class<?>> sups = new ArrayList<Class<?>>();
			for (int i = 0; i < mc.getInterfaces().length; ++i) {
				sups.add(mc.getInterfaces()[i]);
			}
			sups.add(mc.getSuperclass());
			if (sups != null) {
				toDelete.addAll(sups);
			}
		}
		commonClasses.removeAll(toDelete);
		if (commonClasses.isEmpty()) {
			return null;
		}
		return commonClasses.iterator().next();
	}
	
	public abstract Collection<? extends EClassifier> getBase();
	
	public Class<?> getUpperType() { //Get-Typ
		Set<Class<?>> superClass = new HashSet<>();
		for (EClassifier baseType: getBase()) {
			if (("http://www.eclipse.org/emf/2002/Ecore".equals(baseType.eResource().getURI().toString()))) {
				String typ = simpleMap.get(baseType.getName());
				System.out.println(baseType.getName() + " => " + typ);
				try {
					superClass.add(Class.forName("java.lang."+typ));
				} catch (ClassNotFoundException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			} else {
				if (baseType.getInstanceClass() != null) {
					superClass.add(baseType.getInstanceClass());
				} else {
					if (true) {
						System.out.println("Hallo!");
					}
					System.err.println("Could nott find instance class for type " + baseType.getName() + " - mb TODO: Generate it?");
				}
			}
		}
		Class<?> sup = getCommonSuperClass(superClass);
		return sup;
	}
	
	public Set<Class<?>> getSuper(Class<?> sup) {
		Set<Class<?>> ret = new HashSet<Class<?>>();
		Set<Class<?>> cur = new HashSet<Class<?>>();
		cur.add(sup);
		ret.add(sup);
		while (!cur.isEmpty()) {
			Set<Class<?>> newCur = new HashSet<>();
			for (Class<?> c: cur) {
				if (ret.add(c.getSuperclass())) {
					newCur.add(c.getSuperclass());
				}
				for (Class<?> s: c.getInterfaces()) {
					if (ret.add(s)) {
						newCur.add(s);
					}
				}
			}
			cur = newCur;
		}
		return ret;
	}
	
	public Class<?> getLowerType() { //Set-Typ - geht nur, wenn eine direkte Unter/Obertypbeziehung zwischen allen Typen besteht
		Set<Class<?>> superClass = new HashSet<>();
		for (EClassifier baseType: getBase()) {
			if (("http://www.eclipse.org/emf/2002/Ecore".equals(baseType.eResource().getURI().toString()))) {
				String typ = simpleMap.get(baseType.getName());
				System.out.println(baseType.getName() + " => " + typ);
				try {
					superClass.add(Class.forName("java.lang."+typ));
				} catch (ClassNotFoundException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			} else {
				if (baseType.getInstanceClass() != null) {
					superClass.add(baseType.getInstanceClass());
				} else {
					System.err.println("Could not3 find instance class for type " + baseType.getName() + " - mb TODO: Generate it?");
				}
			}
		}
		Set<Class<?>> current = new HashSet<Class<?>>();
		Class<?> ret = null;
		for (Class<?> cl: superClass) {
			int remSize = current.size();
			boolean haveadded = false;
			Set<Class<?>> added = getSuper(cl);
			for (Class<?> c: added) {
				if (!current.add(c)) {
					--remSize;
				} else {
					haveadded = true;
				}
			}
			if (haveadded && remSize > 0) {
				return null; //No common subclass
			}
			if (ret == null || haveadded) {
				ret = cl;
			}
		}
		return ret; 
	}

}
