package at.cosimo.verocl.metamodel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.impl.EEnumImpl;

public class MergeDataType extends MergeClassifier {
	private Set<EDataType> base;
	
	public Set<EDataType> getBase() {
		return base;
	}

	protected MergeDataType(MergeManager manager, Set<EDataType> base) {
		super(manager);
		this.base = base;
	}

	@Override
	public String getName() {
		String ret = null;
		for (ENamedElement cl: base) {
			if (ret == null || ret.compareTo(cl.getName()) > 0) {
				ret = cl.getName();
			}
		}
		return ret;
	}

	public boolean isBasicDataType() {
		for (EDataType dt: base) {
			if (!("http://www.eclipse.org/emf/2002/Ecore".equals(dt.eResource().getURI().toString()))) {
				return false;
			}
		}
		return true;
	}
	
	
	
	private static final Map<String,String> simpleMap = new HashMap<>();
	static {
		simpleMap.put("int", "Integer");
		simpleMap.put("boolean", "Boolean");
		simpleMap.put("double", "Double");
		simpleMap.put("char", "Character");
		simpleMap.put("long", "Long");
		simpleMap.put("byte", "Byte");
	}
	
	public static final String unrollType(String type) {
		String ret = simpleMap.get(type);
		if (ret == null) {
			ret = type;
		}
		return ret;
	}
	
	public static Class<?> getCommonSuperClass(Set<Class<?>> baseClasses) {
		Set<Class<?>> commonClasses = null;
		Set<Class<?>> referenceTypes = new HashSet<>();
		referenceTypes.addAll(baseClasses);
		for (Class<?> curClass: referenceTypes) {
			Set<Class<?>> closedClasses = new HashSet<>();
			Set<Class<?>> openClasses = new HashSet<>();
			openClasses.add(curClass);
			closedClasses.add(curClass);
			while (!openClasses.isEmpty()) {
				Iterator<Class<?>> iter = openClasses.iterator();
				Class<?> mc = iter.next();
				iter.remove();
				Collection<Class<?>> sups = new ArrayList<Class<?>>();
				for (int i = 0; i < mc.getInterfaces().length; ++i) {
					sups.add(mc.getInterfaces()[i]);
				}
				if (mc.getSuperclass() != null) {
					sups.add(mc.getSuperclass());
				}
				if (sups != null) {
					for (Class<?> sup: sups) {
						if (closedClasses.add(sup)) {
							openClasses.add(sup);
						}
					}
				}
			}
			if (commonClasses == null) {
				commonClasses = closedClasses;
			} else {
				commonClasses.retainAll(closedClasses);
			}
		}
		if (commonClasses == null) {
			return null;
		}
		Set<Class<?>> toDelete = new HashSet<>();
		for (Class<?> mc: commonClasses) {
			Collection<Class<?>> sups = new ArrayList<Class<?>>();
			for (int i = 0; i < mc.getInterfaces().length; ++i) {
				sups.add(mc.getInterfaces()[i]);
			}
			sups.add(mc.getSuperclass());
			if (sups != null) {
				toDelete.addAll(sups);
			}
		}
		commonClasses.removeAll(toDelete);
		if (commonClasses.isEmpty()) {
			return null;
		}
		return commonClasses.iterator().next();
	}
	
	public Class<?> getUpperType() { //Get-Typ
		Set<Class<?>> superClass = new HashSet<>();
		for (EDataType baseType: base) {
			if (("http://www.eclipse.org/emf/2002/Ecore".equals(baseType.eResource().getURI().toString()))) {
				superClass.add(baseType.getInstanceClass());
			} else {
				if (baseType.getInstanceClass() != null) {
					superClass.add(baseType.getInstanceClass());
				} else {
					System.err.println("Could no4t find instance class for type " + baseType.getName() + " - mb TODO: Generate it?");
				}
			}
		}
		Class<?> sup = getCommonSuperClass(superClass);
		return sup;
	}
	
	public Set<Class<?>> getSuper(Class<?> sup) {
		Set<Class<?>> ret = new HashSet<Class<?>>();
		Set<Class<?>> cur = new HashSet<Class<?>>();
		cur.add(sup);
		ret.add(sup);
		while (!cur.isEmpty()) {
			Set<Class<?>> newCur = new HashSet<>();
			for (Class<?> c: cur) {
				if (ret.add(c.getSuperclass())) {
					newCur.add(c.getSuperclass());
				}
				for (Class<?> s: c.getInterfaces()) {
					if (ret.add(s)) {
						newCur.add(s);
					}
				}
			}
			cur = newCur;
		}
		return ret;
	}
	
	public Class<?> getLowerType() { //Set-Typ - geht nur, wenn eine direkte Unter/Obertypbeziehung zwischen allen Typen besteht
		Set<Class<?>> superClass = new HashSet<>();
		for (EDataType baseType: base) {
			if (("http://www.eclipse.org/emf/2002/Ecore".equals(baseType.eResource().getURI().toString()))) {
				superClass.add(baseType.getInstanceClass());
			} else {
				if (baseType.getInstanceClass() != null) {
					superClass.add(baseType.getInstanceClass());
				} else {
					System.err.println("Could no2t find instance class for type " + baseType.getName() + " - mb TODO: Generate it?");
				}
			}
		}
		Set<Class<?>> current = new HashSet<Class<?>>();
		Class<?> ret = null;
		for (Class<?> cl: superClass) {
			int remSize = current.size();
			boolean haveadded = false;
			Set<Class<?>> added = getSuper(cl);
			for (Class<?> c: added) {
				if (!current.add(c)) {
					--remSize;
				} else {
					haveadded = true;
				}
			}
			if (haveadded && remSize > 0) {
				return null; //No common subclass
			}
			if (ret == null || haveadded) {
				ret = cl;
			}
		}
		return ret; 
	}
	
	
	public String getInstanceName() {
		String ret = null;
		for (EDataType cl: base) {
			if (cl.getInstanceClass() != null) {
				if (ret == null || ret.compareTo(cl.getInstanceClassName()) > 0) {
					ret = cl.getInstanceClassName();
				}
			}
		}
		if (ret == null) {
			for (EDataType cl: base) {
				if (cl.getInstanceClass() != null && cl.getInstanceTypeName() != null) {
					if (ret == null || ret.compareTo(cl.getInstanceTypeName()) > 0) {
						ret = cl.getInstanceTypeName();
					}
				}
			}
		}
		if (ret == null) {
			for (EDataType cl: base) {
				if (cl.getInstanceClass() != null) {
					if (ret == null || ret.compareTo(cl.getInstanceClass().getCanonicalName()) > 0) {
						ret = cl.getInstanceClass().getCanonicalName();
					}
				}
			}
		}
		if (ret == null) {
			for (EDataType cl: base) {
				if (cl.getName() != null) {
					if (ret == null || ret.compareTo(cl.getName()) > 0) {
						ret = cl.getName();
					}
				}
			}
		}
		return unrollType(ret);
	}

}
