package at.cosimo.verocl.metamodel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.ENamedElement;

public class MergeEnum extends MergeClassifier {
	protected MergeEnum(MergeManager manager, Set<EEnum> base) {
		super(manager);
		this.base = base;
		System.out.println("MergeEnum contains Set with " + base.size() + " elements!");
		if (base.size() < 2) {
			new Exception().printStackTrace();
		}
	}

	private Set<EEnum> base = new HashSet<>();
	
	public static class MergeEnumLiteral {
		public MergeEnumLiteral(Set<EEnumLiteral> base, String name, String literal, int value) {
			this.baseLiterals = base;
			this.name = name;
			this.literal = literal;
			this.value = value;
		}
		
		public MergeEnumLiteral(Set<EEnumLiteral> eenumLiterals) {
			this.baseLiterals = eenumLiterals;
		}
		
		public void set(String name, String literal, int value) {
			this.name = name;
			this.literal = literal;
			this.value = value;
		}
		
		public String getName() {
			return name;
		}
		
		public int getValue() {
			return value;
		}
		
		public String getLiteral() {
			return literal;
		}
		
		public Set<EEnumLiteral> getBase() {
			return baseLiterals;
		}
		
		public void setName(String name) {
			this.name = name;
		}
		
		public void setLiteral(String literal)  {
			this.literal = literal;
		}
		
		public void setValue(int value) {
			this.value = value;
		}

		private Set<EEnumLiteral> baseLiterals = new HashSet<EEnumLiteral>();
		private String name;
		private String literal;
		private int value;
		private MergeEnum enumContainer;
		private MergeEnum getEnumContainer;
		private MergeEnum setEnumContainer;
		
		public void setEnumContainer(MergeEnum enumContainer) {
			this.enumContainer = enumContainer;
		}
		
		public void setSetEnumContainer(MergeEnum enumContainer) {
			this.setEnumContainer = enumContainer;
		}
		
		public void setGetEnumContainer(MergeEnum enumContainer) {
			this.getEnumContainer = enumContainer;
		}
		
		public MergeEnum getEnumContainer() {
			return enumContainer;
		}

		public MergeEnum getGetEnumContainer() {
			return getEnumContainer;
		}
		
		public MergeEnum getSetEnumContainer() {
			return setEnumContainer;
		}
	}

	public boolean isStrictPseudoSubtypeOf(MergeEnum other) {
		return base.size()>other.base.size() && base.containsAll(other.getBase()); 
	}
	
	Set<MergeEnum> pseudoSupertypes = new HashSet<MergeEnum>();
	Set<MergeEnum> pseudoSubtypes = new HashSet<MergeEnum>();
	
	public void addStrictPseudoSupertype(MergeEnum other) {
		if (pseudoSupertypes.add(other)) {
			other.addStrictPseudoSubtype(this);
		}
	}
	
	public void addStrictPseudoSubtype(MergeEnum other) {
		if (pseudoSubtypes.add(other)) {
			other.addStrictPseudoSupertype(this);
		}
	}
	
	@Override
	public Collection<EEnum> getBase() {
		return base;
	}

	@Override
	public String getName() {
		String ret = null;
		for (ENamedElement cl: base) {
			if (ret == null || ret.compareTo(cl.getName()) > 0) {
				ret = cl.getName();
			}
		}
		return ret;
	}
	
	
	List<MergeEnumLiteral> literals = new ArrayList<MergeEnum.MergeEnumLiteral>();
	public void addLiteral(MergeEnumLiteral literal) {
		literals.add(literal);
	}
	
	List<MergeEnumLiteral> realLiterals = null;
	List<MergeEnumLiteral> realUnionLiterals = null;
	
	public List<MergeEnumLiteral> getLiterals() {
		return realLiterals;
	}
	
	public List<MergeEnumLiteral> getGetLiterals() {
		return realLiterals;
	}
	
	public List<MergeEnumLiteral> getSetLiterals() {
		return realUnionLiterals;
	}
	
	private MergeEnum getEnum;
	private MergeEnum setEnum;
	
	public MergeEnum getGetEnum() {
		if (getEnum == null) {
			calculateRealLiterals();
			MergeEnum ret = getManager().getMergeEnum(new HashSet<>(realUnionLiterals));
			if (ret == null) {
				ret = new MergeEnum(getManager(),base);
				getEnum = ret;
				if (hasSetIdentity()) {
					setEnum = ret;
				}
				ret.base = base;
				ret.getEnum = ret;
				ret.realLiterals = realUnionLiterals;
				ret.setEnum = getSetEnum();
				ret.instanceName = instanceName;
				getManager().registerEnum(ret,hasSetIdentity(),true);
			} else {
				getEnum = ret;
			}
		}
		return getEnum;
	}
	
	public MergeEnum getSetEnum() {
		if (setEnum == null) {
			calculateRealLiterals();
			MergeEnum ret = getManager().getMergeEnum(new HashSet<>(realLiterals));
			if (ret == null) {
				ret = new MergeEnum(getManager(),base);
				ret.base = base;
				setEnum = ret;
				if (hasSetIdentity()) {
					getEnum = ret;
				}
				ret.setEnum = ret;
				ret.realLiterals = realLiterals;
				ret.getEnum = getGetEnum();
				ret.instanceName = instanceName;
				getManager().registerEnum(ret,hasSetIdentity(),false);
			} else {
				setEnum = ret;
			}	
		}
		return setEnum;
	}
	
	public boolean hasSetIdentity() {
		return realLiterals.size()==realUnionLiterals.size();
	}
	
	private void calculateRealLiterals(Set<MergeEnumLiteral> haveLiterals, List<MergeEnumLiteral> realLiterals) {
		if (realLiterals == null) {
			calculateRealLiterals();
		}
		for (MergeEnumLiteral literal: realLiterals) {
			if (haveLiterals.add(literal)) {
				realLiterals.add(literal);
			}
		}
	}
	
	public void calculateRealLiterals() {
		if (realLiterals == null) {
			realLiterals = new ArrayList<MergeEnum.MergeEnumLiteral>();
			Set<MergeEnumLiteral> haveLiterals = new HashSet<MergeEnum.MergeEnumLiteral>(literals);
			//Meine plus die aller Subs
			realLiterals.addAll(literals);
			for (MergeEnum pseudoSub: pseudoSubtypes) {
				pseudoSub.calculateRealLiterals(haveLiterals, realLiterals);
			}
		}
		if (realUnionLiterals == null) {
			realUnionLiterals = new ArrayList<MergeEnum.MergeEnumLiteral>();
			Set<MergeEnumLiteral> haveLiterals = new HashSet<MergeEnum.MergeEnumLiteral>(realLiterals);
			//Meine plus die aller Subs
			realUnionLiterals.addAll(realLiterals);
			Set<MergeEnum> mes = new HashSet<MergeEnum>(pseudoSupertypes);
			Set<MergeEnum> processed = new HashSet<MergeEnum>(pseudoSupertypes);			
			while (!mes.isEmpty()) {
				Iterator<MergeEnum> iter = mes.iterator();
				MergeEnum next = iter.next();
				for (MergeEnumLiteral mel: next.literals) {
					if (haveLiterals.add(mel)) {
						realUnionLiterals.add(mel);
					}
				}
				iter.remove();
				for (MergeEnum sup: next.pseudoSupertypes) {
					if (processed.add(sup)) {
						mes.add(sup);
					}
				}
			}
		}
	}
	
	public List<MergeEnumLiteral> getSpecificLiterals() {
		return literals;
	}
	
	private void getIndirectPseudoSubtypes(Set<MergeEnum> enums) {
		for (MergeEnum sub: pseudoSubtypes) {
			if (enums.add(sub)) {
				sub.getIndirectPseudoSubtypes(enums);
			}
		}
	}
	
	private void getIndirectPseudoSupertypes(Set<MergeEnum> enums) {
		for (MergeEnum sub: pseudoSupertypes) {
			if (enums.add(sub)) {
				sub.getIndirectPseudoSupertypes(enums);
			}
		}
	}

	public void cleanSubSuper() {
		Set<MergeEnum> subTypes = new HashSet<MergeEnum>();
		for (MergeEnum enu: pseudoSubtypes)  {
			enu.getIndirectPseudoSubtypes(subTypes);
		}
		pseudoSubtypes.removeAll(subTypes);
		Set<MergeEnum> superTypes = new HashSet<MergeEnum>();
		for (MergeEnum enu: pseudoSupertypes)  {
			enu.getIndirectPseudoSupertypes(superTypes);
		}
		pseudoSupertypes.removeAll(superTypes);		
	}

	public Set<EEnum> calcEnumUnionInter(Set<EEnum> union, int direction) {
		Set<EEnum> ret = new HashSet<EEnum>(this.getBase());
		union.addAll(this.getBase());
		if (direction >= 0) {
			for (MergeEnum sup: pseudoSupertypes) {
				ret.retainAll(calcEnumUnionInter(union,1));
			}
		}
		if (direction <= 0) {
			for (MergeEnum sub: pseudoSubtypes) {
				ret.retainAll(calcEnumUnionInter(union,-1));
			}
		}
		return ret;
	}

	private void getTopEnums(Set<MergeEnum> topEnums) {
		if (pseudoSupertypes.isEmpty()) {
			topEnums.add(this);
		} else {
			for (MergeEnum sup: pseudoSupertypes) {
				sup.getTopEnums(topEnums);
			}
		}
	}
	
	public Set<MergeEnum> getTopEnums() {
		Set<MergeEnum> ret = new HashSet<MergeEnum>();
		getTopEnums(ret);
		return ret;
	}
	
	private void getBotEnums(Set<MergeEnum> botEnums) {
		if (pseudoSubtypes.isEmpty()) {
			botEnums.add(this);
		} else {
			for (MergeEnum sub: pseudoSubtypes) {
				sub.getBotEnums(botEnums);
			}
		}
	}
	
	public Set<MergeEnum> getBotEnums() {
		Set<MergeEnum> ret = new HashSet<MergeEnum>();
		getBotEnums(ret);
		return ret;
	}
	
	public boolean isGetter() {
		return getGetEnum() == this;
	}
	
	public boolean isSetter() {
		return getSetEnum() == this;
	}
	
	public boolean isGetterAndSetter() {
		return getGetEnum() == getSetEnum();
	}
	
	public void setInstanceName(String instanceName) {
		this.instanceName = instanceName;
	}
	
	private String instanceName;

	public String getInstanceName() {
		//TODO: Stimmt sehr oft nicht
		return instanceName;
	}

	public Collection<MergeEnum> getPseudoSuper() {
		return pseudoSupertypes;
	}

	public Collection<MergeEnum> getPseudoSub() {
		return pseudoSubtypes;
	}

	public boolean isTop() {
		return pseudoSupertypes.isEmpty();
	}
	
}
 