package at.cosimo.verocl.metamodel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EGenericType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.impl.EClassImpl;
import org.eclipse.emf.ecore.resource.Resource;

import at.cosimo.verocl.metamodel.MergeEnum.MergeEnumLiteral;

public class MergeManager {
	
	public MergeManager() {
		System.out.println();
	}
	
	private String ecoreURI = "http://www.eclipse.org/emf/2002/Ecore";
	
	private Map<EClass,MergeClass> matchingClasses = new HashMap<>();
	private Map<String,EClass> partToEclass = new HashMap<>();
	private Map<EPackage,MergePackage> matchingPackages = new HashMap<>();
	private Map<EGenericType,MergeGenericType> matchingGenericTypes = new HashMap<>();
	private Map<EStructuralFeature, MergeStructuralFeature> matchingStructuralFeatures = new HashMap<>();
	private Map<EAttribute,MergeAttribute> matchingAttributes = new HashMap<>();
	private Map<EReference,MergeReference> matchingReference = new HashMap<>();
	private Map<EObject,Map<Resource,EObject>> setTable = new HashMap<>();
	private Map<EObject,Map<EObject,Resource>> setRTable = new HashMap<>();
	private Map<EDataType,MergeDataType> matchingDataTypes = new HashMap<>();
	private List<Resource> resources;
	
	
	private Set<MergeManager> subManagers = new HashSet<MergeManager>();
	private Set<MergeManager> superManagers = new HashSet<MergeManager>();
	
	
	private String basePackage;
	
	public void setBasePackage(String basePackage) {
		this.basePackage = basePackage;
	}
	
	public String getBasePackage() {
		return this.basePackage;
	}
	
	public void addSubManager(MergeManager manager) {
		subManagers.add(manager);
		manager.superManagers.add(this);
	}
	
	
	
	public Set<MergeManager> getSubManagers() {
		return subManagers;
	}
	
	public Set<MergeManager>  getSuperManagers() {
		return superManagers;
	}
	
	
	public static MergeManager getMergeManager(
			String packageName,
			List<Resource> resources,
			Map<EObject,Map<Resource,EObject>> setTable, Map<EObject,Map<EObject,Resource>> setRTable) {
		MergeManager ret = new MergeManager();
		ret.setMatchingInfo(resources, setTable, setRTable);
		ret.basePackage = packageName;
		
		Map<EObject,Map<EObject,Resource>> eobjectRSetCopy = new HashMap<>();
		for (Entry<EObject,Map<EObject,Resource>> entry: setRTable.entrySet()) {
			Map<EObject,Resource> copy = new HashMap<>();
			if (entry.getValue() == null) {continue;}
			for (Entry<EObject,Resource> entry2: entry.getValue().entrySet()) {
				copy.put(entry2.getKey(), entry2.getValue());
			}
			eobjectRSetCopy.put(entry.getKey(), copy);
		}
		List<MergeEnumLiteral> mergeLiterals = new ArrayList<MergeEnum.MergeEnumLiteral>();
		Map<Set<EObject>,MergeEnumLiteral> enumLiteralMap = new HashMap<Set<EObject>, MergeEnum.MergeEnumLiteral>();
		Map<EEnumLiteral,MergeEnumLiteral> oldNewLiteralMap = new HashMap<EEnumLiteral, MergeEnum.MergeEnumLiteral>();
		Map<MergeEnumLiteral,Set<EEnum>> baseEnum = new HashMap<MergeEnum.MergeEnumLiteral, Set<EEnum>>();
		Set<EEnum> existingEnums = new HashSet<EEnum>();
		//First: Generate MergeLiteral
		for (Entry<EObject,Map<EObject,Resource>> entry: eobjectRSetCopy.entrySet()) {
			if (entry.getKey() instanceof EEnum) {
				existingEnums.add((EEnum)entry.getKey());
			}
			Map<EObject,Resource> map = entry.getValue();
			Set<EEnumLiteral> eenumLiterals = new HashSet<EEnumLiteral>();
			for (EObject obj: map.keySet()) {
				if (obj instanceof EEnumLiteral) {
					eenumLiterals.add((EEnumLiteral)obj);
				} else if (obj != null) {
					eenumLiterals = null;
					break;
				}
			}
			if (eenumLiterals != null && !eenumLiterals.isEmpty()) {
				if (!enumLiteralMap.containsKey(map.keySet())) {
					MergeEnumLiteral enumLiteral = new MergeEnumLiteral(eenumLiterals);
					mergeLiterals.add(enumLiteral);
					enumLiteralMap.put(map.keySet(), enumLiteral);
					for (EEnumLiteral literal: eenumLiterals) {
						oldNewLiteralMap.put(literal,enumLiteral);
					}
					Set<EEnum> enumSet = new HashSet<EEnum>();
					for (EEnumLiteral literal: eenumLiterals) {
						enumSet.add(literal.getEEnum());
					}
					baseEnum.put(enumLiteral, enumSet);
				}
			}
		}
		//Hier sind lauter schnittmengen drin - Genau die Instanzen einer bestimmten Klasse. Ein Set<EEnum> --> MergeEnum
		Map<Set<EEnum>,MergeEnum> allEnums = new HashMap<>();
		Map<EEnum,Set<MergeEnum>> singleEnumRef = new HashMap<>();
		for (Entry<MergeEnumLiteral,Set<EEnum>> entry: baseEnum.entrySet()) {
			MergeEnum cur = allEnums.get(entry.getValue());
			if (cur == null) {
				allEnums.put(entry.getValue(), cur = new MergeEnum(ret, entry.getValue()));
				for (EEnum enu: entry.getValue()) {
					Set<MergeEnum> curEnums = singleEnumRef.get(enu);
					if (curEnums == null) { 
						singleEnumRef.put(enu, curEnums = new HashSet<MergeEnum>());
					}
					curEnums.add(cur);
				}				
			}
			cur.addLiteral(entry.getKey());
			entry.getKey().setEnumContainer(cur);
		}
		//Add singleton enums if they don't exist
		/*for (EEnum eenu: existingEnums) {
			
			Set<EEnum> singleton = new HashSet<EEnum>();
			singleton.add(eenu);
			if (allEnums.get(singleton) == null) {
				MergeEnum cur = new MergeEnum(ret,singleton);
				allEnums.put(singleton, cur);
				for (EEnum enu: singleton) {
					Set<MergeEnum> curEnums = singleEnumRef.get(enu);
					if (curEnums == null) { 
						singleEnumRef.put(enu, curEnums = new HashSet<MergeEnum>());
					}
					curEnums.add(cur);
				}
				for (EEnumLiteral lit: eenu.getELiterals()) {
					cur.addLiteral(oldNewLiteralMap.get(lit));
				}
			}
		}*/
		//Hier sind alle Enums, baue Sub/Supertyp-Relationship
		for (Entry<EEnum,Set<MergeEnum>> entry: singleEnumRef.entrySet()) {
			//Alle in dem Set sind mgliche sub/supertypen
			for (MergeEnum first: entry.getValue()) {
				for (MergeEnum second: entry.getValue()) {
					if (first.isStrictPseudoSubtypeOf(second)) {
						first.addStrictPseudoSupertype(second);
					}
				}
			}
		}
		//Entferne zustzliche Sub/Supertyp-Relationen		
		for (MergeEnum mergeEnum: allEnums.values()) {
			mergeEnum.cleanSubSuper();
		}
		//Es kann noch immer Diamantsachen geben und es wurde noch kein Vereinigungs-Get gebildet
		//Es gibt nur mglicherweise ein Durchschnitts-Set - wenn es keines gibt, ist es leer
		//Die mengen sind die Vereinigung aller Enums, die vorkommen
		//Und deren Schnittmenge (falls sie nicht existiert)
		//Alle tatschlichen Instanzen ergeben sich aus den eigenen Instanzen + aller Instanzen aller Unterenums
		
		for (MergeEnum mergeEnum: allEnums.values()) {
			mergeEnum.calculateRealLiterals();
		}
		List<MergeEnum> allEnumList = new ArrayList<MergeEnum>();
		//Bilde durchschnitts- und vereinigungsmenge aller Enums
		//Die ergeben dann ein Mergeenum, wenn es nicht schon vorkommt
		//Das leere Merge-Enum kann fter vorkommen und ist sinnlos - kein Setzen mglich
		for (MergeEnum mergeEnum: allEnums.values()) {
			allEnumList.add(mergeEnum);
			//Betrachte Menge aller Top-Enums und menge aller Bot-Enums
			//Wenn die gre 1 hat, passt alles, sonst muss was dazugefgt werden
//			Set<MergeEnum> topEnums = mergeEnum.getTopEnums();
//			if (topEnums.size() > 1) { //Das geht so nicht - das Union-Set passt nicht, denn das wre das Intersection-Set. Intersection-Set geht aber auch nicht ...
//				//Add additional top
//				Set<EEnum> unionEnum = new HashSet<EEnum>();
//				for (MergeEnum enu: topEnums) {
//					unionEnum.addAll(enu.getBase());
//				}
//				MergeEnum additional = new MergeEnum(ret, unionEnum);				
//				for (MergeEnum topEnum: topEnums) {
//					additional.addStrictPseudoSubtype(topEnum);
//				}
//				additional.calculateRealLiterals();
//				allEnumList.add(additional);
//				if (allEnums.put(unionEnum, additional) != null) {
//					assert false:"Adding union enum which already was there!";
//					System.err.println("Adding union enum which already was there!");
//				}
//			}
			
			Set<MergeEnum> botEnums = mergeEnum.getBotEnums(); //new HashSet<MergeEnum>();
//			for (MergeEnum top: topEnums) {
//				botEnums.addAll(top.getBotEnums());
//			}
			if (botEnums.size() > 1) { 
				Iterator<MergeEnum> iter = botEnums.iterator();
				//Add additional bot
				Set<EEnum> intersectEnum = new HashSet<EEnum>(iter.next().getBase());
				while (iter.hasNext()) {
					intersectEnum.addAll(iter.next().getBase());
				}
				MergeEnum additionalBot = new MergeEnum(ret,intersectEnum);
				for (MergeEnum botEnum: botEnums) {
					additionalBot.addStrictPseudoSupertype(botEnum);
				}
				allEnumList.add(additionalBot);
				additionalBot.calculateRealLiterals();
				if (allEnums.put(intersectEnum, additionalBot) != null) {
					assert false:"Adding intersection enum which already was there!";
					System.err.println("Adding intersection enum which already was there!");
				}
			}			
		}
		
		//Jetzt stehen in der EnumList alle Enum-Klassen drin, die generiert werden mssen
		for (MergeEnum enu: allEnumList) {
			if (enu.isTop()) {
				//Set value names etc.
				Set<String> haveStrings = new HashSet<String>();
				for (MergeEnumLiteral mel: enu.getGetLiterals()) {
					//Get the alphabetically first string
					String str = null;
					for (EEnumLiteral lit: mel.getBase()) {
						if (str == null || (lit.getName() != null && str.compareTo(lit.getName()) > 0)) {
							str = lit.getName();
						}
					}
					if (haveStrings.add(str)) {
						mel.setName(str);
						mel.setLiteral(str);
					} else {
						int nmb = 1;
						while (!haveStrings.add(str+nmb));
						mel.setName(str+nmb);
						mel.setLiteral(str+nmb);
					}
				}
			}
			ret.addMergeEnum(enu);
		}
		
		//Irgendwie muss man den Mergedatatype auf so ein MergeEnum mappen
		
		
		
		//Und dann kommt das mergen. Indem man maps cleart kann man sicherstellen, dass eine Map nur 1x verwendet wird
		for (Entry<EObject,Map<EObject,Resource>> entry: eobjectRSetCopy.entrySet()) {
			Map<EObject,Resource> map = entry.getValue();
			for (EObject obj: map.keySet()) {
				if (obj instanceof EClass) {
					ret.getMatching((EClass)obj);
				} /* else if (obj instanceof EDataType) {
					ret.getMatching((EDataType)obj);
				} else if (obj instanceof EEnum) {
					ret.getMatching((EEnum)obj);
				} else if (obj instanceof EEnumLiteral) {
					ret.getMatching((EEnumLiteral)obj);
				}*/
			}
			
			map.clear();
		}
		return ret;
	}
	
	private List<MergeEnum> mergeEnums = new ArrayList<MergeEnum>();
	private Map<Collection<EEnum>,MergeEnum> getMergeEnums = new HashMap<>();
	private Map<Collection<EEnum>,MergeEnum> setMergeEnums = new HashMap<>();
	
	private void addMergeEnum(MergeEnum enu) {
		mergeEnums.add(enu);
		getMergeEnums.put(enu.getBase(), enu.getGetEnum());
		setMergeEnums.put(enu.getBase(), enu.getSetEnum());
	}

	public void setMatchingInfo(List<Resource> resources, Map<EObject,Map<Resource,EObject>> setTable, Map<EObject,Map<EObject,Resource>> setRTable) {
		this.resources = resources;
		this.setTable = setTable;
		this.setRTable = setRTable;
	}
	
	public static String getFullURI(EObject obj) {
		return obj.eResource().getURI().toString()+"/"+obj.eResource().getURIFragment(obj);
	}
	
	public void sortEObjectResources(List<EObject> eobjs, Map<EObject,Resource> resources) {
		Collections.sort(eobjs,new Comparator<EObject>() {

			@Override
			public int compare(EObject arg0, EObject arg1) {
				return Integer.compare(MergeManager.this.resources.indexOf(arg0.eResource()),MergeManager.this.resources.indexOf(arg1.eResource()));
						//indexarg0.eResource().getURI().toString().compareTo(arg1.eResource().getURI().toString());
			}
		});
	}
	
	public MergeClass getForeignEClass(EClass cl) {
		EClass retCl = partToEclass.get(getFullURI(cl));
		MergeClass ret = matchingClasses.get(retCl);
		if (ret == null) {
			Map<EObject,Resource> map = setRTable.get(retCl);
			if (map == null) {
				return null;
			}
			List<EObject> eobjList = new ArrayList<>();
			eobjList.addAll(map.keySet());
			sortEObjectResources(eobjList, map);
			ret = new MergeClass(this, (List<EClass>)(List)eobjList);
			for (EClass subcl: ret.getBase()) {
				matchingClasses.put(subcl, ret);
				partToEclass.put(getFullURI(subcl), subcl);
			}
		}
		return ret;
	}
	
	//Wenn ich ein 
	
	public MergeClass getMatching(EClass cl) {
		MergeClass ret = matchingClasses.get(cl);
		if (ret == null) {
			Map<EObject,Resource> map = setRTable.get(cl);
			if (map == null) {
				throw new RuntimeException("Accessing class " + cl.getName() + " which was not put in before");
			}
			List<EObject> eobjList = new ArrayList<>();
			eobjList.addAll(map.keySet());
			sortEObjectResources(eobjList, map);
			ret = new MergeClass(this, (List<EClass>)(List)eobjList);
			for (EClass subcl: ret.getBase()) {
				matchingClasses.put(subcl, ret);
				partToEclass.put(getFullURI(subcl), subcl);
			}
		}
		return ret;
	}
	
	public MergeDataType getMatching(EDataType cl) {
		//TODO: Ist das wirklich gerechtfertigt oder nicht? EEnums knnte man ja theoretisch matchen
		//d.h. TODO: Mindestens Enums einbauen
		MergeDataType ret = matchingDataTypes.get(cl);
		if (ret == null) {
			Map<EObject,Resource> map = setRTable.get(cl);
			if (map == null) {
				System.err.println("Accessing attribute " + cl.getName() + " which was not put in before");
				matchingDataTypes.put(cl, ret = new MergeDataType(this, Collections.singleton(cl)));
				return ret;
			}
			ret = new MergeDataType(this, (Set<EDataType>)(Set)map.keySet());
			for (EDataType subDt: ret.getBase()) {
				matchingDataTypes.put(subDt, ret);
			}
			
		}
		return ret;
	}
	
	
	public MergeAttribute getMatching(EAttribute cl) {
		MergeAttribute ret = matchingAttributes.get(cl);
		if (ret == null) {
			Map<EObject,Resource> map = setRTable.get(cl);
			if (map == null) {
				throw new RuntimeException("Accessing attr " + cl.getName() + " which was not put in before");
			}
			ret = new MergeAttribute(this, (Set<EAttribute>)(Set)map.keySet());
			for (EAttribute  subcl: ret.getBaseAttributes()) {
				matchingAttributes.put(subcl, ret);
			}
		}
		return ret;
	}
	
	public MergeReference getMatching(EReference cl) {
		MergeReference ret = matchingReference.get(cl);
		if (ret == null) {
			Map<EObject,Resource> map = setRTable.get(cl);
			if (map == null) {
				throw new RuntimeException("Accessing reference " + cl.getName() + " which was not put in before");
			}
			ret = new MergeReference(this, (Set<EReference>)(Set)map.keySet());
			for (EReference subcl: ret.getBaseReferences()) {
				matchingReference.put(subcl, ret);
			}
		}
		return ret;
	}
	
	public MergePackage getMatching(EPackage cl) {
		MergePackage ret = matchingPackages.get(cl);
		if (ret == null) {
			Map<EObject,Resource> map = setRTable.get(cl);
			if (map == null) {
				throw new RuntimeException("Accessing package " + cl.getName() + " which was not put in before");
			}
			ret = new MergePackage(this, (Set<EPackage>)(Set)map.keySet());
			for (EPackage subcl: ret.getBasePackages()) {
				matchingPackages.put(subcl, ret);
			}
		}
		return ret;
	}
	
	public MergeGenericType getMatching(EGenericType cl) {
		MergeGenericType ret = matchingGenericTypes.get(cl);
		if (ret == null) {
			Map<EObject,Resource> map = setRTable.get(cl);
			if (map == null) {
				throw new RuntimeException("Accessing generic type  which was not put in before");
			}
			ret = new MergeGenericType(this, (Set<EGenericType>)(Set)map.keySet());
			for (EGenericType subcl: ret.getBaseGenericTypes()) {
				matchingGenericTypes.put(subcl, ret);
			}
		}
		return ret;
	}
	
	private long curId = 0;

	public long newId() {
		return ++curId;
	}

	public Set<MergeClass> getMergeClasses() {
		Set<MergeClass> ret = new HashSet<MergeClass>();
		ret.addAll(matchingClasses.values());
		return ret;
	}

	public List<Resource> getResources() {
		return this.resources;
	}

	public Collection<String> getResourcesString() {
		Collection<String> ret = new ArrayList<String>();
		for (Resource r: this.resources) {
			ret.add(r.getURI().toString());
		}
		return ret;
	}

	public MergeEnum getSetMergeEnum(Set<EEnum> eenumSet) {
		MergeEnum ret = setMergeEnums.get(eenumSet);
		if (ret == null) {
			//Try to find a more general classifier, i.e. one which is (a) a superset of this classifier and (b) is the least such, i.e. counts the most elements
			int curElementCount = -1;
			for (Entry<Collection<EEnum>,MergeEnum> entry: setMergeEnums.entrySet()) {
				if (entry.getKey().containsAll(eenumSet)) {
					if (entry.getValue().getLiterals().size() > curElementCount) {
						ret = entry.getValue();
						curElementCount = entry.getValue().getLiterals().size();
					}
				}
			}
			if (curElementCount < Integer.MAX_VALUE) {
				setMergeEnums.put(eenumSet, ret);
			}
		}
		return ret;
	}

	public MergeEnum getGetMergeEnum(Set<EEnum> eenumSet) {
		MergeEnum ret = getMergeEnums.get(eenumSet);
		if (ret == null) {
			//Try to find a more general classifier, i.e. one which is (a) a superset of this classifier and (b) is the least such, i.e. counts the least elements
			int curElementCount = Integer.MAX_VALUE;
			for (Entry<Collection<EEnum>,MergeEnum> entry: getMergeEnums.entrySet()) {
				if (entry.getKey().containsAll(eenumSet)) {
					if (entry.getValue().getLiterals().size() < curElementCount) {
						ret = entry.getValue();
						curElementCount = entry.getValue().getLiterals().size();
					}
				}
			}
			if (curElementCount < Integer.MAX_VALUE) {
				getMergeEnums.put(eenumSet, ret);
			}
			
		}
		return ret;  
	}

	private Map<Set<MergeEnumLiteral>, MergeEnum> createdEnums = new HashMap();
	
	public MergeEnum getMergeEnum(Set<MergeEnumLiteral> realLiterals) {
		return createdEnums.get(realLiterals);
	}
	
	private Map<String,MergeEnum> instanceNames = new HashMap<String, MergeEnum>();

	public void registerEnum(MergeEnum ret, boolean wasGetterAndSetter, boolean makeGetter) {
		//Check for sub/Supertype
		for (Entry<Set<MergeEnumLiteral>, MergeEnum> haveEnums: createdEnums.entrySet()) {
			Set<MergeEnumLiteral> mels = haveEnums.getKey();
			if (mels.size() > ret.realLiterals.size() && mels.containsAll(ret.realLiterals)) {
				ret.addStrictPseudoSupertype(haveEnums.getValue());
			} else if (mels.size() < ret.realLiterals.size() && ret.realLiterals.contains(mels)) {
				ret.addStrictPseudoSubtype(haveEnums.getValue());
			}
		}
		String str = ret.getName();
		if (!wasGetterAndSetter) {
			if (makeGetter) {
				str = str+"G";
			} else {
				str = str+"S";
			}
		}
		if (instanceNames.get(str) == null) {
			ret.setInstanceName(str);
			instanceNames.put(str, ret);
		} else {
			int curV = 1;
			while (instanceNames.get(str+curV) != null) {
				++curV;
			}
			ret.setInstanceName(str+curV);
			instanceNames.put(str+curV, ret);
		}
		createdEnums.put(new HashSet<>(ret.realLiterals), ret);
		if (ret.isTop()) {
			if (makeGetter) {
				for (MergeEnumLiteral mel: ret.getLiterals()) {
					mel.setGetEnumContainer(ret);
				}
			} else {
				for (MergeEnumLiteral mel: ret.getLiterals()) {
					mel.setSetEnumContainer(ret);
				}
			}
		}
	}

	public Collection<MergeEnum> getMergeEnums() {
		return instanceNames.values();
	}

	
}
