package at.cosimo.verocl.metamodel;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.resource.Resource;

public class MergeStructuralFeature extends MergeObject {
	
	private Set<? extends EStructuralFeature> baseStructuralFeatures = new HashSet<>();
	
	public Set<? extends EStructuralFeature> getBaseStructuralFeatures() {
		return baseStructuralFeatures;
	}
	
	protected MergeStructuralFeature(MergeManager manager, Set<? extends EStructuralFeature> baseStructuralFeatures) {
		super(manager);
		this.baseStructuralFeatures = baseStructuralFeatures;
	}	
	
	public boolean isOrdered() {
		return ordered;
	}
	
	public boolean isUnique() {
		return unique;
	}
	
	
	public int getSetLowerBound() {
		int maxBound = 0;
		for (EStructuralFeature feat: baseStructuralFeatures) {
			if (feat.getLowerBound() > maxBound) {
				maxBound = feat.getLowerBound();
			}
		}
		return maxBound;
	}
	
	public int getGetLowerBound() {
		int maxBound = Integer.MAX_VALUE;
		for (EStructuralFeature feat: baseStructuralFeatures) {
			if (feat.getLowerBound() < maxBound) {
				maxBound = feat.getLowerBound();
			}
		}
		return maxBound;
	}
	
	public int getGetUpperBound() {
		int maxBound = Integer.MIN_VALUE;
		for (EStructuralFeature feat: baseStructuralFeatures) {
			if (maxBound == Integer.MIN_VALUE || feat.getUpperBound() == -1) {
				maxBound = feat.getUpperBound();
			} else if (maxBound >= 0 && feat.getUpperBound() > maxBound) {
				maxBound = feat.getUpperBound();
			}
		}
		return maxBound;
	}
	
	public int getSetUpperBound() {
		int maxBound = -1;
		for (EStructuralFeature feat: baseStructuralFeatures) {
			if (maxBound == Integer.MIN_VALUE || feat.getUpperBound() == -1) {
				//maxBound = feat.getUpperBound();
			} else if (maxBound == -1 || feat.getUpperBound() < maxBound) {
				maxBound = feat.getUpperBound();
			}
		}
		return maxBound;
	}
	
	public boolean isChangeable() {
		return changeable;
	}
	public void setChangeable(boolean changeable) {
		this.changeable = changeable;
	}
	public boolean isVolatile() {
		return volatile_;
	}
	public boolean isTransient() {
		return transient_;
	}
	public String getDefaultValueLiteral() {
		return defaultValueLiteral;
	}
	
	public Object getDefaultValue() {
		return defaultValue;
	}
	
	public boolean isUnsettable() {
		return unsettable;
	}
	
	public boolean isDerived() {
		return derived;
	}
	
	private boolean ordered;
	private boolean unique;
	
	
	private boolean changeable;
	private boolean volatile_;
	private boolean transient_;
	private String defaultValueLiteral;
	private Object defaultValue;
	private boolean unsettable;
	private boolean derived;
	
	public static MergeClass getCommonSuperType(Collection<MergeClass> referenceTypes) {
		Set<MergeClass> commonClasses = null;
		for (MergeClass curClass: referenceTypes) {
			Set<MergeClass> closedClasses = new HashSet<>();
			Set<MergeClass> openClasses = new HashSet<>();
			openClasses.add(curClass);
			closedClasses.add(curClass);
			while (!openClasses.isEmpty()) {
				Iterator<MergeClass> iter = openClasses.iterator();
				MergeClass mc = iter.next();
				iter.remove();
				for (MergeClass sup: mc.getSuperTypes()) {
					if (closedClasses.add(sup)) {
						openClasses.add(sup);
					}
				}
			}
			if (commonClasses == null) {
				commonClasses = closedClasses;
			} else {
				commonClasses.retainAll(closedClasses);
			}
		}
		if (commonClasses == null) {
			return null;
		}
		Set<MergeClass> toDelete = new HashSet<>();
		for (MergeClass mc: commonClasses) {
			toDelete.addAll(mc.getSuperTypes());
		}
		commonClasses.removeAll(toDelete);
		if (commonClasses.isEmpty()) {
			return null;
		}
		return commonClasses.iterator().next();
	}
	
	public static MergeClass getCommonSubType(Collection<MergeClass> referenceTypes) {
		Set<MergeClass> commonClasses = null;
		for (MergeClass curClass: referenceTypes) {
			Set<MergeClass> closedClasses = new HashSet<>();
			Set<MergeClass> openClasses = new HashSet<>();
			openClasses.add(curClass);
			closedClasses.add(curClass);
			while (!openClasses.isEmpty()) {
				Iterator<MergeClass> iter = openClasses.iterator();
				MergeClass mc = iter.next();
				iter.remove();
				for (MergeClass sup: mc.getSubTypes()) {
					if (closedClasses.add(sup)) {
						openClasses.add(sup);
					}
				}
			}
			if (commonClasses == null) {
				commonClasses = closedClasses;
			} else {
				commonClasses.retainAll(closedClasses);
			}
		}
		if (commonClasses == null) {
			return null;
		}
		Set<MergeClass> toDelete = new HashSet<>();
		for (MergeClass mc: commonClasses) {
			toDelete.addAll(mc.getSubTypes());
		}
		commonClasses.removeAll(toDelete);
		if (commonClasses.isEmpty()) {
			return null;
		}
		return commonClasses.iterator().next();
	}
	
	public MergeClassifier getGetType() {
		Set<MergeDataType> dataTypes = new HashSet<>();
		Set<MergeClass> referenceTypes = new HashSet<>();
		Set<EEnum> eenumSet = new HashSet<EEnum>();
		for (EStructuralFeature feat: baseStructuralFeatures) {
			EClassifier type = feat.getEType();
			if (type instanceof EEnum) {
				if (eenumSet != null) {
					eenumSet.add((EEnum)type);
				}
			} else if (type instanceof EDataType) {
				dataTypes.add(mergeManager.getMatching((EDataType)type));
				eenumSet = null;
			} else if (type instanceof EClass) {
				EClass tClass = (EClass)type;
				referenceTypes.add(mergeManager.getMatching(tClass));
				eenumSet = null;
			} else {
				throw new RuntimeException("Unknown type type " + ((type==null)?"null":type.getClass().getCanonicalName())+ " for feature " + feat.getName()+
						" in file " + feat.eResource().getURI());
			}
		}
		if (eenumSet != null && !eenumSet.isEmpty()) {
			return mergeManager.getGetMergeEnum(eenumSet); 
		}
		if (!dataTypes.isEmpty() && !referenceTypes.isEmpty()) {
			return null /*java.lang.Object*/;
		}
		if (!dataTypes.isEmpty()) {
			if (dataTypes.size() > 1) {
				return null /*java.lang.Object*/;
			}
			return dataTypes.iterator().next(); 
		}
		if (!referenceTypes.isEmpty()) {
			MergeClass common =  getCommonSuperType(referenceTypes);
			if (common != null) {
				//TODO: ...
				return common;
			} else {
				//TODO: ...
				return null;
			}
		}
		return null;
	}
	
	public MergeClassifier getSetType() {
		Set<MergeDataType> dataTypes = new HashSet<>();
		Set<EEnum> eenumSet = new HashSet<EEnum>();
		Set<MergeClass> referenceTypes = new HashSet<>();
		for (EStructuralFeature feat: baseStructuralFeatures) {
			EClassifier type = feat.getEType();
			if (type instanceof EEnum) {
				if (eenumSet != null) {
					eenumSet.add((EEnum)type);
				}
			} else if (type instanceof EDataType) {
				dataTypes.add(mergeManager.getMatching((EDataType)type));
				eenumSet = null;
			} else if (type instanceof EClass) {
				EClass tClass = (EClass)type;
				referenceTypes.add(mergeManager.getMatching(tClass));
				eenumSet = null;
			} else {
				throw new RuntimeException("Unknown type type " + ((type==null)?"null":type.getClass().getCanonicalName()));
			}
		}
		if (eenumSet != null && !eenumSet.isEmpty()) {
			return mergeManager.getSetMergeEnum(eenumSet); 
		}
		if (!dataTypes.isEmpty() && !referenceTypes.isEmpty()) {
			return null /*java.lang.Object*/;
		}
		if (!dataTypes.isEmpty()) {
			if (dataTypes.size() > 1) {
				return null /*java.lang.Object*/;
			}
			return dataTypes.iterator().next(); 
		}
		if (!referenceTypes.isEmpty()) {
			MergeClass common =  getCommonSubType(referenceTypes);
			if (common != null) {
				//TODO: ...
				return common;
			} else {
				//TODO: ...
				return null;
			}
		}
		return null;
	}
	
	public boolean isVirtualSettable() {
		Set<String> haveResources = new HashSet<>();
		for (EStructuralFeature feat: baseStructuralFeatures) {
			haveResources.add(feat.eResource().getURI().toString());
		}
		return haveResources.containsAll(mergeManager.getResourcesString()) && canSet();
	}

	
	@Override
	public String getName() {
		String ret = null;
		//TODO: Besseres Merging
		for (ENamedElement cl: baseStructuralFeatures) {
			if (ret == null || ret.compareTo(cl.getName()) == 1) {
				ret = cl.getName();
			}
		}
		return ret;
	}

	public boolean canSet() {
		return (getSetUpperBound() == -1 || getSetUpperBound() >= getSetLowerBound()) && getSetType() != null;
	}

	public boolean isGetMulti() {
		return getGetUpperBound() == -1 || getGetUpperBound() > 1;
	}
}
