package at.cosimo.verocl.metamodel.output;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.resource.Resource;

import at.cosimo.verocl.metamodel.MergeAttribute;
import at.cosimo.verocl.metamodel.MergeClass;
import at.cosimo.verocl.metamodel.MergeClassifier;
import at.cosimo.verocl.metamodel.MergeDataType;
import at.cosimo.verocl.metamodel.MergeReference;
import at.cosimo.verocl.metamodel.MergeStructuralFeature;
import at.cosimo.verocl.metamodel.output.OutputClass.Field;

public class EClassManager {

	
	private Map<MergeClass,EClassInfo> eclassMap = new HashMap<>();
	private Map<MergeClassifier,EClassifier> eclassifierMap = new HashMap<>();
	private OutputManager om;
	
	public EClassManager(OutputManager om) {
		this.om = om;
	}
	
	
	public EStructuralFeature getStructuralFeature() {
		throw new UnsupportedOperationException();
	}
	
	public static class EClassInfo {
		public EClassInfo(EClass ecl) {
			this.ecl = ecl;
		}
		
		private EClass ecl;
		private Map<String,EStructuralFeature> esfMap = new HashMap<>();
		
		public EClass getEClass() {
			return ecl;
		}
		
		public Map<String,EStructuralFeature> getESFMap() {
			return esfMap;
		}
	}
	
	private Map<Class<?>,EDataType> eDataTypeMap = new HashMap<>();
	
	public EDataType generateEDataType(Class<?> cl, Resource res) {
		EDataType ret = eDataTypeMap.get(cl);
		if (ret == null) {
			//TODO: Konvertiere das ganze ...
			EDataType type = EcoreFactory.eINSTANCE.createEDataType();
			type.setInstanceClass(cl);
			type.setName(cl.getName());
			eDataTypeMap.put(cl, ret = type);
		}
		return ret;
	}
	
	public EClassifier generateEClassifier(MergeClassifier mcl, Resource res) {
		if (mcl instanceof MergeClass) {
			return generateEClass((MergeClass)mcl, res);
		} else if (mcl instanceof MergeDataType) {
			MergeDataType mdt = (MergeDataType)mcl;
			Class<?> lower = mdt.getLowerType();
			Class<?> upper = mdt.getUpperType();
			return generateEDataType(upper, res);
		} else {
			System.err.println("Unknown MergeClassifier " + (mcl==null?"null":mcl.getClass()));
			return null;
		}
	}
	
	//TODO: Klassen, die schon EClass sind darf man nicht mehr verndern ...
	
	public EClass generateEClass(MergeClass mcl, Resource res) {
		EClassInfo retInfo = eclassMap.get(mcl);
		OutputClass clazz = om.getOrCreate(mcl, res);
		if (retInfo == null) {
			EClass ret = EcoreFactory.eINSTANCE.createEClass();
			retInfo = new EClassInfo(ret);
			eclassMap.put(mcl, retInfo);
			ret.setName(mcl.getName());
			ret.setInterface(mcl.isInterface());
			ret.setAbstract(false);
			ret.setInstanceClassName(clazz.getQualifiedName());
			EList<EStructuralFeature> features = ret.getEStructuralFeatures();
			Map<String,EStructuralFeature> nameFeatureMap = new HashMap<String,EStructuralFeature>();
			for (MergeStructuralFeature msf: mcl.getStructuralFeatures()) {
				EStructuralFeature feat = null;
				if (msf instanceof MergeAttribute) {
					EAttribute eattr = EcoreFactory.eINSTANCE.createEAttribute();
					feat = eattr;
					MergeClassifier classifier = msf.getGetType();
					feat.setChangeable(msf.isVirtualSettable());
				} else if (msf instanceof MergeReference) {
					EReference eref = EcoreFactory.eINSTANCE.createEReference();
					feat = eref;
					feat.setChangeable(msf.isVirtualSettable());
				} else {
					System.err.println("Unknown msf type " + (msf==null?"null":msf.getClass()));
				}
				feat.setName(msf.getName());
				feat.setDerived(true);
				feat.setEType(generateEClassifier(msf.getGetType(), res));
				
				nameFeatureMap.put(msf.getName(), feat);
			}
		}
		return retInfo.getEClass();
	}
}
