package at.cosimo.verocl.metamodel.output;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EPackage;

import com.sun.javafx.animation.TickCalculation;

import at.cosimo.verocl.metamodel.MergeClass;
import at.cosimo.verocl.template.TemplateProcessor;
import at.cosimo.verocl.template.TemplateProcessorIFace;

public class OutputClass {
	
	
	private TemplateProcessorIFace fullProcessor = null;
	
	
	
	private List<OutputClass> superClasses = new ArrayList<OutputClass>();
	private String superClassesStr = null;
	private List<String> interfacesStr = new ArrayList<>();
	private OutputPackage pkg;
	private String name;
	private Set<String> imports = new HashSet<>();
	private List<String> initBlock = new ArrayList<String>();
	
	private List<TemplateProcessorIFace> topStringProcessor = new ArrayList<>();
	private List<TemplateProcessorIFace> bottomStringProcessor = new ArrayList<>();
	
	public void addTopProcessor(TemplateProcessorIFace tIface) {
		topStringProcessor.add(tIface);
	}
	
	public void addBottomProcessor(TemplateProcessorIFace bottomProcessor) {
		bottomStringProcessor.add(bottomProcessor);
	}
	
	private boolean useEObject = false;
	
	public void setUseEObject(boolean useEObject) {
		this.useEObject = useEObject;
	}
	
	public boolean useEObject() {
		return useEObject;
	}
	
	public void addInit(String stmt) {
		initBlock.add(stmt);
	}
	
	public CLASS_TYPE getClassType() {
		return classType;
	}
	
	public void addEClassInit(String str) {
		eclassInit.add(str);
	}
	
	private MergeClass mergeClass;
	
	public MergeClass getMergeClass() {
		return mergeClass;
	}
	
	private List<String> eclassInit = new ArrayList<>();
	
	
	public static enum CLASS_TYPE {CLASS, INTERFACE, ENUM};
	
	private CLASS_TYPE classType;

	public OutputClass(TemplateProcessorIFace fullProcessor, String name, OutputPackage pkg, MergeClass from) {
		this.fullProcessor = fullProcessor;
		this.name = name;
		this.pkg = pkg;
		this.mergeClass = from;
	}
	
	public OutputClass(String name, OutputPackage pkg, MergeClass from, CLASS_TYPE classType) {
		this.name = name;
		this.mergeClass = from;
		this.pkg = pkg;
		this.classType = classType;
	}
	
	public OutputClass(String name, OutputPackage pkg, MergeClass from, boolean isInterface) {
		this.name = name;
		this.mergeClass = from;
		this.pkg = pkg;
		this.classType = isInterface?CLASS_TYPE.INTERFACE:CLASS_TYPE.CLASS;
	}
	
	public OutputPackage getPackage() {
		return pkg;
	}
	
	/**Not to be called*/
	void setPackage(OutputPackage pkg) {
		if (this.pkg != pkg) {
			this.pkg = pkg;
		}
	}
	
	public String getName() {
		return name;
	}
	
	public void addImport(String str) {
		imports.add(str);
	}
	
	public static class NamedType {
		public NamedType(String type, String name) {
			this.type = type;
			this.name = name;
		}
		
		private String type;
		private String name;
		
		public void toString(StringBuilder builder) {
			builder.append(type);
			builder.append(" ");
			builder.append(name);
		}
	}
	
	public static class Field {
		public Field(NamedType type, String defaultValue) {
			this.type = type;
			this.defaultValue = defaultValue;
		}
		
		public final NamedType type;
		public final String defaultValue;
	}
	
	public static class Method {
		public Method(String returnType, String name, List<NamedType> parameter, String implementation) {
			this.returnType = returnType;
			this.name = name;
			this.parameter = parameter;
			this.implementation = implementation;
		}
		
		public String returnType;
		public List<NamedType> parameter = new ArrayList<>();
		public String name;
		public List<String> exceptions = new ArrayList<String>(); 
		public String implementation;
	}
	
	public static class Constructor {
		public Constructor(List<NamedType> parameter, String implementation) {
			this.parameter = parameter;
			this.implementation = implementation;
		}
		
		private List<NamedType> parameter = new ArrayList<>();
		private String implementation;
	}
	
	private List<Field> fields = new ArrayList<>();
	private List<Method> methods = new ArrayList<>();
	private List<Constructor> constructors = new ArrayList<>();
	
	public void addField(Field field) {
		this.fields.add(field);
	}
	
	public List<Field> getFields() {
		return this.fields;
	}
	
	public void addMethod(Method method) {
		this.methods.add(method);
	}
	
	public void addConstructor(Constructor con) {
		this.constructors.add(con);
	}
	
	public String getQualifiedName() {
		return pkg.getFullName()+"."+name;
	}
	
	public String foreignName(OutputClass cl) {
		if (cl.pkg == null || cl.pkg == pkg) {
			return cl.name;
		} else {
			return cl.pkg.getFullName() + "."+cl.name;
		}
	}
	
	public void expandParameters(StringBuilder builder, List<NamedType> parameters) {
		boolean first = true;
		for (NamedType parameter: parameters) {
			if (first) {first = false;} else {builder.append(", ");}
			parameter.toString(builder);
		}
	}
	
	public void addSuperClass(OutputClass sup) {
		this.superClasses.add(sup);
	}
	
	private StringBuilder topWrite = new StringBuilder();
	private StringBuilder bottomWrite = new StringBuilder();
	
	public void addStringTop(String str) {
		topWrite.append(str);
	}
	
	public void addStringBottom(String str) {
		bottomWrite.append(str);
	}
	
	public void putOut(StringBuilder builder, int indent) {
		if (fullProcessor != null) {
			try {
				fullProcessor.process(builder,new int[]{0},this);
				return;
			} catch (ParseException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		builder.append("package ");
		builder.append(pkg.getFullName());
		builder.append(";");
		Util.newLine(builder, indent);
		if (!imports.isEmpty()) {
			Util.newLine(builder, indent);
			List<String> sortedList = new ArrayList<>();
			sortedList.addAll(imports);
			Collections.sort(sortedList);
			for (String str: sortedList) {
				builder.append("import ");
				builder.append(str);
				builder.append(";");
				Util.newLine(builder, indent);
			}
		}
		
		//TODO: Manchmal wren imports ganz nett ...
		builder.append("\npublic ");
		switch (classType) {
			case CLASS: builder.append("class"); break;
			case INTERFACE: builder.append("interface"); break;
			case ENUM: builder.append("enum"); break;
		}
		
		builder.append(" ");
		builder.append(name);
		List<OutputClass> superInterfaces = new ArrayList<>();
		OutputClass superClass = null;
		
		for (OutputClass cl: superClasses) {
			if (cl.classType == CLASS_TYPE.INTERFACE) {
				superInterfaces.add(cl);
			} else {
				if (cl.classType == CLASS_TYPE.INTERFACE || cl.classType == CLASS_TYPE.ENUM) {
					throw new RuntimeException("Interface " + name + " wants to extend class " + cl.getName());
				}
				if (superClass != null) {
					throw new RuntimeException("Class " + name + " has multiple superClasses, e.g. " + superClass.getName() + " and " + cl.getName());
				}
				superClass = cl;
			}
		}
		
		if (classType == CLASS_TYPE.INTERFACE) {
			if (!superInterfaces.isEmpty() || !interfacesStr.isEmpty()) {
				builder.append(" extends ");
				boolean first = true;
				for (OutputClass cl: superInterfaces) {
					if (first) {first = false;} else {builder.append(", ");}
					builder.append(foreignName(cl));
				}
				for (String str: interfacesStr) {
					if (first) {first = false;} else {builder.append(", ");}
					builder.append(str);
				}
			}
		} else {
			if (superClass != null || superClassesStr != null) {
				builder.append(" extends ");
				if (superClassesStr != null) {
					builder.append(superClassesStr);
				} else {
					builder.append(foreignName(superClass));
				}
			}
			if (!superInterfaces.isEmpty() || !interfacesStr.isEmpty()) {
				builder.append(" implements ");
				boolean first = true;
				for (OutputClass cl: superInterfaces) {
					if (first) {first = false;} else {builder.append(", ");}
					builder.append(foreignName(cl));
				}
				for (String str: interfacesStr) {
					if (first) {first = false;} else {builder.append(", ");}
					builder.append(str);
				}
			}
		}
		
		
		builder.append(" {");
		++indent;
		Util.newLine(builder, indent);		
		
		builder.append(topWrite);
		
		for (TemplateProcessorIFace tproc: topStringProcessor) {
			try {
				tproc.process(builder,new int[]{0},this);
			} catch (ParseException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		for (Constructor constructor: constructors) {
			builder.append("public ");
			builder.append(name);
			builder.append("(");
			expandParameters(builder, constructor.parameter);
			builder.append(") {");
			++indent;
			Util.newLine(builder, indent);
			builder.append(Util.expandNewLine(constructor.implementation, indent));			
			--indent;
			Util.newLine(builder, indent);
			builder.append("}");
			Util.newLine(builder, indent);
			Util.newLine(builder, indent);
		}
		
		
		for (Field field: fields) {
			builder.append("public ");
			field.type.toString(builder);
			if (field.defaultValue != null) {
				builder.append(" = ");
				builder.append(field.defaultValue);
			}
			builder.append(";");
			Util.newLine(builder, indent);
		}
		
		if (!eclassInit.isEmpty()) {
			builder.append("static {");
			for (String str: eclassInit) {
				builder.append("\n\t\t");
				builder.append(str);
			}
			builder.append("\n}\n");
		}
		
		/*if (useEObject) {
			
			builder.append("public static final EClass _eclass");
			builder.append("public static final EClass _staticeclass");
			builder.append("public static final EPackage _epkg");
			
			
			
			public static final EClass eclass;
			public static final EClass staticeclass;
			
			public static final EAttribute attribute;
			
			public static final EPackage epkg;
		}*/
		
		if (!initBlock.isEmpty()) {
			Util.newLine(builder, indent);
			builder.append("{");
			++indent;
			for (String str: initBlock) {
				Util.newLine(builder, indent);
				builder.append(Util.expandNewLine(str,indent));
			}
			--indent;
			Util.newLine(builder, indent);
			builder.append("}");
			Util.newLine(builder, indent);
		}
		
		
		Util.newLine(builder, indent);
		for (Method method: methods) {
			builder.append("public ");
			if (classType == CLASS_TYPE.CLASS && method.implementation == null) {
				builder.append("abstract ");
			}	
			if (classType == CLASS_TYPE.INTERFACE && method.implementation != null) {
				builder.append("default ");
			}
			builder.append(method.returnType);
			builder.append(" ");
			builder.append(method.name);
			builder.append("(");
			expandParameters(builder, method.parameter);
			builder.append(")");
			if (!method.exceptions.isEmpty()) {
				builder.append(" throws ");
				boolean first = true;
				for (String exception: method.exceptions) {
					if (first) {first = false;} else {builder.append(", ");}
					builder.append(exception);
				}
			}
			if (method.implementation == null) {
				builder.append(";");
			} else {
				builder.append(" {");
				++indent;
				Util.newLine(builder, indent);
				builder.append(Util.expandNewLine(method.implementation, indent));			
				--indent;
				Util.newLine(builder, indent);
				builder.append("}");
			}
			Util.newLine(builder, indent);
			Util.newLine(builder, indent);
		}
		for (TemplateProcessorIFace tproc: bottomStringProcessor) {
			try {
				tproc.process(builder,new int[]{0},this);
			} catch (ParseException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		builder.append(bottomWrite);
		--indent;
		Util.newLine(builder, indent);
		builder.append("}");
	}

	public List<OutputClass> getSuperClasses() {
		return superClasses;
	}
	
	public void addInterface(String string) {
		this.interfacesStr.add(string);
	}

	public void setSuperClass(String string) {
		this.superClassesStr = string;
	}
	

}
