package at.cosimo.verocl.metamodel.output;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.security.spec.PSSParameterSpec;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.UnknownFormatConversionException;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.emf.ecore.EcorePackage.Literals;
import org.eclipse.emf.ecore.impl.EClassImpl;
import org.eclipse.emf.ecore.impl.MinimalEObjectImpl;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.EcoreResourceFactoryImpl;
import org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl;

import at.cosimo.verocl.metamodel.ManagerManager;
import at.cosimo.verocl.metamodel.MergeAttribute;
import at.cosimo.verocl.metamodel.MergeClass;
import at.cosimo.verocl.metamodel.MergeClassifier;
import at.cosimo.verocl.metamodel.MergeDataType;
import at.cosimo.verocl.metamodel.MergeEnum;
import at.cosimo.verocl.metamodel.MergeGenericType;
import at.cosimo.verocl.metamodel.MergeManager;
import at.cosimo.verocl.metamodel.MergeReference;
import at.cosimo.verocl.metamodel.MergeStructuralFeature;
import at.cosimo.verocl.metamodel.MergeEnum.MergeEnumLiteral;
import at.cosimo.verocl.metamodel.output.OutputClass.CLASS_TYPE;
import at.cosimo.verocl.metamodel.output.OutputClass.Constructor;
import at.cosimo.verocl.metamodel.output.OutputClass.Field;
import at.cosimo.verocl.metamodel.output.OutputClass.Method;
import at.cosimo.verocl.metamodel.output.OutputClass.NamedType;
import at.cosimo.verocl.template.TemplateProcessor;

public class OutputManager {

	private Map<String,OutputPackage> output = new HashMap<>();
	private Map<String,String> basicNames = new HashMap<>();
	private Map<MergeClass,Map<Resource,OutputClass>> retMap = new HashMap<>();
	private Map<MergeEnum,Map<Resource,OutputClass>> retEnumMap = new HashMap<>();
	private boolean useEClass;
	
	public void setUseEClass(boolean useEClass) {
		this.useEClass = useEClass;
	}
	
	public boolean getUseEClass() {
		return useEClass;
	}
	
	public String mergeTypeToTypeDef(MergeGenericType type) {
		//TODO: ....
		return null;
	}
	
	public Map<String,OutputPackage> getOutput() {
		return output;
	}

	public Map<String,Collection<String>> subtypeMap() {
		Map<String,Collection<String>> ret = new HashMap<>();
		for (Entry<MergeClass,Map<Resource,OutputClass>> entry: retMap.entrySet()) {
			OutputClass oc = entry.getValue().get(null);
			String fullName = oc.getQualifiedName();
			if (!oc.getSuperClasses().isEmpty()) {
				List<String> supList = new ArrayList<>();
				for (OutputClass sup: oc.getSuperClasses()) {
					supList.add(sup.getQualifiedName());
				}
				ret.put(fullName,supList);
			}
		}
		return ret;
	}
	
	public Map<String,String> classToEclassMap() {
		Map<String,String> ret = new HashMap<>();
		for (Entry<MergeClass,Map<Resource,OutputClass>> entry: retMap.entrySet()) {
			for (EClass cl: entry.getKey().getBase()) {
				for (Entry<Resource,OutputClass> entr2: entry.getValue().entrySet()) {
					if (entr2.getKey() == null) {
						ret.put(entr2.getValue().getQualifiedName(),null);
					} else if (entr2.getKey().getURI().toString().equals(cl.eResource().getURI().toString())) {
						ret.put(entr2.getValue().getQualifiedName(),PseudoOutputManager.getFullName(cl));
					}
				}
			}
		}
		return ret;
	}
	
	public Collection<OutputPackage> getOutputPkg() {
		return output.values();
	}
	
	public OutputPackage getOutputPkg(Resource res) {
		return output.get((res==null)?null:res.getURI().toString());
	}
	
	public String getInstanceName(EClass cl) {
		if (cl == null && "EInt".equals(cl.getName())) {
			System.out.println("Found");
		}
		if (cl.getInstanceClass() != null) {
			return cl.getInstanceClass().getCanonicalName();
		}
		if (cl.getInstanceClassName() != null) {
			return cl.getInstanceClassName();
		}
		//TODO: Mit irgendeinem Paket davor
		if (cl.getName() != null) {
			return cl.getName();
		}
		return "EObject";
	}
	
	public static String digitToString(int index) {
		char letter = (char) ((index%26)+'A');
		if (index/26 > 0) {
			int remaining = (index/26)+1;
			return letter+String.valueOf(remaining);
		} else {
			return ""+letter;
		}
	}
	
	public String getJavaTypeName(String baseName, Resource resource) {
		OutputPackage pkg = getOutputPkg(resource);
		if (pkg == null) {
			System.err.println("No output pkg for package " + resource);
			for (String str: output.keySet()) {
				System.out.println(str);
			}
		}
		return pkg.getFullName()+"."+baseName;
	}
	
	public String getJavaTypeName(EClass ecl) {
		return getJavaTypeName(getInstanceName(ecl),ecl.eResource());
	}
	
	/**Statische Factory-Methoden wren sinnvoll - im Manager! - hmm, oder auch woanders.
	 * Eigentlich will man z.B. eine Referenz hinzufgen und der Typ soll "passen" zum aktuellen
	 * Typ
	 * addNewReference() //Vom direkten typ
	 * addNewReference(class) //Vom spezifizierten Typ, kann aber ein Interface sein - vielleicht wre da ein eigenes Enum praktischer
	 * 
	 * Man kann nur Dinge erstellen, die in allen BasisModellen vorhanden sind ...
	 * */
	
	public enum Settability {SETTABLE,NOT_SETTABLE};
	
	public String getJavaTypeCol(MergeStructuralFeature attr, Resource resource, OutputClass forClass, Boolean doCollection, boolean doGet, Settability settability) {
		MergeClassifier classifier = doGet?attr.getGetType():attr.getSetType();
		String ret = null;
		if (classifier == null) {
			classifier = (doGet?attr.getGetType():attr.getSetType());
			ret = "java.lang.Object";
		} else {
			//TODO: Der Name, den der Typ dann haben wird ...
			//Das hngt natrlich vom Ausgabepaket ab
			//TODO: Hier fehlt jetzt noch die komplette Paketverwaltung
			if (classifier instanceof MergeDataType) {
				if (((MergeDataType) classifier).isBasicDataType()) {
					ret = ((MergeDataType)classifier).getInstanceName();
				}
			} else if (classifier instanceof MergeEnum) {
				ret = forClass.foreignName(getOrCreate(((MergeEnum)classifier), null));
			}
			if (ret == null) {
				ret = getJavaTypeName(classifier.getName(),resource);
			}
		}
		int upperBound = doGet?attr.getGetUpperBound():attr.getSetUpperBound();
		if ((doCollection==null)?(upperBound > 1 || upperBound == -1):doCollection) {
			forClass.addImport("java.util.Collection");
			if (settability == Settability.NOT_SETTABLE) {
				return "Collection<? extends " + ret + ">";
			} else {
				return "Collection<" + ret + ">";
			}
		}
		return ret;
	}
	
	public String getJavaTypeNoCol(MergeStructuralFeature attr, Resource resource, OutputClass forClass, boolean doGet, Settability settability) {
		return getJavaTypeCol(attr,resource,forClass,null,doGet,settability);		
	}
	
	public String getBasicName(Resource res) {
		return basicNames.get((res==null)?null:res.getURI().toString());
	}
	
	public OutputClass getOrCreate(MergeClass mcl, Resource res) {
		Map<Resource,OutputClass> map = retMap.get(mcl);
		if (map == null) {
			retMap.put(mcl, map = new HashMap<>());
		}
		OutputClass ret = map.get(res);
		if (ret == null) {
			OutputPackage pkg = getOutputPkg(res);
			if (res == null || mcl.isInterface()) {
				ret = generateInterface(mcl, pkg, res, map);
			} else {
				ret = generateClass(mcl, pkg, res, map);
			}
			map.put(res, ret);
		}
		return ret;
	}
	
	public OutputClass getOrCreate(MergeEnum mcl, Resource res) {
		Map<Resource,OutputClass> map = retEnumMap.get(mcl);
		if (map == null) {
			retEnumMap.put(mcl, map = new HashMap<>());
		}
		OutputClass ret = map.get(res);
		if (ret == null) {
			OutputPackage pkg = getOutputPkg(res);
			ret = generateEnum(mcl, pkg, res, map);
			map.put(res, ret);
		}
		return ret;
	}
	
	//Bei der ganzen virtuellen Geschichte ist es wichtig, dass man jederzeit auf Subklassen zurckgreifen kann, d.h.
	//Sein virtuelles modell spezialisieren - das ganze hier
	//Methoden SpecializedMergeModel specialize(Condition) - bringt nicht viel, weil man es dann statisch nicht kennt
	//d.h. solche Sachen sollte man vorkompilieren, z.B. mit Annotationen
	//EObject getBaseObject()
	//SpecialClass getSpecialOf<X>() //oder null wenn es das nicht ist
	
	//Man sollte dynamisch methoden hinzufgen knnen, die man benutzen will ...

	static final String vmanager = "Manager";
	
	public final String classInPackageToNamePart(MergeStructuralFeature attr, Resource res, boolean doGet) {
		MergeClassifier type = doGet?attr.getGetType():attr.getSetType();
		if (type instanceof MergeClass) {
			return classInPackageToNamePart((MergeClass)type, res);
		}
		throw new RuntimeException("attr must be of type MergeClass but was " + attr.getClass());
		
	}
	
	public final String classInPackageToNamePart(MergeClass type, Resource res) {
		OutputClass clazz = getOrCreate(type, res);
		OutputPackage pkg = clazz.getPackage();
		if (pkg != null) {
			return pkg.getFullName().replace('.', '_')+"_"+type.getName();
		} else {
			return "_"+type.getName();
		}
	}
	
	
	private OutputClass virtualManagerClass;
	private OutputClass virtualBasicObject;
	private OutputClass modelFactoryIFace;
	private Map<String,OutputClass> modelFactoryImpl = new HashMap<>();
	
	private Map<String,Method> virtualManagerCreateMethods = new HashMap<>();
	private Method virtualManagerGetFactoryMethod;
	
	public String getvmanagerMethodName(MergeStructuralFeature feat, Resource res) {
		return "getOrCreate_" + /*classInPackageToNamePart(feat,res)+"_"+*/getBasicName(res);
	}
	
	
	public String getvmanagerMethodName(MergeClass cl, Resource res) {
		return "getOrCreate_" + /*classInPackageToNamePart(cl,res)+"_"+*/getBasicName(res);
	}
	
	public void getOrCreateCreateMethod(MergeClass type, Resource res) {
		String methodName = (res!=null)?"getOrCreate_"+getBasicName(res):"getOrCreate";
		if (virtualManagerCreateMethods.containsKey(methodName)) {
			return;
		}
		virtualManagerCreateMethods.put(methodName, null);
		OutputClass returned = getOrCreate((MergeClass)type, res);
		String retType = virtualManagerClass.foreignName(returned);
		List<NamedType> parList = new ArrayList<>();
		parList.add(new NamedType("EObject", "eobj"));
		StringBuilder implBuilder = new StringBuilder();
		implBuilder.append("\nif (eobj == null) {\n\treturn null;\n}");
		implBuilder.append("\nBasicObject ret = eobjmap.get(eobj);");
		implBuilder.append("\nif (ret == null) {");
		if (res == null) {
			implBuilder.append("\n\t"+virtualManagerClass.foreignName(modelFactoryIFace)+" factory = getFactory(eobj.eClass().eResource().getURI().toString());");
			implBuilder.append("\n\teobjmap.put(eobj, ret = factory.createObjFromFragment(eobj));");
		} else {
			implBuilder.append("\n\teobjmap.put(eobj, ret = factory_"+getBasicName(res)+".createObjFromFragment(eobj));");
		}
		implBuilder.append("\n}");
		implBuilder.append("\nreturn ret;");
		Method method = new Method("BasicObject",
				methodName, parList, implBuilder.toString());
		virtualManagerClass.addMethod(method);
		virtualManagerCreateMethods.put(methodName, method);
		
	}
	
	private OutputClass modelFactoryClass = null;
	
	public void getOrCreateCreateMethod(MergeStructuralFeature cl, Resource res) {
		
		String methodName = getvmanagerMethodName(cl,res);
		if (virtualManagerCreateMethods.containsKey(methodName)) {
			return;
		}
		virtualManagerCreateMethods.put(methodName, null);
		MergeClassifier type = cl.getGetType();
		/*EStructuralFeature thisFeat = null;
		for (EStructuralFeature feat: cl.getBaseStructuralFeatures()) {
			if (res.getURI().toString().equals(feat.eResource().getURI().toString())) {
				thisFeat = feat;
			}
		}
		if (thisFeat == null) {
			throw new RuntimeException("Could not find feature for " + cl.getName() + " in resource " + res);
		}*/
		;
		if (type instanceof MergeClass) {
			getOrCreateCreateMethod((MergeClass)type, res);
		} else {
			throw new RuntimeException("attr must be of type MergeClass but was " + type.getClass());
		}
	}
	
	private OutputPackage basePackage;
	
	public String vmanager() {
		return basePackage.getFullName()+"."+vmanager;
	}
	
	private static EObjectBase test = null;
	
	private Map<MergeClass,OutputClass> ecoreClassMap = new HashMap<>();
	
	/**Proxy-Collection notwendig*/
	
	private OutputClass generateClass(MergeClass mcl, OutputPackage pkg, Resource curResource, Map<Resource, OutputClass> map) {
;		OutputClass ret = new OutputClass(mcl.getName(), pkg, mcl,  false /*mcl.isInterface()*/);
		OutputClass ecoreClass = ecoreClassMap.get(mcl);
		boolean createEClass = false;
		if (ecoreClass == null) {
			ecoreClassMap.put(mcl,  ecoreClass = ret);
			createEClass = true;
		}
	

		
		map.put(curResource, ret);
		ret.addImport("org.eclipse.emf.ecore.EStructuralFeature");
		ret.addImport("org.eclipse.emf.ecore.EObject");
		ret.addImport("org.eclipse.emf.ecore.EClass");
		ret.addImport("org.eclipse.emf.ecore.EcoreFactory");
		ret.addImport("org.eclipse.emf.ecore.EAttribute");
		ret.addImport("org.eclipse.emf.ecore.EReference");
		ret.addImport("org.eclipse.emf.ecore.EDataType");
		
		pkg.addContainedClass(ret);
		
		getOrCreateCreateMethod(mcl,curResource);
		getOrCreateCreateMethod(mcl,null);
		
		
		boolean hasSuperClas = false;
		
		for (MergeClass mc: mcl.getSuperTypes()) {
			OutputClass sup = getOrCreate(mc, curResource);
			ret.addSuperClass(sup);
			if (sup.getClassType() != CLASS_TYPE.INTERFACE) {
				hasSuperClas = true;
			}
		}
		
		if (!hasSuperClas) {
			ret.setSuperClass("org.eclipse.emf.ecore.impl.MinimalEObjectImpl");
		}
		//Zustzlich - das jeweilige Interface
		ret.addSuperClass(getOrCreate(mcl, null));
		EClass thisClass = null;
				
		ret.addMethod(new Method("EObject","baseObject",Collections.<NamedType>emptyList(),"return _thiz;"));
		if (!createEClass) {
			ret.addStringTop("static {try {Class.forName(\""+ret.foreignName(ecoreClass)+"\");} catch (Exception e){}}\n");
		}
		
		ret.addField(new Field(new NamedType("static EClass", "_eclass"), createEClass?"EcoreFactory.eINSTANCE.createEClass()":(ecoreClass.getQualifiedName()+"._eclass")));
		if (createEClass) {
			ret.addEClassInit("_eclass.setName(\""+mcl.getName()+"\");");
			ret.addEClassInit("_eclass.setInterface("+mcl.isInterface()+");");
			ret.addEClassInit("_eclass.setAbstract("+mcl.isAbstract()+");");
			ret.addEClassInit("_eclass.setInstanceClass("+mcl.getName()+".class);");
			ret.addEClassInit(vmanager()+"._pkg.getEClassifiers().add(_eclass);");
		}
		List<NamedType> ntList = new ArrayList<NamedType>();
		ntList.add(new NamedType("int", "featureID"));
		ntList.add(new NamedType("boolean", "resolve"));
		ntList.add(new NamedType("boolean", "coreType"));
		
	

		
		ret.addInit("eSetClass(_eclass);");
		
		//Die Spezialisierungs-Methoden
		for (EClass ecl: mcl.getBase()) {
			String basicName = getBasicName(ecl.eResource());
			String implementation = "return null;";
			OutputPackage baseClassPkg = getOutputPkg(ecl.eResource());
			if (pkg == baseClassPkg) {
				implementation = "return ("+getJavaTypeName(mcl.getName(),ecl.eResource())+")this;";
				thisClass = ecl;
				ret.addMethod(new Method(getJavaTypeName(mcl.getName(),ecl.eResource()),"getSpecialOf"+Util.firstUpperCase(basicName), Collections.<NamedType> emptyList(), implementation));
			}
			ret.addImport("java.util.Collections");
			
		}
		
		ret.addMethod(new Method("EClass ","eStaticClass",Collections.<NamedType>emptyList(),"return _eclass;"));
		StringBuilder eSetImpl = new StringBuilder();
		StringBuilder eGetImpl = new StringBuilder();
		
		//eGetImpl.append("switch (featureID) {\n");
		//eSetImpl.append("switch (featureID) {\n");
		
		//Virtuelle variable: _thiz;
		//NamedType managerType = new NamedType(vmanager(), "vmanager");
		/*NamedType thizType = new NamedType(getJavaTypeName(thisClass), "thiz");
		ret.addField(new Field(new NamedType(getJavaTypeName(thisClass), "_thiz"),null));*/
		NamedType thizType = new NamedType("EObject", "thiz");
		ret.addImport("org.eclipse.emf.ecore.EObject");
		ret.addField(new Field(new NamedType("EObject", "_thiz"),null));
		//ret.addField(new Field(new NamedType(vmanager(), "_vmanager"),null));
		
		List<NamedType> connamedTypes = new ArrayList<>();
		//namedTypes.add(managerType);
		connamedTypes.add(thizType);
		StringBuilder constructionBuilder = new StringBuilder();
		if (hasSuperClas) {
			constructionBuilder.append("super(thiz);\n");
		}
		constructionBuilder.append("this._thiz = thiz;");
		
				
		
		
		//this._vmanager().getOrCreate<Paket>_<Typ>(returnwert)
	
		List<NamedType> ntsList = new ArrayList<>();
		ntsList.add(new NamedType("int", "featureID"));
		ntsList.add(new NamedType("Object", "newValue"));	
		
		for (MergeClass scl: mcl.getSuperTypes()) {
			ret.addEClassInit("try {Class.forName("+ret.foreignName(getOrCreate(scl, curResource))+".class.getCanonicalName());} catch(Exception e){}");
			ret.addEClassInit("_eclass.getESuperTypes().add("+ret.foreignName(getOrCreate(scl, curResource))+"._eclass);");
		}
		StringBuilder allExistingFeatures = new StringBuilder();
		int curFeatureId = 0;
		//Die Structural-Features-Methoden
		for (MergeStructuralFeature attr: mcl.getStructuralFeatures()) {
			if ("weekday".equals(attr.getName().toLowerCase())) {
				System.out.print("");
			}
			//Getter und has? auf jeden Fall
			//Kompatibel: Verweisen aufs selbe Attribut oder auf eines mit gemeinsamen Supertypen
			//Wobei es eigentlich egal ist - ein Object (bzw. List<Object>) ist es immer
			String getImpl = "return null;";
			String hasImpl = "return false;";
			//Reflection oder direkt?
			//Reflection geht immer, direkt ist hier einfacher zu implementieren, knnte aber woanders Schwierigkeiten
			//bereiten
			
			//Probiere ich es mal mit Reflection
			StringBuilder implBuilder = new StringBuilder();
			EStructuralFeature subFeat = null;
			for (EStructuralFeature feat: attr.getBaseStructuralFeatures()) {
				if (feat.getEContainingClass() == thisClass) {
					subFeat = feat;
				}
			}
			
			EStructuralFeature testF = null;
			
			if (attr instanceof MergeAttribute) {
				ret.addField(new Field(new NamedType("static EAttribute", "_"+attr.getName()),createEClass?"EcoreFactory.eINSTANCE.createEAttribute()":
					(ecoreClass.getQualifiedName()+"._"+attr.getName())));
				EAttribute testA = null;
			} else if (attr instanceof MergeReference) {
				ret.addField(new Field(new NamedType("static EReference", "_"+attr.getName()),createEClass?"EcoreFactory.eINSTANCE.createEReference()":
					(ecoreClass.getQualifiedName()+"._"+attr.getName())));				
				EReference testA = null;
				//TODO: EOpposite - aber da braucht man irgendsoein lazy zeug und ich msste das erst in MergeReference implementieren
			}
			
			//ret.addField(new Field(new NamedType("ESFGetter", "_getter_"+attr.getName()),null));
			
			//ret.addEClassInit("_getter_"+attr.getName()+" = _thiz.eClass().getEStructuralFeature(\""+attr.getName()+"\");\n");
			//ret.addField(new Field(new NamedType("static final int", "__fid_"+attr.getName()),null));
			Settability settability = (attr.isVirtualSettable() || subFeat != null)?Settability.SETTABLE:Settability.NOT_SETTABLE;
			if (createEClass && subFeat != null) {
				ret.addEClassInit("_eclass.getEStructuralFeatures().add(_"+attr.getName()+");");
				ret.addEClassInit("_"+attr.getName()+".setName(\""+attr.getName()+"\");");
				ret.addEClassInit("_"+attr.getName()+".setChangeable("+(attr.isChangeable() && attr.isVirtualSettable())+");");
				ret.addEClassInit("_"+attr.getName()+".setDerived(true);");
				ret.addEClassInit("_"+attr.getName()+".setUpperBound("+attr.getGetUpperBound()+");");
				ret.addEClassInit("_"+attr.getName()+".setLowerBound("+attr.getGetLowerBound()+");");
				
				
				
				if (attr.getGetType() instanceof MergeClass) {
					MergeClass tcl = (MergeClass)attr.getGetType();
					ret.addEClassInit("_"+attr.getName()+".setEType("+getJavaTypeCol(attr, curResource, ret, false,true,settability)+"._eclass"+");");
				} else if (attr.getGetType() instanceof MergeDataType) {
					//TODO: Mehr als nur die Standard-Datentypen supporten
					Class<?> cl = attr.getGetType().getUpperType();
					String ecoreTypeName = null;
					if (cl == String.class) {
						ecoreTypeName = "ESTRING";
					} else if (cl == Integer.class || cl == int.class) {
						ecoreTypeName = "EINT";
					} else if (cl == Double.class || cl == double.class) {
						ecoreTypeName = "EDOUBLE";
					} else if (cl == Long.class || cl == long.class) {
						ecoreTypeName = "ELONG";
					} else if (cl == Boolean.class || cl == boolean.class) {
						ecoreTypeName = "EBOOLEAN";
					} 
					MergeDataType mdt = (MergeDataType) attr.getGetType();
					if (ecoreTypeName == null) {
						System.err.println("Unrcceognized Data Type " + cl + " for Data Type " + attr.getName() + " - only primitve supported");
						ret.addEClassInit("EDataType temp = EcoreFactory.eINSTANCE.createEDataType();");
						ret.addEClassInit("temp.setName(\""+cl.getName()+"\");");
						ret.addEClassInit("temp.setInstanceClass("+cl.getCanonicalName()+".class);");
						ret.addEClassInit("_"+attr.getName()+".setEType(temp);");
					} else {
						ret.addEClassInit("_"+attr.getName()+".setEType(org.eclipse.emf.ecore.EcorePackage.Literals."+ecoreTypeName+");");
						
					}
					EStructuralFeature f;
				} else if (attr.getGetType() instanceof MergeEnum) {
					System.err.println("EType generation for enum not yet supported");
				}
			}
			MergeClassifier getAAType = attr.getGetType();
			//ret.addEClassInit("__fid_"+attr.getName()+" = _"+attr.getName()+".getFeatureID();");
			eGetImpl.append("\n\tif (featureID == _"+attr.getName()+".getFeatureID()) { return get"+Util.firstUpperCase(attr.getName())+"();}");
			
			Resource getResource = attr.isVirtualSettable()?null:curResource;
			//Wenn das ESF kein basis-ESF ist, dann:
			//
			String getType = getJavaTypeNoCol(attr,getResource,ret,true,settability);
			if (getType.startsWith("Collection<") && (attr.getGetType() instanceof MergeClass) && subFeat != null &&  (subFeat.getUpperBound() > 1 || subFeat.getUpperBound() == -1)) {
				getType = "CollectionProxy<"+getType.substring("Collection<".length());
				ret.addImport(basePackage.getFullName()+".CollectionProxy");
			}
			if (subFeat != null) {
				
				ret.addField(new Field(new NamedType("static EStructuralFeature", "_feat_"+attr.getName()),null));
				ret.addImport(basePackage.getFullName()+".ESFGetter");
				constructionBuilder.append("\n_getter_"+attr.getName()+" = new ESFGetter(this,_feat_"+attr.getName()+");\n");
				ret.addField(new Field(new NamedType("ESFGetter","_getter_"+attr.getName()),null));
				allExistingFeatures.append("\n_feat_"+attr.getName()+ " = ecl.getEStructuralFeature(\""+subFeat.getName()+"\");\n");
				implBuilder.append("EStructuralFeature subFeat = _thiz.eClass().getEStructuralFeature(\""+subFeat.getName()+"\");\n");
				if (attr.getGetUpperBound() > 1 || attr.getGetUpperBound() == -1) {
					//Konvertiere in Multizeugs
					ret.addImport("org.eclipse.emf.common.util.EList");
					ret.addImport("org.eclipse.emf.common.util.BasicEList");
					if (attr.getGetType() instanceof MergeClass) {
						
						String addString = null;						
						if (subFeat.getUpperBound() > 1 || subFeat.getUpperBound() == -1) {
							implBuilder.append("CollectionProxy<");
							implBuilder.append(getJavaTypeCol(attr,getResource,ret,false,true,settability));
							implBuilder.append("> ret = new CollectionProxy<>(_vmanager,(Collection)_thiz.eGet(subFeat));\n");
							implBuilder.append("return ret;\n");							
						} else {
							implBuilder.append("return new CollectionFromAttribute(_getter_"+attr.getName()+",_getter_"+attr.getName()+""+");\n");
							ret.addImport(basePackage.getFullName()+".CollectionFromAttribute");
						}
						//implBuilder.append(addString);
						//implBuilder.append("\nreturn ret;");
					} else {
						if (subFeat.getUpperBound() > 1 || subFeat.getUpperBound() == -1) {
							implBuilder.append("return (Collection)_thiz.eGet(subFeat);\n"); 
						} else {
							implBuilder.append("return new CollectionFromAttribute(_getter_"+attr.getName()+",_getter_"+attr.getName()+""+");\n");
							ret.addImport(basePackage.getFullName()+".CollectionFromAttribute");
						}
					}
				} else {
					String addString = null;
					if (attr.getGetType() instanceof MergeClass) {
						getOrCreateCreateMethod(attr, curResource);
						addString = "(" + getJavaTypeNoCol(attr,getResource,ret,true,settability)+")_vmanager."+getvmanagerMethodName(attr,curResource)
								+"((EObject)_thiz.eGet(subFeat))";
					} else if (attr.getGetType() instanceof MergeEnum) {
						ret.addImport("org.eclipse.emf.ecore.EEnumLiteral");
						implBuilder.append("EEnumLiteral subValue = (EEnumLiteral)_thiz.eGet(subFeat);");
						addString = getType+".nameToValueMap.get(subValue.getName())";
					} else {
						addString = "(" + getJavaTypeNoCol(attr,getResource,ret,true,settability)+")_thiz.eGet(subFeat)";
					}
					implBuilder.append("return "+addString+";");
				}
				getImpl = implBuilder.toString();
				hasImpl = "return true;";
			}
			if (subFeat != null) {
				ret.addMethod(new Method(getType,"get"+Util.firstUpperCase(attr.getName()),Collections.<NamedType>emptyList(),getImpl));
			}
			ret.addMethod(new Method("boolean","has"+Util.firstUpperCase(attr.getName()),Collections.<NamedType>emptyList(),hasImpl));
		
			
			
			
			//Speed improvement: Lege Table eClass -> EStructuralFeature an. Diese Tabelle kann dann immer verwendet werden
			//Mit index statt was anderem
			
			//Setter nur wenn passt
			//Eigentlich muss man hier auf kompatible Typen achten - set muss ja ein subtyp von dem Zeugs sein
			if (subFeat != null && attr.canSet()) {
				//Auerdem muss man auf Multiplizitten achten ...
				
				/*Biete alles an, um sicher zu gehen, dass das ganze mit dem Interface stimmt
				 * Interface: 1, multiplicity				
				 */
				//TODO: Vielleicht ist es sinnvoller, da andere Berechnungen fr die upperBound zu nehmen
				
				
				implBuilder = new StringBuilder();
				implBuilder.append("EStructuralFeature subFeat = _thiz.eClass().getEStructuralFeature(\""+subFeat.getName()+"\");\n");
				
				String addBaseObject = (attr.getGetType() instanceof MergeClass)?".baseObject()":"";
				if (attr.getGetType() instanceof MergeEnum) {
					String id = getBasicName(curResource);
					addBaseObject = ".getEnumLiteral("+id.charAt(id.length()-1)+")";
				}
				
				if (subFeat.isMany()) {
					ret.addImport("org.eclipse.emf.common.util.EList");
					ret.addImport("org.eclipse.emf.common.util.BasicEList");
					String type = (attr.getGetType() instanceof MergeClass)?"EObject":getJavaTypeCol(attr,null,ret,false,false,settability);
					String collectionName = "Collection<EObject>";
					if ("".equals(addBaseObject)) {
						collectionName = "Collection<"+getJavaTypeCol(attr,getResource,ret,false,false,settability)+">";
					}
					implBuilder.append("\n"+collectionName+" cmpList = get"+Util.firstUpperCase(attr.getName())+"()"+("".equals(addBaseObject)?"":".getReal()")+";");
					implBuilder.append("\n"+collectionName+" eList = cmpList;");
					if (attr.getSetUpperBound() == -1 || attr.getSetUpperBound() > 1) {
						implBuilder.append("\nif ("+attr.getName()+" instanceof CollectionProxy && (((CollectionProxy<?>)"+attr.getName()+").getReal() == cmpList)) {\n\treturn;\n}\n");
					}
					//Ansonsten sollte es sich nicht auf die gleiche Collection beziehen, man kann daher die beiden unabhngig behandeln
					implBuilder.append("if (cmpList != null) {\n\tcmpList.clear();\n} else {\ncmpList = new BasicEList<>();\n}\n"); //Atomar ist das ganze natrlich nicht mehr
					if (attr.getSetUpperBound() == -1 || attr.getSetUpperBound() > 1) {
						//Beides many
						//Konvertiere das Many in das Zeug. Das geht mit baseClass()
						implBuilder.append("\nfor ("+getJavaTypeCol(attr,getResource,ret,false,false,settability)+" o: "+attr.getName()+") {");
						if ("".equals(addBaseObject)) {
							implBuilder.append("\n\tcmpList.add(o"+addBaseObject+");");
						} else {
							implBuilder.append("\n\tcmpList.add(((o == null)?null:(o"+addBaseObject+")));");
						}
						implBuilder.append("}\n");
						
					} else {
						//Virtuelles single, das many
						if ("".equals(addBaseObject)) {
							implBuilder.append("\n\tcmpList.add("+attr.getName()+addBaseObject+");");
						} else {
							implBuilder.append("\n\tcmpList.add((("+attr.getName()+" == null)?null:("+attr.getName()+addBaseObject+")));");
						}
					}
					
					implBuilder.append("\nif (eList == null) {\n");
					implBuilder.append("\n_thiz.eSet(subFeat,cmpList);");
					implBuilder.append("\n}");
				} else {
					//Das single, aber es kommt many rein
					if (attr.getSetUpperBound() == -1 || attr.getSetUpperBound() > 1) {
						implBuilder.append("\nif (" + attr.getName()+ ".isEmpty()");
						if ("".equals(addBaseObject)) {
							implBuilder.append(" || "+attr.getName()+".iterator().next() == null");
						}
						implBuilder.append(") {");
						implBuilder.append("\n\t_thiz.eSet(subFeat,null);");
						implBuilder.append("\n} else {");
						implBuilder.append("\n\t_thiz.eSet(subFeat,"+attr.getName()+".iterator().next()"+addBaseObject+");");
						implBuilder.append("\n}");
					} else {
						//Beides single
						if ("".equals(addBaseObject)) {
							implBuilder.append("\n_thiz.eSet(subFeat,"+attr.getName()+");");
						} else {
							implBuilder.append("\n_thiz.eSet(subFeat,(("+attr.getName()+"==null)?null:("+attr.getName()+addBaseObject+")));");
						}
					}
				}
				
				List<NamedType> parameter = new ArrayList<>();
				parameter.add(new NamedType(getJavaTypeNoCol(attr,getResource,ret, false,settability),attr.getName()));
				ret.addMethod(new Method("void","set"+Util.firstUpperCase(attr.getName()),parameter,implBuilder.toString()));
				eSetImpl.append("\n\tif (featureID == _"+attr.getName()+".getFeatureID()) { set"+Util.firstUpperCase(attr.getName())+"(("+
						getJavaTypeNoCol(attr,getResource,ret, false,settability)+")newValue); return;}");
				
				//TODO: Type
				//Hier kann ich nur auf Interface-Sachen zugreifen
				//Mglicherweise sinnvoll: MergeZeugs --> TypName?
				
				
				implBuilder.append("_thiz.eSet(subFeat,"+attr.getName()+");");
				
				
			}
		}
		eGetImpl.append("\n");
		eSetImpl.append("\n");
		if (hasSuperClas) {
			eGetImpl.append("return super.eGet(featureID, resolve, coreType);\n");
			eSetImpl.append("super.eSet(featureID, newValue);\n");
		} else {
			eGetImpl.append("return eDynamicGet(featureID, resolve, coreType);\n");
			eSetImpl.append("eDynamicSet(featureID, newValue);\n");
		}
		
		ret.addMethod(new Method("Object","eGet",ntList,eGetImpl.toString()));
		ret.addMethod(new Method("void","eSet",ntsList,eSetImpl.toString()));
		
		
		Set<String> havePkg = new HashSet<>();
		
		for (Entry<String,MergeClass> entry: mcl.getPseudoSubclasses().entrySet()) {
			OutputManager man = ManagerManager.getOrCreateOutput(entry.getValue().getManager());
			havePkg.add(man.getPackageName());
			StringBuilder impl = new StringBuilder();
			//TODO: Nur wenn es es gibt
			
			String retType = getRetType(ret,man,entry.getValue());
			impl.append("return ("+retType+")"+man.vmanager()+".instance.getOrCreate(this.baseObject());");
			
			ret.addMethod(new Method(retType,"as" + Util.firstUpperCase(possiblyShortified(entry.getKey()).replace('.', '_')),Collections.EMPTY_LIST,impl.toString()));
		}
		
		for (Entry<String,MergeClass> entry: mcl.getPseudoSuperclasses().entrySet()) {
			OutputManager man = ManagerManager.getOrCreateOutput(entry.getValue().getManager());
			havePkg.add(man.getPackageName());
			StringBuilder impl = new StringBuilder();
			//TODO: Nur wenn es es gibt
			
			String retType = getRetType(ret,man,entry.getValue());
			impl.append("return ("+retType+")"+man.vmanager()+".instance.getOrCreate(this.baseObject());");
			
			ret.addMethod(new Method(retType,"as" + Util.firstUpperCase(possiblyShortified(entry.getKey()).replace('.', '_')),Collections.EMPTY_LIST,impl.toString()));
		}
		
		for (PseudoOutputManager pom: pseudoSubs) {
			if (havePkg.contains(pom.getBaseName())) {continue;}
			
			StringBuilder impl = new StringBuilder();
			
			String className = pom.getCommonSuperType(mcl.getBase());
			
			String retType = className;
			if (retType != null) {
			
				impl.append("return ("+retType+")"+pom.vmanager()+".instance.getOrCreate(this.baseObject());");
							
				ret.addMethod(new Method(retType,"as" + Util.firstUpperCase(possiblyShortified(pom.getBaseName()).replace('.', '_')),Collections.EMPTY_LIST,impl.toString()));
			}
			
		}
		
		for (PseudoOutputManager pom: pseudoSups) {
			if (havePkg.contains(pom.getBaseName())) {continue;}
			
			StringBuilder impl = new StringBuilder();
			
			String className = pom.getCommonSuperType(mcl.getBase());
			
			String retType = className;
			if (retType != null) {
			
				impl.append("return ("+retType+")"+pom.vmanager()+".instance.getOrCreate(this.baseObject());");
							
				ret.addMethod(new Method(retType,"as" + Util.firstUpperCase(possiblyShortified(pom.getBaseName()).replace('.', '_')),Collections.EMPTY_LIST,impl.toString()));
			}
			
		}
		
		ret.addConstructor(new Constructor(connamedTypes, constructionBuilder.toString()));
		
		TemplateProcessor cloneNewMethodM;
		try {
			cloneNewMethodM = new TemplateProcessor(new File("templates/class/staticMethods.java"));
			cloneNewMethodM.putParameter("mcl", mcl);
			cloneNewMethodM.putParameter("outputClass", ret);
			cloneNewMethodM.putParameter("cloneStateName", basePackage.getFullName()+".CloneState");
			cloneNewMethodM.putParameter("basicObjectName", basePackage.getFullName()+".BasicObject");
			cloneNewMethodM.putParameter("managerType", vmanager());
			cloneNewMethodM.putParameter("allExistingFeatures", allExistingFeatures.toString());
			ret.addBottomProcessor(cloneNewMethodM);
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		return ret;
	}
	
	public void simplePkgNames1(Collection<PseudoOutputManager> pom, Set<String> occuring, Set<String> duplicate) {
		for (PseudoOutputManager p: pom) {
			String baseName = p.getBaseName();
			if (baseName.lastIndexOf('.') != -1) {
				baseName = baseName.substring(baseName.indexOf('.')+1);
			}
			if (!occuring.add(baseName)) {
				duplicate.add(baseName);
			}
		}
	}
	
	public void simplePkgNames2(Collection<MergeManager> pom, Set<String> occuring, Set<String> duplicate) {
		for (MergeManager man: pom) {
			String baseName = man.getBasePackage();
			if (baseName.lastIndexOf('.') != -1) {
				baseName = baseName.substring(baseName.indexOf('.')+1);
			}
			if (!occuring.add(baseName)) {
				duplicate.add(baseName);
			}
		}
	}
		
	
	public String getRetType(OutputClass ret, OutputManager man, MergeClass mergeClass) {
	
		return ret.foreignName(man.getOrCreate(mergeClass, null));
	}
	
	private Set<String> duplicateSimplePackageNames = new HashSet<>();
	private Set<String> occurringSimplePackageNames = new HashSet<>();
	
	private List<PseudoOutputManager> pseudoSubs = new ArrayList<>();
	private List<PseudoOutputManager> pseudoSups = new ArrayList<>();
	private OutputClass cloneState;
	private OutputClass transferType;
	private OutputClass valueSetter;
	private OutputClass valueGetter;
	private OutputClass collectionProxy;
	private OutputClass esfGetter;
	private OutputClass collectionFromAttribute;
	
		
	public String getPackageName() {
		return this.basePackage.getFullName();
	}
	

	private OutputClass generateEnum(MergeEnum menum, OutputPackage inPackage, Resource res, Map<Resource,OutputClass> map) {
		OutputClass ret = new OutputClass(menum.getInstanceName(), inPackage, null, CLASS_TYPE.INTERFACE);
		map.put(res, ret);
		inPackage.addContainedClass(ret);
		ret.addImport("java.util.List");
		ret.addImport("java.util.ArrayList");
		ret.addImport("java.util.Map");
		ret.addImport("java.util.HashMap");
		ret.addImport("org.eclipse.emf.ecore.EEnumLiteral");
		for (MergeEnum sup: menum.getPseudoSuper()) {
			OutputClass osup = getOrCreate(sup, res);
			ret.addSuperClass(osup);
		}
		/*ret.addConstructor(new Constructor(Arrays.asList(new NamedType("String", "name"),
				new NamedType("String", "literal"), new NamedType("int", "value")
				), "this.name = name\nthis.literal = literal;\nthis.value = value;"));*/
		ret.addMethod(new Method("String", "getName", Arrays.asList(), null));
		ret.addMethod(new Method("String", "getLiteral", Arrays.asList(), null));
		ret.addMethod(new Method("int", "getValue", Arrays.asList(), null));
		ret.addMethod(new Method("void", "addEnumLiteral", Arrays.asList(new NamedType("int", "idx"), new NamedType("EEnumLiteral", "literal")), null));
		ret.addMethod(new Method("EEnumLiteral", "getEnumLiteral", Arrays.asList(new NamedType("int", "idx")), null));
		ret.addField(new Field(new NamedType("static final Map<String,"+menum.getInstanceName()+">","nameToValueMap"),"new HashMap<>()"));
		StringBuilder initializer = new StringBuilder();
		//If this is a topmost class, it generates all subenums
		List<MergeEnumLiteral> mels = menum.getLiterals();
		if (menum.getPseudoSuper().isEmpty()) {
			if (mels != null) {
				for (MergeEnumLiteral mel: mels) {
					String melClass = mel.getGetEnumContainer().getInstanceName();
					initializer.append("nameToValueMap.put(\""+mel.getName()+"\","+mel.getName()+");\n");
					//Erzeuge eines
					ret.addField(new Field(new NamedType("static final "+melClass,mel.getName()),"new "+melClass+"() {"
							+ "private List<EEnumLiteral> literals = new ArrayList<>();\n"
							+ "\tpublic String getName() {return \""+mel.getName()+"\";}\n\tpublic String getLiteral() {return \""+mel.getLiteral()+"\";}\n\tpublic int getValue() { return "+mel.getValue()+";}\n"
									+ "\tpublic void addEnumLiteral(int idx, EEnumLiteral literal) {while (this.literals.size()<=idx) {this.literals.add(null);}\nthis.literals.set(idx,literal);}\n"
									+ "\tpublic EEnumLiteral getEnumLiteral(int idx) {return this.literals.get(idx);}\n"
							+ "}"));
				}
			}
		} else {
			//Get superType
			MergeEnum superType = menum;
			while (!superType.getPseudoSuper().isEmpty()) {
				superType = superType.getPseudoSuper().iterator().next();
			}
			OutputClass superClOp = getOrCreate(superType, res);
			for (MergeEnumLiteral mel: mels) {
				String melClass = mel.getGetEnumContainer().getInstanceName();
				initializer.append("nameToValueMap.put(\""+mel.getName()+"\","+mel.getName()+");\n");
				//That's the target class now
				ret.addField(new Field(new NamedType("static final "+melClass,mel.getName()),ret.foreignName(superClOp)+"."+mel.getName()));
				
			}
		}
		List<String> names = new ArrayList<String>();
		if (mels != null) {
			for (MergeEnumLiteral mel: mels) {
				names.add(mel.getName());
			}
		}
		StringBuilder namesStr = new StringBuilder();
		{
			boolean first = true;
			for (String name: names) {
				if (first) {first = false;} else {namesStr.append(", ");}
				namesStr.append(name);
			}
		}
		for (EEnum base: menum.getBase()) {
			String uri = base.eResource().getURI().toString();
			OutputClass fcl = modelFactoryImpl.get(uri);
			System.out.println("Having base: " + uri + " for package " + ret.foreignName(fcl));
			fcl.addInit("try {Class.forName(\""+fcl.foreignName(ret)+"\");} catch (Exception e) {}");
			fcl.addInit("uriToClass.put(\""+base.eResource().getURIFragment(base)+"\","+fcl.foreignName(getOrCreate(menum,null))+".class);");
		}
		
		ret.addField(new Field(new NamedType("static final "+menum.getInstanceName()+"[]","values"), "new "+menum.getInstanceName()+"[]{"+namesStr+"}"));
		ret.addStringBottom("static final MyInit myInit = new MyInit(); static class MyInit {public MyInit() {\n"+initializer+"}}\n");
		return ret;
	}
	
	/**Manager braucht:
	 *  - Factorymethoden fr Klassen
	 *  - Initial: Einlesen der gegebenen Ecore-Dateien (d.h. die mssen auch provided werden (aber indirekt mit den Ressourcen geht da vielleicht was))
	 *  - 
	 */

	public OutputClass generateInterface(MergeClass mcl, OutputPackage inPackage, Resource curResource, Map<Resource, OutputClass> map) {
		OutputClass ret = new OutputClass(mcl.getName(), inPackage, mcl, true);
		map.put(curResource, ret);
		inPackage.addContainedClass(ret);
		ret.addImport("org.eclipse.emf.ecore.EObject");
		ret.addSuperClass(virtualBasicObject);

		
		if (curResource == null && !mcl.isInterface()) {
			//Generate factory Method
			StringBuilder fmb = new StringBuilder();
			
			List<NamedType> factoryParameter = new ArrayList<>();
			factoryParameter.add(new NamedType(virtualManagerClass.foreignName(virtualBasicObject), "modelObject"));
			
			fmb.append("String resource = modelObject.baseObject().eResource().getURI().toString();"); //Hole den Resourcenstring, wie immer
			//Je nachdem erzeuge unterschiedliche EObjects und dann unterschiedliche Objekte selbst
			fmb.append("\n"+virtualManagerClass.foreignName(modelFactoryIFace)+" factory = getFactory(resource);");
			fmb.append("\nreturn factory.create"+classInPackageToNamePart(mcl, null)+"();");
			
			
			
			String methodName = "create"+classInPackageToNamePart(mcl, null);
			
			modelFactoryIFace.addMethod(new Method(modelFactoryIFace.foreignName(ret),
					methodName,Collections.emptyList(), "throw new UnsupportedOperationException();"));
			
			Method factoryMethod = new Method(virtualManagerClass.foreignName(ret)
					, methodName, factoryParameter, fmb.toString());
			virtualManagerClass.addMethod(factoryMethod);
			
			Set<String> remaining = new HashSet<String>();
			remaining.addAll(basicNames.keySet());
			remaining.remove(null);
			remaining.remove(vmanager()); //Bh ...
			
			for (EClass base: mcl.getBase()) {
				String uri = base.eResource().getURI().toString();
				remaining.remove(uri);
				OutputClass fcl = modelFactoryImpl.get(uri);
				StringBuilder impl = new StringBuilder();
				impl.append("return new " + 
						fcl.foreignName(getOrCreate(mcl, base.eResource()))+"("+
						"createFromFragment(\"" + base.eResource().getURIFragment(base) + "\"));");
				fcl.addMethod(new Method(fcl.foreignName(ret),methodName,Collections.emptyList()/*Collections.singletonList(new
						NamedType(fcl.foreignName(virtualManagerClass),"vmanager"))*/,impl.toString()));
				fcl.addInit("try {Class.forName(\""+fcl.foreignName(getOrCreate(mcl,base.eResource()))+"\");} catch (Exception e) {}");
				fcl.addInit("uriToClass.put(\""+base.eResource().getURIFragment(base)+"\","+fcl.foreignName(getOrCreate(mcl,base.eResource()))+".class);");
			}
			
			String impl = "throw new UnsupportedOperationException();";
			
//			for (String str: remaining) {
//				
//				OutputClass fcl = modelFactoryImpl.get(str);
//				if (fcl == null) {
//					System.out.println("No class found for " + str);
//					for (String st: modelFactoryImpl.keySet()) {
//						System.out.println(st);
//					}
//				}
//				fcl.addMethod(new Method(fcl.foreignName(ret),methodName,Collections.emptyList()/*,Collections.singletonList(new
//						NamedType(fcl.foreignName(virtualManagerClass),"vmanager"))*/,impl));
//			}
		}
		
		for (MergeClass mc: mcl.getSuperTypes()) {
			OutputClass cl = getOrCreate(mc, curResource);
			ret.addSuperClass(cl);
		}
		
		
		
		ret.addMethod(new Method("EObject","baseObject",Collections.<NamedType>emptyList(),null));
		
		for (EClass ecl: mcl.getBase()) {
			String basicName = getBasicName(ecl.eResource());
			ret.addMethod(new Method(getJavaTypeName(mcl.getName(),ecl.eResource()),"getSpecialOf"+Util.firstUpperCase(basicName), Collections.<NamedType> emptyList(), "return null;"));
		}
		
		/*for (Entry<String,MergeClass> sub: mcl.getPseudoSubclasses().entrySet()) {
			MergeClass subClass = sub.getValue();
			OutputManager man = ManagerManager.getOrCreateOutput(subClass.getManager());
			String basicName = man.getBasicName(null);
			ret.addMethod(new Method(man.getJavaTypeName(subClass.getName(), null), "pseudoSubOf"+basicName, Collections.<NamedType>emptyList(), null));
		}*/
		StringBuilder setAttrImpl = new StringBuilder();
		for (MergeStructuralFeature attr: mcl.getStructuralFeatures()) {
			//Getter und has? auf jeden Fall
			//Kompatibel: Verweisen aufs selbe Attribut oder auf eines mit gemeinsamen Supertypen
			//Wobei es eigentlich egal ist - ein Object (bzw. List<Object>) ist es immer
			String namePart = Util.firstUpperCase(attr.getName());
			String impl = attr.isGetMulti()?"return Collections.emptyList();":"return null;";
			if (attr.isGetMulti()) {
				ret.addImport("java.util.Collections");
			}
			Settability settability = attr.isVirtualSettable()?Settability.SETTABLE:Settability.NOT_SETTABLE;
			ret.addMethod(new Method(getJavaTypeNoCol(attr,null,ret, true, settability),"get"+namePart,Collections.<NamedType>emptyList(),
					impl));
			ret.addMethod(new Method("boolean","has"+namePart,Collections.<NamedType>emptyList(),null));
			
			//Setter nur wenn passt
			//Eigentlich muss man hier auf kompatible Typen achten - set muss ja ein subtyp von dem Zeugs sein
			if (attr.isVirtualSettable()) {
				List<NamedType> parameter = new ArrayList<>();
				//TODO: Type
				//Hier kann ich nur auf Interface-Sachen zugreifen
				//Mglicherweise sinnvoll: MergeZeugs --> TypName?
				parameter.add(new NamedType(getJavaTypeNoCol(attr,null,ret, false,settability),attr.getName()));
				String implementation = null;
				ret.addMethod(new Method("void","set"+namePart,parameter,implementation));
				if ("Matrnr".equals(namePart)) {
					System.out.print("");
				}
				boolean reduceMultiplicity = (attr.getGetUpperBound()==-1 || attr.getGetUpperBound() > 1) && (attr.getSetUpperBound() >= 0 && attr.getSetUpperBound() <= 1);
				String reduceStr = "";
				String ensureType = "";
				if (attr.getGetType() != attr.getSetType()) {
					if (attr instanceof MergeReference) {
						if (reduceMultiplicity) {
							ensureType = "ensureSingle("+getJavaTypeCol(attr, null, ret, false, false,settability)+".class,cs.";
						} else {
							if (attr.getGetUpperBound() == -1 || attr.getGetUpperBound() > 1) {
								ensureType = "ensureCol("+getJavaTypeCol(attr, null, ret, false, false,settability)+".class,cs.";
							} else {
								ensureType = "ensureBasic("+getJavaTypeCol(attr, null, ret, false, false,settability)+".class,cs.";
							}
						}
					} else {
						if (reduceMultiplicity) {
							ensureType = "cs.ensureSingle("+getJavaTypeCol(attr, null, ret, false, false,settability)+".class,";
						} else {
							if (attr.getGetUpperBound() == -1 || attr.getGetUpperBound() > 1) {
								ensureType = "cs.ensureCol("+getJavaTypeCol(attr, null, ret, false, false,settability)+".class,";
							} else {
								ensureType = "cs.ensureBasic("+getJavaTypeCol(attr, null, ret, false, false,settability)+".class,";
							}
						}
					}
					reduceStr = ")";
				} else {
					if (reduceMultiplicity) {
						if (attr instanceof MergeReference) {
							ensureType = "ensureSingle(cs.";
						} else {
							ensureType = "cs.ensureSingle(";
						}
						reduceStr = ")";
					}
				}
				//String ensureType = (attr.getGetType() != attr.getSetType())?(); 
				if (attr instanceof MergeReference) {
					setAttrImpl.append("if (has"+namePart+"()) {\n\tset"+namePart+"(cs."+ensureType+"clone(other.get"+namePart+"())"+reduceStr+");\n}\n");
				} else {
					setAttrImpl.append("if (has"+namePart+"()) {\n\tset"+namePart+"("+ensureType+"other.get"+namePart+"()"+reduceStr+");\n}\n");
				}
			}
		}
		
		Set<String> havePkg = new HashSet<>();
		
		
		for (Entry<String,MergeClass> entry: mcl.getPseudoSubclasses().entrySet()) {
			OutputManager man = ManagerManager.getOrCreateOutput(entry.getValue().getManager());
			havePkg.add(man.getPackageName());
			//TODO: Nur wenn es es gibt
			
			String retType = getRetType(ret,man,entry.getValue());
			
			ret.addMethod(new Method(retType,"as" + Util.firstUpperCase(possiblyShortified(entry.getKey()).replace('.', '_')),Collections.EMPTY_LIST,"return null;"));
		}
		
		for (Entry<String,MergeClass> entry: mcl.getPseudoSuperclasses().entrySet()) {
			OutputManager man = ManagerManager.getOrCreateOutput(entry.getValue().getManager());
			havePkg.add(man.getPackageName());
			//TODO: Nur wenn es es gibt
			
			String retType = getRetType(ret,man,entry.getValue());
			
			ret.addMethod(new Method(retType,"as" + Util.firstUpperCase(possiblyShortified(entry.getKey()).replace('.', '_')),Collections.EMPTY_LIST,"return null;"));
		}
		
		for (PseudoOutputManager pom: pseudoSubs) {
			if (havePkg.contains(pom.getBaseName())) {continue;}
			
			String className = pom.getCommonSuperType(mcl.getBase());			
			String retType = className;
			if (retType != null) {
				ret.addMethod(new Method(retType,"as" + Util.firstUpperCase(possiblyShortified(pom.getBaseName()).replace('.', '_')),Collections.EMPTY_LIST,"return null;"));
			}
			
		}
		
		for (PseudoOutputManager pom: pseudoSups) {
			if (havePkg.contains(pom.getBaseName())) {continue;}
			
			String className = pom.getCommonSuperType(mcl.getBase());			
			String retType = className;
			if (retType != null) {
				ret.addMethod(new Method(retType,"as" + Util.firstUpperCase(possiblyShortified(pom.getBaseName()).replace('.', '_')),Collections.EMPTY_LIST,"return null;"));
			}
			
		}
		if (ret != virtualBasicObject) {
			for (OutputClass cl: ret.getSuperClasses()) {
				if (cl != virtualBasicObject) {
					setAttrImpl.append("\n"+ret.foreignName(cl)+".super.setAttributesFrom(other,cs);\n");
				}
			}
		}
		
		Method setAttributes = new Method("void","setAttributesFrom",Arrays.asList(new NamedType(ret.getName(),"other"), new NamedType(basePackage.getFullName()+".CloneState","cs")),setAttrImpl.toString());
		ret.addMethod(setAttributes);
		
		TemplateProcessor cloneNewMethodM;
		try {
			cloneNewMethodM = new TemplateProcessor(new File("templates/class/cloneNew.java"));
			cloneNewMethodM.putParameter("mcl", mcl);
			cloneNewMethodM.putParameter("outputClass", ret);
			cloneNewMethodM.putParameter("cloneStateName", basePackage.getFullName()+".CloneState");
			cloneNewMethodM.putParameter("basicObjectName", basePackage.getFullName()+".BasicObject");
			cloneNewMethodM.putParameter("managerType", vmanager());
			String createMethod = classInPackageToNamePart(mcl, null);
			cloneNewMethodM.putParameter("createMyself", createMethod);
			ret.addBottomProcessor(cloneNewMethodM);
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return ret;
	}
	
	public String possiblyShortified(String fullName) {
		int lind = fullName.lastIndexOf('.');
		String name = fullName;
		if (lind != -1) {
			name = fullName.substring(lind+1);
			if (duplicateSimplePackageNames.contains(name)) {
				return fullName;
			}
		}
		return name;
	}
	
	public OutputClass createFactoryStub(String resourceURI, OutputPackage virtual) {
		OutputClass ret = new OutputClass("Factory"+basicNames.get(resourceURI),virtual,null, false);
		ret.addSuperClass(modelFactoryIFace);
		virtual.addContainedClass(ret);
		ret.addImport("java.util.Map");
		ret.addImport("java.util.HashMap");
		ret.addImport("java.io.IOException");
		ret.addImport("org.eclipse.emf.common.util.TreeIterator");
		ret.addImport("org.eclipse.emf.ecore.EClass");
		ret.addImport("org.eclipse.emf.ecore.EObject");
		ret.addImport("org.eclipse.emf.ecore.EPackage");
		ret.addImport("org.eclipse.emf.ecore.EcoreFactory");
		ret.addImport("org.eclipse.emf.ecore.EEnum");
		ret.addImport("org.eclipse.emf.ecore.EEnumLiteral");
		ret.addImport("org.eclipse.emf.ecore.resource.Resource");
		ret.addImport("org.eclipse.emf.ecore.resource.ResourceSet");
		ret.addImport("org.eclipse.emf.ecore.resource.impl.ResourceSetImpl");
		ret.addImport("org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl");
		ret.addImport("org.eclipse.emf.ecore.xmi.impl.EcoreResourceFactoryImpl");
		ret.addImport("org.eclipse.emf.common.util.URI");
		modelFactoryImpl.put(resourceURI, ret);
		ret.addField(new Field(new NamedType("Map<String,Class<?>>","uriToClass"),"new HashMap<>()"));
		//Init
		NamedType vmanagerType = new NamedType(vmanager(),"manager");
		
		ret.addField(new Field(new NamedType("ResourceSet","resourceSet"),null));
		ret.addField(new Field(vmanagerType,null));
		
		ret.addConstructor(new Constructor(Arrays.asList(new NamedType("ResourceSet","resourceSet"), vmanagerType),
				"this.resourceSet = resourceSet;\nthis.manager = manager;\ninit();"));
		
		//initMethod.append("ResourceSet resourceSet = new ResourceSetImpl();");
		//initMethod.append("\nresourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(\"xmi\", new XMIResourceFactoryImpl());");
		//initMethod.append("\nresourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(\"ecore\", new EcoreResourceFactoryImpl());");
		String ecoreFile = resourceURI;
		
		TemplateProcessor initProcessor;
		StringBuilder initMethod = new StringBuilder();
		try {
			initProcessor = new TemplateProcessor(new File("templates/class/factoryInitMethod.java"),"ecoreFile");
			String idx = basicNames.get(resourceURI);
			idx = idx.substring(idx.length()-1);
			initProcessor.putParameter("myNmb", idx);
			initProcessor.process(initMethod,new int[]{0}, ecoreFile);
		} catch (IOException | ParseException e) {
			e.printStackTrace();
		}
		
		
		ret.addMethod(new Method("void","init",Collections.<NamedType>emptyList(),initMethod.toString()));
		
		StringBuilder implBuild = new StringBuilder();
		
		implBuild.append("String uriFragment = eobj.eClass().eResource().getURIFragment(eobj.eClass());");
		implBuild.append("\nClass<?> cl = uriToClass.get(uriFragment);");
		implBuild.append("\ntry {\n\tObject ret = cl.getConstructor(EObject.class).newInstance(eobj);");
		implBuild.append("\n\treturn ("+ret.foreignName(virtualBasicObject)+")ret;");
		implBuild.append("\n} catch (Exception e) {\n\te.printStackTrace();\n}\nreturn null;");
		
		
		List<NamedType> param = new ArrayList<>();
		param.add(new NamedType("EObject","eobj"));
		//param.add(new NamedType(ret.foreignName(virtualManagerClass),"vmanager"));
		
		ret.addMethod(new Method(ret.foreignName(virtualBasicObject),"createObjFromFragment",param,implBuild.toString()));
		
		ret.addField(new Field(new NamedType("Resource", "mainEcore"),null));
		ret.addField(new Field(new NamedType("Map<String,EClass>", "urifragmentToClass"),"new HashMap<>()"));
		
		ret.addConstructor(new Constructor(Arrays.asList(vmanagerType), "this.manager = manager;\ninit();"));
		
		List<NamedType> parameters = new ArrayList<>();
		
		StringBuilder createFromFragment = new StringBuilder();
		createFromFragment.append("EClass cl = urifragmentToClass.get(uriFragment);");
		createFromFragment.append("\nif (cl != null) {");
		createFromFragment.append("\n\treturn new org.eclipse.emf.ecore.impl.DynamicEObjectImpl(cl);");
		createFromFragment.append("\n} else {\n\treturn null;\n}");
		ret.addMethod(new Method("EObject","createFromFragment",
				Collections.singletonList(new NamedType("String","uriFragment")),
						createFromFragment.toString()));
		
		virtualManagerGetFactoryMethod.implementation =
				"if (\""+resourceURI+"\".equals(resStr)) {\n"+
				"\treturn factory_"+basicNames.get(resourceURI)+";"+
				"\n}\n"+virtualManagerGetFactoryMethod.implementation;
		String vt = basicNames.get(resourceURI);
		virtualManagerClass.addField(new Field(new NamedType(virtualManagerClass.foreignName(ret),"factory_"+
		basicNames.get(resourceURI)),null));
		virtualManagerClass.addMethod(new Method("CloneState", "getCloneState"+Util.firstUpperCase(vt),
				Collections.emptyList(), "return new CloneState(this, factory_"+vt+", new java.util.IdentityHashMap<BasicObject, BasicObject>());"));
		
		ResourceSet resourceSet = new ResourceSetImpl();
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("xmi", new XMIResourceFactoryImpl());
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put("ecore", new EcoreResourceFactoryImpl());
		virtualManagerClass.addInit("factory_" + basicNames.get(resourceURI)+ " = new "+virtualManagerClass.foreignName(ret)+"(resourceSet,this);");
				
		return ret;
	}
	
	
	/**TODO: Schnere Factories durch Singletons*/
	
	/*Jetzt muss man irgendwie Modelle laden knnen - wobei eigentlich immer nur ein Modell,
	 * das wird dann virtualisiert
	 */
	
	//Erzeuge: 1 gemeinsame Interface-hierarchie fr alle (Resource "null")
	
	//Pro resource 1 implementierung fr die Interface-hierarchie
	//Parameter: Set<Resourcen>, die tatschlich verwendet werden
	//Theoretisch wre es sogar denkbar, hier noch einschrnkungen zu machen
	//Wodurch man noch genaueren Zugriff htte
	//Aber praktisch drfte das nicht sehr relevant sein
	public static OutputManager getOutput(MergeManager manager, OutputManager ret, List<PseudoOutputManager> oms) {
		OutputPackage basePkg = new OutputPackage(manager.getBasePackage(), null);
		ret.basePackage = basePkg;
		
		for (PseudoOutputManager om: oms) {
			if (om.getEcoreFiles().size() < manager.getResourcesString().size() && manager.getResourcesString().containsAll(om.getEcoreFiles())) {
				ret.pseudoSubs.add(om);
			}
			System.out.println(manager.getBasePackage()+": " + Arrays.toString(manager.getResourcesString().toArray()) + " VS " + Arrays.toString(om.getEcoreFiles().toArray()));
			if (om.getEcoreFiles().size() > manager.getResourcesString().size() && om.getEcoreFiles().containsAll(manager.getResourcesString())) {
				ret.pseudoSups.add(om);
			}
		}
		
		{
			int id = 1; 
			
			ret.basicNames.put(null, "minterface");
			for (Resource res: manager.getResources()) {
				ret.basicNames.put(res.getURI().toString(),"vt"+id);
				++id;
			}
		}
		
		ret.duplicateSimplePackageNames = new HashSet<>();
		ret.occurringSimplePackageNames = new HashSet<>();
		ret.simplePkgNames1(ret.pseudoSubs, ret.occurringSimplePackageNames, ret.duplicateSimplePackageNames);
		ret.simplePkgNames1(ret.pseudoSups, ret.occurringSimplePackageNames, ret.duplicateSimplePackageNames);
		ret.simplePkgNames2(manager.getSuperManagers(), ret.occurringSimplePackageNames, ret.duplicateSimplePackageNames);
		ret.simplePkgNames2(manager.getSubManagers(), ret.occurringSimplePackageNames, ret.duplicateSimplePackageNames);
		
		OutputPackage virtual = new OutputPackage("virtual", basePkg);
		ret.virtualManagerClass = new OutputClass("Manager", basePkg, null,false);
		
		ret.virtualManagerClass.addImport("java.util.Map");
		ret.virtualManagerClass.addImport("java.util.HashMap");
		ret.virtualManagerClass.addImport("java.io.File");
		ret.virtualManagerClass.addImport("org.eclipse.emf.ecore.EObject");
		ret.virtualManagerClass.addImport("org.eclipse.emf.ecore.EcoreFactory");
		ret.virtualManagerClass.addImport("org.eclipse.emf.ecore.EPackage");
		
		ret.virtualManagerClass.addField(new Field(new NamedType("static final EPackage","_pkg"),"EcoreFactory.eINSTANCE.createEPackage()"));
		ret.virtualManagerClass.addEClassInit("_pkg.setName(\""+basePkg.getFullName().replace('.', '_')+"\");");
		
		try {
			TemplateProcessor tproc = new TemplateProcessor(new File("templates/package/CloneState.java"),"outputClass");
			OutputClass cloneState = new OutputClass(tproc, "CloneState",basePkg,null);
			ret.cloneState = cloneState;
			tproc.putParameter("managerType",ret.cloneState.foreignName(ret.virtualManagerClass));
			
			basePkg.addContainedClass(ret.cloneState);
		} catch (IOException e) {
			e.printStackTrace();
		}
		try {
			OutputClass transferType = new OutputClass(new TemplateProcessor(new File("templates/package/TransferType.java"), "outputClass"),"TransferType",basePkg,null);
			ret.transferType = transferType;
			basePkg.addContainedClass(ret.transferType);
		} catch (IOException e) {
			e.printStackTrace();
		}
		

		try {
			OutputClass valueSetter = new OutputClass(new TemplateProcessor(new File("templates/package/ESFGetter.java"), "outputClass"),"ESFGetter",basePkg,null);
			ret.esfGetter = valueSetter;
			basePkg.addContainedClass(ret.esfGetter);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		try {
			OutputClass valueGetter = new OutputClass(new TemplateProcessor(new File("templates/package/ValueGetter.java"), "outputClass"),"ValueGetter",basePkg,null);
			ret.valueGetter = valueGetter;
			basePkg.addContainedClass(ret.valueGetter);			
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		try {
			OutputClass valueSetter = new OutputClass(new TemplateProcessor(new File("templates/package/ValueSetter.java"), "outputClass"),"ValueSetter",basePkg,null);
			ret.valueSetter = valueSetter;
			basePkg.addContainedClass(ret.valueSetter);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		try {
			OutputClass valueSetter = new OutputClass(new TemplateProcessor(new File("templates/package/CollectionProxy.java"), "outputClass"),"CollectionProxy",basePkg,null);
			ret.collectionProxy = valueSetter;
			basePkg.addContainedClass(ret.collectionProxy);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		try {
			OutputClass valueSetter = new OutputClass(new TemplateProcessor(new File("templates/package/CollectionFromAttribute.java"), "outputClass"),"CollectionFromAttribute",basePkg,null);
			ret.collectionFromAttribute = valueSetter;
			basePkg.addContainedClass(ret.collectionFromAttribute);
		} catch (IOException e) {
			e.printStackTrace();
		}
				
		OutputClass basicObject = new OutputClass("BasicObject", basePkg, null,true);
		basicObject.addInterface("EObject");
		ret.virtualBasicObject = basicObject;
		
		ret.virtualManagerClass.addMethod(ret.virtualManagerGetFactoryMethod = new Method(virtual.getFullName()+".ModelFactory","getFactory",Collections.singletonList(new NamedType("String","resStr")),"return null;"));
				
		basicObject.addMethod(new Method("BasicObject", "cloneNew", Arrays.asList(new NamedType("CloneState", "cs")), null));
		
				
		
		

		
		basePkg.addContainedClass(ret.virtualManagerClass);
		
		Map<String,OutputPackage> op = new HashMap<>();
		
		OutputClass modelFactory = new OutputClass("ModelFactory",virtual,null,true);
		ret.modelFactoryIFace = modelFactory;
		modelFactory.addImport("org.eclipse.emf.ecore.EObject");
		modelFactory.addImport("java.util.Map");
		modelFactory.addImport("java.util.HashMap");
		modelFactory.addMethod(new Method("EObject","createFromFragment",Collections.singletonList(new NamedType("String","uriFragment")),null));
		
		List<NamedType> param = new ArrayList<>();
		param.add(new NamedType("EObject","eobj"));
		//param.add(new NamedType(modelFactory.foreignName(ret.virtualManagerClass),"vmanager"));
		
		modelFactory.addMethod(new Method(modelFactory.foreignName(ret.virtualBasicObject),"createObjFromFragment",param,null));
		
		
		
		

		virtual.addContainedClass(modelFactory);
		
		
		basicObject.addImport("org.eclipse.emf.ecore.EObject");
		basicObject.addMethod(new Method("EObject", "baseObject", Collections.<NamedType>emptyList(), null));
		basePkg.addContainedClass(basicObject);
		
		op.put(basePkg.getFullName(), basePkg);
		
		op.put(ret.vmanager(), virtual);
		ret.basicNames.put(ret.vmanager(), "vm");
		ret.output = op;
		
		ret.virtualManagerClass.addImport("org.eclipse.emf.ecore.resource.Resource");
		ret.virtualManagerClass.addImport("org.eclipse.emf.ecore.resource.ResourceSet");
		ret.virtualManagerClass.addImport("org.eclipse.emf.common.util.TreeIterator");
		ret.virtualManagerClass.addImport("org.eclipse.emf.common.util.URI");
		ret.virtualManagerClass.addImport("java.util.List");
		ret.virtualManagerClass.addImport("java.util.ArrayList");
		ret.virtualManagerClass.addImport(basePkg.getFullName()+".BasicObject");;
		ret.virtualManagerClass.addImport("org.eclipse.emf.ecore.resource.impl.ResourceSetImpl");
		ret.virtualManagerClass.addImport("org.eclipse.emf.ecore.xmi.impl.XMIResourceFactoryImpl");
		
		ret.virtualManagerClass.addField(new Field(new NamedType("Map<EObject,BasicObject>","eobjmap"),"new HashMap<>()"));
		ret.virtualManagerClass.addField(new Field(new NamedType("ResourceSet","resourceSet"),"new ResourceSetImpl()"));
		ret.virtualManagerClass.addInit("resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(\"xmi\", new XMIResourceFactoryImpl());");
		ret.virtualManagerClass.addInit("resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(\"ecore\", new EcoreResourceFactoryImpl());");
		ret.virtualManagerClass.addImport("org.eclipse.emf.ecore.xmi.impl.EcoreResourceFactoryImpl");
		
		StringBuilder impl = new StringBuilder();
		impl.append("\nResource resource = resourceSet.createResource(URI.createFileURI(file.getAbsolutePath()));");
		impl.append("\nresource.load(new HashMap<>());");
		impl.append("\nList<BasicObject> ret = new ArrayList<>();");
		impl.append("\nfor (EObject obj: resource.getContents()) {");		
		impl.append("\n\tif (obj != null && obj.eClass() != null) {");
		impl.append("\n\t\tString uri = obj.eClass().eResource().getURI().toString();");
		impl.append("\n\t\t"+basePkg.getFullName()+".virtual.ModelFactory"+" fact = getFactory(uri);");
		impl.append("\n\t\tif (fact == null) {");
		impl.append("\n\t\t\tthrow new java.util.UnknownFormatConversionException(\"Could not load object with URI \" + uri +");
		impl.append("\n\t\t\t\", supported are: \"");
		boolean first = true;
		for (String str: manager.getResourcesString()) {
			impl.append("\n\t\t\t + \""+(first?"":", ")+str+"\"");
			first = false;
		}
		impl.append(");\n\t\t}");
		impl.append("\n\t\tret.add(fact.createObjFromFragment(obj));");
		impl.append("\n\t}");
		impl.append("\n}");
		impl.append("\nreturn ret;");
		Method loadFile = new Method("Collection<? extends "+ret.virtualManagerClass.foreignName(ret.virtualBasicObject)+">","loadFile",Collections.singletonList(new NamedType("File","file")),
				impl.toString());
		loadFile.exceptions.add("java.io.IOException");
		
		
		Method loadFileStatic = new Method("static Collection<? extends "+ret.virtualManagerClass.foreignName(ret.virtualBasicObject)+">","load",Collections.singletonList(new NamedType("File","file")),
				"return instance.loadFile(file);");
		loadFileStatic.exceptions.add("java.io.IOException");
		
		StringBuilder saveFileImpl = new StringBuilder();
		
		saveFileImpl.append("Resource resource = null;\n");
		saveFileImpl.append("java.util.Map<Object,Object> options = new HashMap<>();\n");
		saveFileImpl.append("options.put(org.eclipse.emf.ecore.xmi.XMLResource.OPTION_SCHEMA_LOCATION, true);\n");
		saveFileImpl.append("for (BasicObject bo: boCol) {\n");
		
		saveFileImpl.append("\tif (resource == null) {\n");
		saveFileImpl.append("\t\tresource = resourceSet.createResource(bo.baseObject().eClass().eResource().getURI());\n");
		saveFileImpl.append("\t}");
		saveFileImpl.append("\tresource.getContents().add(bo.baseObject());\n");
		saveFileImpl.append("}\nresource.setURI(URI.createURI(resource.getURI().toString()+\" asdf\"));\n");
		saveFileImpl.append("resource.save(new java.io.FileOutputStream(file), options);\n");
		List<NamedType> saveFileList = new ArrayList<>();
		saveFileList.add(new NamedType("Collection<? extends "+ret.virtualManagerClass.foreignName(ret.virtualBasicObject)+">","boCol"));
		saveFileList.add(new NamedType("File","file"));
		Method saveFile = new Method("void","saveFile",saveFileList,
				saveFileImpl.toString());
		saveFile.exceptions.add("java.io.IOException");
		ret.virtualManagerClass.addImport("org.eclipse.emf.common.util.URI");
		ret.virtualManagerClass.addMethod(loadFile);
		ret.virtualManagerClass.addMethod(loadFileStatic);
		ret.virtualManagerClass.addMethod(saveFile);
		
		ret.virtualManagerClass.addImport("java.util.Collection");
		
		
		
		ret.virtualManagerClass.addField(new Field(new NamedType("static Manager", "instance"),"new Manager()"));
		
		
		{
			op.put(null, new OutputPackage("minterface", basePkg));
			int id = 1;
			for (Resource res: manager.getResources()) {
				OutputPackage newPkg = new OutputPackage("mvirtual"+id, basePkg);
				op.put(res.getURI().toString(),newPkg);
				++id;
				ret.createFactoryStub(res.getURI().toString(), virtual);
			}
		}
		
		for (MergeClass mcl: manager.getMergeClasses()) {
			//Interface: Enthlt Union der Attribute frs Get
			//Enthlt Intersection der Attribute frs Set/Add
			//Enthlt Union der Attribute frs Get
			//Methoden: hasXY --> gibt es das Attribut prinzipiell
			//Wenn Attribute gleich sind, werden sie gemerged (und Name angepasst)
			//Wenn Attribute unterschiedlich sind, aber gleichen Namen haben, wird name angepasst
			//Zustzlich in jeder Implementierung ein Virtuelles-Modell-Manager notwendig
			//Im Idealfall: Wenn eine Ressource vllig gleich einer anderen Ressource ist, sollte
			//Diese Implementierung vom Interface her der Ecore-Implementierung entsprechen
			
			ret.getOrCreate(mcl, null);
			
			
			for (EClass ecl: mcl.getBase()) {
				Resource res = ecl.eResource();
				ret.getOrCreate(mcl, res);
				//OutputPackage pkg = 
			}
		}
		
		for (MergeEnum menum: manager.getMergeEnums()) {
			ret.getOrCreate(menum, null);
		}
		
	
		ret.output = op;
		return ret;
	}
	

	public void outputIn(File folder) throws IOException {
		for (OutputPackage pkg: output.values()) {
			File addFile = new File(folder.getAbsolutePath()+File.separator+pkg.getFullName().replace('.', '/'));
			addFile.mkdirs();
			for (OutputClass ecl: pkg.getContainedClasses()) {
				File outputFile = new File(addFile.getAbsolutePath()+File.separator+ecl.getName()+".java");
				FileOutputStream os = new FileOutputStream(outputFile);
				StringBuilder output = new StringBuilder();
				ecl.putOut(output, 0);
				os.write(output.toString().getBytes());
				os.flush();
				os.close();
			}
		}
	}
}
