package at.cosimo.verocl.metamodel.output;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EClass;

import at.cosimo.verocl.metamodel.MergeClass;

public class PseudoOutputManager implements SomeManager {
	private Map<String,Collection<String>> superClasses = new HashMap<String,Collection<String>>();
	private Map<String,String> eclassToName = new HashMap<String,String>();
	private Map<String,String> nameToEclass = new HashMap<String,String>();
	private List<String> ecoreFiles = new ArrayList<>();
	private String baseName;
	
	public Collection<String> getClassNames() {
		return nameToEclass.keySet();
	}
	
	
	public PseudoOutputManager(String baseName, String... ecoreFiles) {
		this.baseName = baseName;
		for (String ecoreFile: ecoreFiles) {
			this.ecoreFiles.add(ecoreFile);
		}
	}
	
	public void addSubRelation(String superClass, String subClass) {
		Collection<String> scl = superClasses.get(subClass);
		if (scl == null) {
			superClasses.put(subClass, scl = new HashSet<>());
		}
		scl.add(superClass);
	}
	
	public void setNameToEClass(String name, String eclass) {
		eclassToName.put(eclass, name);
		nameToEclass.put(name, eclass);
	} 
	
	public static String getFullName(EClass cl) {
		return cl.eResource().getURI().toString()+"."+cl.eResource().getURIFragment(cl);
	}
	
	public List<String> getEcoreFiles() {
		return ecoreFiles;
	}
	
	public String getBaseName() {
		return baseName;
	}
	
	public String vmanager() {
		return baseName+"."+OutputManager.vmanager;
	}

	public String getCommonSuperType(Collection<EClass> baseClasses) {
		Set<String> commonClasses = null;
		Set<String> referenceTypes = new HashSet<>();
		for (EClass cl: baseClasses) {
			String classFullName = getFullName(cl);
			String myCl = eclassToName.get(classFullName);
			if (myCl != null) {
				referenceTypes.add(myCl);
			}
		}
		for (String curClass: referenceTypes) {
			Set<String> closedClasses = new HashSet<>();
			Set<String> openClasses = new HashSet<>();
			openClasses.add(curClass);
			closedClasses.add(curClass);
			while (!openClasses.isEmpty()) {
				Iterator<String> iter = openClasses.iterator();
				String mc = iter.next();
				iter.remove();
				Collection<String> sups = superClasses.get(mc);
				if (sups != null) {
					for (String sup: sups) {
						if (closedClasses.add(sup)) {
							openClasses.add(sup);
						}
					}
				}
			}
			if (commonClasses == null) {
				commonClasses = closedClasses;
			} else {
				commonClasses.retainAll(closedClasses);
			}
		}
		if (commonClasses == null) {
			return null;
		}
		Set<String> toDelete = new HashSet<>();
		for (String mc: commonClasses) {
			Collection<String> sups = superClasses.get(mc);
			if (sups != null) {
				toDelete.addAll(sups);
			}
		}
		commonClasses.removeAll(toDelete);
		if (commonClasses.isEmpty()) {
			return null;
		}
		return commonClasses.iterator().next();
	}

	public void addEcore(String ecore) {
		if (!this.ecoreFiles.contains(ecore)) {this.ecoreFiles.add(ecore);}
	}

}
