package at.cosimo.verocl.template;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class CheckProcessor implements TemplateProcessorIFace {
	
	private String myTemplate;
	private List<String> parameterNames;

	@Override
	public void setTemplate(String myTemplate) {
		this.myTemplate = myTemplate;
	}

	@Override
	public String getTemplate() {
		return myTemplate;
	}

	@Override
	public List<String> getParameterNames() {
		return parameterNames;
	}
	

	@Override
	public void process(String template, Map<String, Object> parameters,
			StringBuilder output, int[] curPosA) throws ParseException {
		TemplateProcessorIFace originalProcessor = (TemplateProcessorIFace)parameters.get("#.");
		
		int startVal = curPosA[0];
		int startPar = 1;
		List<Object> allPars = new ArrayList<Object>();
		for(;;) {
			++startPar;
			Object o = parameters.get("#"+(startPar));
			if (o == null && !parameters.containsKey("#"+(startPar))) {
				break;
			}
			allPars.add(o);
		}
		FunctionCall fCall = originalProcessor.getFunctionCall((String)parameters.get("#1"));
		if (fCall == null) {
			System.err.println("Function " + (String)parameters.get("#1") + " not found!");
		} else {
			Object ret = fCall.call(allPars.toArray());			
			parameters.put("#1", true);
			parameters.put("#2", ret);
		}
		
	}

	private Map<String,TemplateProcessorIFace> processorMap = new HashMap<String, TemplateProcessorIFace>();
	private Map<String,FunctionCall> functionMap = new HashMap<String, FunctionCall>();
	private Map<String,Object> parameterMap = new HashMap<String, Object>();
	
	@Override
	public Map<String, Object> getParameterMap() {
		return parameterMap;
	}
	
	@Override
	public Map<String, FunctionCall> getFunctionMap() {
		return functionMap;
	}

	@Override
	public Map<String, TemplateProcessorIFace> getTemplateProcessorMap() {
		return processorMap;
	}
	
	private TemplateProcessorIFace parent;
	
	@Override
	public void setParentProcessor(TemplateProcessorIFace parent) {
		this.parent = parent;
	}

	@Override
	public TemplateProcessorIFace getParentProcessor() {
		return parent;
	}

	public static class Attribute {
		public Attribute(String name, String type) {
			this.name = name;
			this.typ = type;
		}
		
		public String name;
		public String typ;
		
		public String getType() {
			return typ;
		}
	}
	
	public static void main(String[] args) throws ParseException, IOException {
		TemplateProcessor takeDefault = new TemplateProcessor(new File("templates/class/foralltest.java"), "cl", "attrs");
		String output = takeDefault.process("MyClass",Arrays.asList(new Attribute("count","int"), new Attribute("name","String")));
		System.out.println(output);
	}
}
