package at.cosimo.verocl.template;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class ForAllProcessor implements TemplateProcessorIFace {
	
	private String myTemplate;
	private List<String> parameterNames;

	@Override
	public void setTemplate(String myTemplate) {
		this.myTemplate = myTemplate;
	}

	@Override
	public String getTemplate() {
		return myTemplate;
	}

	@Override
	public List<String> getParameterNames() {
		return parameterNames;
	}
	
	

	@Override
	public void process(String template, Map<String, Object> parameters,
			StringBuilder output, int[] curPosA) throws ParseException {
		TemplateProcessorIFace originalProcessor = (TemplateProcessorIFace)parameters.get("#.");
		String varName = (String)parameters.get("#1");
		Collection<?> iterateOver = (Collection<?>)parameters.get("#2");
		Object originalPar = parameters.get(varName);
		
		int startVal = curPosA[0];
		
		if (iterateOver.isEmpty()) {
			parameters.put(varName, "");
			originalProcessor.process(template, parameters, new StringBuilder(), curPosA);
		} else {
			for (Object over: iterateOver) {
				curPosA[0] = startVal;
				parameters.put(varName, over);
				originalProcessor.process(template, parameters, output, curPosA);
			}
		}
		
		parameters.put(varName, originalPar);
	}

	private Map<String,TemplateProcessorIFace> processorMap = new HashMap<String, TemplateProcessorIFace>();
	private Map<String,FunctionCall> functionMap = new HashMap<String, FunctionCall>();
	
	@Override
	public Map<String, FunctionCall> getFunctionMap() {
		return functionMap;
	}

	@Override
	public Map<String, TemplateProcessorIFace> getTemplateProcessorMap() {
		return processorMap;
	}
	
	private TemplateProcessorIFace parent;
	
	@Override
	public void setParentProcessor(TemplateProcessorIFace parent) {
		this.parent = parent;
	}

	@Override
	public TemplateProcessorIFace getParentProcessor() {
		return parent;
	}

	public static class Attribute {
		public Attribute(String name, String type) {
			this.name = name;
			this.typ = type;
		}
		
		public String name;
		public String typ;
		
		public String getType() {
			return typ;
		}
	}
	
	private Map<String,Object> parameterMap = new HashMap<String, Object>();
	
	@Override
	public Map<String, Object> getParameterMap() {
		return parameterMap;
	}
	
	public static void main(String[] args) throws ParseException, IOException {
		TemplateProcessor takeDefault = new TemplateProcessor(new File("templates/class/foralltest.java"), "mcl");
		FakeMergeClass fmc = new FakeMergeClass();
		fmc.name = "Person";
		fmc.basePackage = "Blabla2";
		fmc.isAbstract = false;
		fmc.isInterface = false;
		FakeMergeStructuralFeature nameFeat = new MergeAttribute();
		nameFeat.name = "name";
		fmc.structuralFeatures = Arrays.asList(nameFeat);
		String output = takeDefault.process(fmc);
		System.out.println(output);
	}
}
