package at.cosimo.verocl.template;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.ListCellRenderer;

public class TemplateProcessor implements FunctionCallProvider, TemplateProcessorIFace {

	public TemplateProcessor() {
		
	}
	
	public TemplateProcessor(String template, String... parameterNames) {
		setTemplate(template);
		for (String pName: parameterNames) {
			this.parameterNames.add(pName);
		}
	}
	
	static {
		TemplateInserter.execute();
	}
	
	public TemplateProcessor(File templateFile, String... parameterNames) throws IOException {
		setTemplate(templateFile);
		for (String pName: parameterNames) {
			this.parameterNames.add(pName);
		}
	}
	
	private List<String> parameterNames = new ArrayList<String>();
	
	private List<TemplateProcessorIFace> subProcessors = new ArrayList<>();
	private Map<String,FunctionCall> functions = new HashMap<String, FunctionCall>();
	public Map<String,FunctionCall> getFunctionMap() {
		return functions;
	}
	private List<FunctionCallProvider> functionCallProviders = new ArrayList<FunctionCallProvider>();
	{
		functionCallProviders.add(FunctionCallProvider.defaultProvider());
	}
	
	private Map<String,TemplateProcessorIFace> templateProcessorMap = new HashMap<>();
	
	public Map<String,TemplateProcessorIFace> getTemplateProcessorMap() {
		return templateProcessorMap;
	}
	
	private List<TemplateProcessorProvider> templateProcessorProviders = new ArrayList<>();
	{
		templateProcessorProviders.add(TemplateProcessorProvider.defaultProvider());
	}
	
	
	@Override
	public FunctionCall getFunctionCall(String name) {
		//Search in this
		FunctionCall ret = functions.get(name);
		if (ret != null) {
			return ret;
		}
		for (FunctionCallProvider fcallProv: functionCallProviders) {
			ret = fcallProv.getFunctionCall(name);
			if (ret != null) {
				return ret;
			}
		}
		if (parent != null) {
			return parent.getFunctionCall(name);
		}
		return null;
	}
	
	public TemplateProcessorIFace getTemplateProcessor(String name) {
		//Search in this
		TemplateProcessorIFace ret = templateProcessorMap.get(name);
		if (ret != null) {
			return ret;
		}
		for (TemplateProcessorProvider tprocProv: templateProcessorProviders) {
			ret = tprocProv.getTemplateProcessor(name);
			if (ret != null) {
				return ret;
			}
		}
		if (parent != null) {
			return parent.getTemplateProcessor(name);
		}
		return null;
	}
	
	
	public static enum READ_TYPE {NORMAL, FUNCTION, LONG, PARSEFUNC, PARSEFUNCPAR};
	
	public Object getFieldCall(String procString, Map<String,Object> parameters, StringBuilder builder, TemplateProcessorIFace processor) {
		if (procString.trim().equals(".")) {
			return parameters;
		} else if (procString.trim().equals("..")) {
			return builder;
		} else if (procString.trim().equals(".p")) {
			return processor;
		}
		String[] spl = procString.split("[.]");
		String base = spl[0].trim();
		Object ret;
		 if (base.startsWith("\"")) {
			//It's a string
			ret = base.substring(1,base.lastIndexOf('"'));
		} else if (Character.isDigit(base.charAt(0))) {
			if (base.contains(".")) {
				ret = Double.valueOf(base);
			} else {
				if (base.endsWith("L")) {
					ret = Long.valueOf(base);
				} else {
					ret = Integer.valueOf(base);
				}
			}
		} else {
			if ("true".equals(base) || "false".equals(base)) {
				ret = Boolean.valueOf(base);
			} else {
				ret = getParameter(spl[0].trim(),parameters);
			}
		}
		
		mainLoop: for (int i = 1; i< spl.length; ++i) {
			if (ret == null) {return "";}
			Class<?> cl = ret.getClass();
			for (Method method: cl.getMethods()) {
				if (method.getParameterCount() > 0) {
					continue;
				}
				if (method.getName().toLowerCase().equals(spl[0].trim().toLowerCase()) || method.getName().startsWith("get") && method.getName().toLowerCase().substring(3).equals(spl[i].trim().toLowerCase())) {
					try {
						ret = method.invoke(ret);
					} catch (IllegalAccessException | IllegalArgumentException
							| InvocationTargetException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					continue mainLoop;
				}
			}
			
			do {
				for (Field field: cl.getDeclaredFields()) {
					if (spl[i].trim().equals(field.getName())) {
						field.setAccessible(true);
						try {
							ret = field.get(ret);
						} catch (IllegalArgumentException
								| IllegalAccessException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
						continue mainLoop;
					}
				}
				cl = cl.getSuperclass();
			} while (cl != null);
			return "";
		}
		return ret;
	}
	
	private String myTemplate = null;
	
	/* (non-Javadoc)
	 * @see at.cosimo.verocl.template.TemplateProcessorIFace#setTemplate(java.lang.String)
	 */
	@Override
	public void setTemplate(String myTemplate) {
		this.myTemplate = myTemplate;
	}
	
	public String getTemplate() {
		return myTemplate;
	}
	
	
	
	
	
	
	/* (non-Javadoc)
	 * @see at.cosimo.verocl.template.TemplateProcessorIFace#process(java.lang.String, java.util.Map, java.lang.StringBuilder, int[])
	 */
	@Override
	public void process(String template, Map<String,Object> parameters, StringBuilder output, int[] curPosA) throws ParseException {
		
		int curPos = curPosA[0];
		mainLoop: for (; curPos < template.length(); ) {
			char cur = template.charAt(curPos);
			if (cur == '#') {
				//Process something
				StringBuilder processString = new StringBuilder();
				READ_TYPE readType = READ_TYPE.NORMAL;
				++curPos;
				cur = template.charAt(curPos);
				if (cur == '<') {
					readType = READ_TYPE.LONG;
				} else if (cur == ':') {
					readType = READ_TYPE.FUNCTION;
				} else if (cur == '!') {
					readType = READ_TYPE.PARSEFUNCPAR;
				} else {
					if (cur == '#') {
						output.append(cur);
						continue mainLoop;
					}
					if (Character.isAlphabetic(cur)) {
						readType = READ_TYPE.NORMAL;
					} else if (Character.isDigit(cur)){
						readType = READ_TYPE.PARSEFUNC;
					}
					processString.append(cur);
										
				}
				readLoop: for (++curPos; curPos < template.length(); ++curPos) {
					cur = template.charAt(curPos);
					switch(readType) {
					case NORMAL:
					case PARSEFUNC:
						if (Character.isAlphabetic(cur) || Character.isDigit(cur)) {
							processString.append(cur);
						} else {
							break readLoop;
						}
						break;
					case FUNCTION:
					case PARSEFUNCPAR:
						if (Character.isAlphabetic(cur) || Character.isDigit(cur) || cur == '.' || cur == '(' || cur == ',' || cur == '"' ) {
							processString.append(cur);
						} else {
							if (cur == ')') {
								++curPos;
								break readLoop;
							}
							if (!Character.isWhitespace(cur)) {
								break readLoop;
							}
						}
						break;
					case LONG:
						if (cur == '>') {
							++curPos;
							break readLoop;
						}
						processString.append(cur);
						break;						
					}
				}
				if (readType == READ_TYPE.LONG) {
					if (processString.length() == 0) {
						continue mainLoop;
					}
					if (Character.isAlphabetic(processString.charAt(0))) {
						readType = READ_TYPE.NORMAL;
					} else if (Character.isDigit(processString.charAt(0))) {
						readType = READ_TYPE.PARSEFUNC;
					} else if (processString.charAt(0) == ':'){
						readType = READ_TYPE.FUNCTION;
						processString.delete(0, 1);
					} else if (processString.charAt(0) == '!'){
						readType = READ_TYPE.PARSEFUNCPAR;
						processString.delete(0, 1);
					} else {
						throw new ParseException("Could not process " + processString, curPos);
					}
				}
				switch (readType) {
				case NORMAL:
					if ("end".equals(processString.toString())) {
						curPosA[0] = curPos;
						return;
					}
					//Just replace thing by a template
					Object ret = getFieldCall(processString.toString(), parameters, output, this);
					output.append((ret==null)?"":String.valueOf(ret));
					break;
				case PARSEFUNC:
					Integer processorNmb = Integer.valueOf(processString.toString().trim());
					curPosA[0] = curPos;
					subProcessors.get(processorNmb-1).process(template, parameters, output, curPosA);
					curPos = curPosA[0];
					break;
				case FUNCTION:
				case PARSEFUNCPAR:
					String proString = processString.toString();
					if (proString.lastIndexOf(')') != -1) {
						proString = proString.substring(0, proString.lastIndexOf(')'));
					}
					String[] spl = proString.toString().split("[(]",2);
					String funcName = spl[0].trim();
					String[] paramStrings;
					if (spl.length == 1 || spl[1].trim().isEmpty()) {
						paramStrings = new String[]{};
					} else {
						paramStrings = spl[1].split("[,]");
					}
					if ("end".equals(funcName)) {
						curPosA[0] = curPos;
						return;
					}
					
					Object[] objs = new Object[paramStrings.length];
					for (int i = 0; i < paramStrings.length; ++i) {
						objs[i] = getFieldCall(paramStrings[i], parameters, output, this);
					}
					if (readType == READ_TYPE.FUNCTION) {
						FunctionCall fCall = getFunctionCall(funcName);
						if (fCall == null) {
							throw new ParseException("Could not find function with name " + funcName, curPos);
						}
						Object addObj = fCall.call(objs);
						output.append((addObj==null)?"":String.valueOf(addObj));
					} else {
						TemplateProcessorIFace subProc = getTemplateProcessor(funcName);
						if (subProc == null) {
							throw new ParseException("Could not find template processor with name " + funcName, curPos);
						}
						subProc.setParentProcessor(this);
						parameters.put("#.", this);
						for (int i = 0; i < paramStrings.length; ++i) {
							parameters.put("#"+(i+1), objs[i]);
						}
						curPosA[0] = curPos;
						subProc.process(template,parameters,output,curPosA);
						curPos = curPosA[0];
					}
					break;
				default:
					throw new RuntimeException();
				}
			} else {
				output.append(cur);
				++curPos;
			}
		}
	}
	
	public static void main(String[] test) throws ParseException, IOException {
		TemplateProcessorIFace proc = new TemplateProcessor(new File("templates/BasicObject.java"),"package");
		System.out.println(proc.process("Blabla2"));
		
	}

	@Override
	public List<String> getParameterNames() {
		return parameterNames;
	}
	
	private Map<String,Object> parameterMap = new HashMap<String, Object>();
	
	
	@Override
	public Map<String, Object> getParameterMap() {
		return parameterMap;
	}

	
	private TemplateProcessorIFace parent;
	
	@Override
	public void setParentProcessor(TemplateProcessorIFace parent) {
		this.parent = parent; 
	}

	@Override
	public TemplateProcessorIFace getParentProcessor() {
		return parent;
	}

}
