package at.cosimo.verocl.template;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public interface TemplateProcessorIFace {

	public void setTemplate(String myTemplate);
	
	public String getTemplate();
	
	public List<String> getParameterNames();

	default Map<String,Object> namesToMap(Object... parameters) {
		Map<String,Object> objMap = new HashMap<String, Object>();
		for (int i = 0; i < parameters.length; ++i) {
			if (getParameterNames().size()>i) {
				objMap.put(getParameterNames().get(i), parameters[i]);
			} else {
				objMap.put("par"+(i+1), parameters[i]);
			}
		}
		return objMap;
	}
	
	public default TemplateProcessorIFace putProcessor(String name, TemplateProcessorIFace processor) {
		return getTemplateProcessorMap().put(name, processor);
	}
	
	
	public default FunctionCall putFunction(String name, FunctionCall func) {
		return getFunctionMap().put(name, func);
	}
	
	public default Object putParameter(String name, Object parameter) {
		return getParameterMap().put(name, parameter);
	}
	
	public void setParentProcessor(TemplateProcessorIFace parent);
	
	public TemplateProcessorIFace getParentProcessor();
	
	public static Map<String,String> templateMap = new HashMap<String, String>();
	public static void addTemplateMap(String key, String value) {
		templateMap.put(key, value);
	}

	public default void setTemplate(File file) {
		try {
		BufferedReader reader;
		String oriPath = file.getPath();
		InputStream is = null;
		if (file.exists()) {
			is = new FileInputStream(file);
		}
		if (is == null) {
			ClassLoader classLoader = getClass().getClassLoader();
			try {is = classLoader.getResourceAsStream(oriPath);} catch (Exception e) {
			}
			if (is == null) {
				try {is = classLoader.getResourceAsStream("/"+oriPath);} catch (Exception e) {
				}
			}
			if (is == null) {
				try {is = getClass().getResourceAsStream("/"+oriPath);} catch (Exception e) {
				}
			}
			if (is == null) {
				try {is = getClass().getResourceAsStream(oriPath);} catch (Exception e) {
				}
			}
			if (is == null) {
				String str = templateMap.get(oriPath);
				if (str != null) {
					is = new ByteArrayInputStream(str.getBytes());
				}
			}
			if (is == null) {
				throw new RuntimeException("Could not load file " + file.getPath() + " where oriPath: " + oriPath);
			}
		}
		reader = new BufferedReader(new InputStreamReader(is));
		StringBuilder templateBuilder = new StringBuilder();
		String s;
		while ((s = reader.readLine()) != null) {
			templateBuilder.append(s);
			templateBuilder.append("\n");
		}
		reader.close();
		setTemplate(templateBuilder.toString());
		} catch (Exception e) {
			setTemplate(e.getMessage());
		}
	}

	public Map<String,FunctionCall> getFunctionMap();
	
	public default Object getParameter(String name) {
		Object ret = getParameterMap().get(name);
		if (ret == null) {
			if (getParameterMap().containsKey(name)) {
				return null;
			}
		}
		if (getParentProcessor() != null) {
			return getParentProcessor().getParameter(name);
		}
		return ret;
	}
	
	public default Object getParameter(String name, Map<String,Object> localMap) {
		Object ret = localMap.get(name);
		if (ret == null && !localMap.containsKey(name)) {
			return getParameter(name);
		}
		return ret;
	}
	
	public Map<String,Object> getParameterMap();
	
	
	public Map<String,TemplateProcessorIFace> getTemplateProcessorMap();
	
	public default FunctionCall getFunctionCall(String name) {
		//Search in this
		FunctionCall ret = getFunctionMap().get(name);
		if (ret != null) {
			return ret;
		}
		if (getParentProcessor() != null) {
			return getParentProcessor().getFunctionCall(name);
		}
		return null;
	}
	
	public default TemplateProcessorIFace getTemplateProcessor(String name) {
		//Search in this
		TemplateProcessorIFace ret = getTemplateProcessorMap().get(name);
		if (ret != null) {
			return ret;
		}
		if (getParentProcessor() != null) {
			return getParentProcessor().getTemplateProcessor(name);
		}
		return null;
	}
		
	public default String process(Object... parameters) throws ParseException {
		return process(getTemplate(),namesToMap(parameters));
	}
	
	public default String process(Map<String,Object> parameters) throws ParseException {
		return process(getTemplate(),parameters);
	}
	
	public default String process(String template, Map<String,Object> parameters) throws ParseException {
		StringBuilder ret = new StringBuilder();
		process(template, parameters, ret, new int[]{0});
		return ret.toString();
	}

	
	public default void process(Map<String,Object> parameters, StringBuilder output, int[] curPosA) throws ParseException {
		process(getTemplate(),parameters,output,curPosA);
	}
	
	public default void process(StringBuilder output, int[] curPosA, Object... parameters) throws ParseException {
		process(getTemplate(),namesToMap(parameters),output,curPosA);
	}
	

	public default void process(String template, StringBuilder output, int[] curPosA, Object... parameters) throws ParseException {
		process(getTemplate(),namesToMap(parameters),output,curPosA);
	}

	public void process(String template,
			Map<String, Object> parameters, StringBuilder output, int[] curPosA)
			throws ParseException;

}