package at.cosimo.verocl.util.ecore;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.tools.JavaFileObject;

import at.cosimo.verocl.metamodel.ManagerManager;
import at.cosimo.verocl.metamodel.MergeManager;
import at.cosimo.verocl.metamodel.output.OutputClass;
import at.cosimo.verocl.metamodel.output.OutputManager;
import at.cosimo.verocl.metamodel.output.OutputPackage;
import at.cosimo.verocl.metamodel.output.PseudoOutputManager;

public class AnnotationIface {

	
	public static class EcoreInput {
		public EcoreInput(String name, String... files) {
			this.name = name;
			this.files = files;
		}
		
		private String name;
		private String[] files;
	}
	
	public static void doMain(BufferedReader input, BufferedWriter output) throws IOException {
		List<EcoreInput> ecoreInputs = new ArrayList<>();
		String s;
		String name = null;
		List<String> files = new ArrayList<>();
		List<PseudoOutputManager> poms = new ArrayList<>();
		PseudoOutputManager curPom = null;
		while ((s = input.readLine()) != null) {
			if (s.isEmpty()) {continue;}
			if (s.startsWith("+")) {
				if (name != null) {
					ecoreInputs.add(new EcoreInput(name, files.toArray(new String[]{})));
				}
				files = new ArrayList<>();
				name = s.substring(1);
			} else if (s.startsWith("#")) {
				files.add(s.substring(1));
			} else if (s.startsWith("?")) {
				if (s.startsWith("?=")) { //Neue PackageInfo
					String baseName = s.substring(2);
					curPom = new PseudoOutputManager(baseName);
					poms.add(curPom);
				} else if (s.startsWith("?+")) { //Ecore-File in Package-Info
					curPom.addEcore(s.substring(2));
				} else if (s.startsWith("?:")) { //EClass-URI to Class mapping
					String[] spl = s.substring(2).split(":",2);
					curPom.setNameToEClass(spl[0],spl[1]);
				} else if (s.startsWith("?<")) { //Subtype relationship
					String[] spl = s.substring(2).split("<",2);
					curPom.addSubRelation(spl[1], spl[0]);
				}
			}
		}
		if (name != null)  {
			ecoreInputs.add(new EcoreInput(name, files.toArray(new String[]{})));
		}
		
		ManagerManager man = new ManagerManager();
		man.setPseudoOM(poms);
		
		Map<String, MergeManager> mmmap = new HashMap<>();
		for (PseudoOutputManager pseudoOOm: poms) {
			MergeManager mergeMan = man.createManager(pseudoOOm.getBaseName(), pseudoOOm.getEcoreFiles().toArray(new String[]{}));
			mmmap.put(pseudoOOm.getBaseName(),mergeMan);
		}
		for (EcoreInput ip: ecoreInputs) {
			MergeManager mergeMan = man.createManager(ip.name, ip.files);
			mmmap.put(ip.name,mergeMan);
		}
		
		for (Entry<String,MergeManager> entry: mmmap.entrySet()) {
			OutputManager om = ManagerManager.getOrCreateOutput(entry.getValue());
			for (Entry<String,OutputPackage> entr: om.getOutput().entrySet()) {
				for (OutputClass cl: entr.getValue().getContainedClasses()) {
					StringBuilder builder = new StringBuilder();
					cl.putOut(builder, 0);
					output.write("+"+cl.getPackage().getFullName()+"."+cl.getName()+"\n");
					output.write("#"+builder.toString().replace("\n", "\n#")+"\n");
				}
			}
			//General output information
			String str = "+"+om.getPackageName()+".info-txt"+"\n";
			output.write(str);
			//1. Ecore-Files
			for (String ecore: entry.getValue().getResourcesString()) {
				output.write("#+"+ecore+"\n");
			}
			//2. Eclass to class mapping
			for (Entry<String,String> entr: om.classToEclassMap().entrySet()) {
				output.write("#:" + entr.getKey()+":"+entr.getValue()+"\n");
			}
			//3. Subtype relationship
			for (Entry<String,Collection<String>> entr: om.subtypeMap().entrySet()) {
				for (String superClass: entr.getValue()) {
					output.write("#<"+entr.getKey()+"<"+superClass+"\n");
				}
			}
		}
		
	}
	
	public static void main(String[] args) throws IOException {
		BufferedReader input = new BufferedReader(new InputStreamReader(System.in));
		BufferedWriter output = new BufferedWriter(new PrintWriter(System.out));
		doMain(input, output);
		input.close();
		output.flush();
		output.close();
	}
}
