package at.cosimo.verocl.util.ecore;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.impl.EClassImpl;

public class VirtualManager {
	
	public static class Parameters {
		public Parameters(List<Object> parameterValues) {
			this.parameters = parameterValues;
		}
		
		private List<Object> parameters = new ArrayList<>();
		
		public boolean equals(Object o) {
			return o != null && (o.getClass() == Parameters.class) && (((Parameters)o).parameters.equals(parameters));
		}
		
		public int hashCode() {
			return parameters.hashCode();
		}
	}
	
	public interface Instanciator {
		public EObject instanciate(Parameters par);
	}
	
	public static class InstantiationIndex {
		private Map<Parameters,EObject> parametersToEobject = new HashMap<>();
		private Instanciator inst;
		private EClass eclass;
		
		public InstantiationIndex(EClass cl, Instanciator inst) {
			this.eclass = cl;
			this.inst = inst;
		}
		
		
		public EObject getOrCreate(List<Object> parameters) {
			Parameters par = new Parameters(parameters);
			EObject ret = parametersToEobject.get(par);
			if (ret == null) {
				ret = inst.instanciate(par);
			}
			return ret;
		}
	}
	
	
	private Map<String,InstantiationIndex> classToInstanceMap = new HashMap<>();
	
	public EObject getOrCreate(EClass cl, List<Object> param) {
		InstantiationIndex ind = classToInstanceMap.get(cl.getName());
		if (ind == null) {
			System.err.println("No instanciationindex for class " + cl.getName());
			return null;
		}
		return ind.getOrCreate(param);
	}
 
}
