package #<outputClass.pkg.name>;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.eclipse.emf.ecore.EObject;

public class CollectionProxy<T extends BasicObject> implements Collection<T>{
	private Collection<EObject> proxy;
	private Manager manager;
	
	public CollectionProxy(Manager manager, Collection<EObject> proxy) {
		this.proxy = proxy; 
		this.manager = manager;
	}
	
	public Collection<EObject> getReal() {
		return proxy;
	}

	@Override
	public boolean add(T e) {
		return proxy.add(e.baseObject());
	}

	@Override
	public boolean addAll(Collection<? extends T> c) {
		boolean ret = false;
		for (T t: c) {
			ret |= add(t);
		}
		return ret;
	}

	@Override
	public void clear() {
		proxy.clear();
	}

	@Override
	public boolean contains(Object o) {
		return (o instanceof BasicObject) && proxy.contains(((BasicObject)o).baseObject());
	}

	@Override
	public boolean containsAll(Collection<?> c) {
		for (Object o: c) {
			if (!contains(o)) {
				return false;
			}
		}
		return true;
	}

	@Override
	public boolean isEmpty() {
		return proxy.isEmpty();
	}

	@Override
	public Iterator<T> iterator() {
		return new Iterator<T>() {
			Iterator<EObject> proxyIter = proxy.iterator();

			@Override
			public boolean hasNext() {
				return proxyIter.hasNext();
			}

			@Override
			public T next() {
				return (T) manager.getOrCreate(proxyIter.next());
			}
			
			public void remove() {
				proxyIter.remove();
			}
			
		};
	}

	@Override
	public boolean remove(Object o) {
		return (o instanceof BasicObject) && proxy.remove(((BasicObject)o).baseObject());
	}

	@Override
	public boolean removeAll(Collection<?> c) {
		boolean ret = false;
		for (Object o: c) {
			ret|= remove(o);
		}
		return ret;
	}

	@Override
	public boolean retainAll(Collection<?> c) {
		//This has not good performance
		List<EObject> base = new ArrayList<>(c.size());
		for (Object o: c) {
			if (o == null || o instanceof BasicObject) {
				base.add(((BasicObject)o).baseObject());
			}
		}
		return proxy.retainAll(base);
	}

	@Override
	public int size() {
		return proxy.size();
	}

	@Override
	public Object[] toArray() {
		Object[] ret = proxy.toArray();
		for (int i = 0; i < ret.length; ++i) {
			if (ret[i] != null) {
				ret[i] = manager.getOrCreate((EObject)ret[i]);
			}
		}
		return ret;
		
	}

	@Override
	public <T> T[] toArray(T[] a) {
		Object[] subArr = proxy.toArray();
		if (subArr.length <= a.length) {
			for (int i = 0; i < subArr.length; ++i) {
				a[i] = (T)manager.getOrCreate((EObject)subArr[i]);
			}
			for (int i = subArr.length; i < a.length; ++i) {
				a[i] = null;
			}
			return a;
		} else {
			T[] ret = Arrays.copyOf(a, subArr.length);
			for (int i = 0; i < subArr.length; ++i) {
				ret[i] = (T)manager.getOrCreate((EObject)subArr[i]);
			}
			return ret;
		}
	}

}
